theory GTT
  imports TA.Tree_Automata Tree_Automata_Utils Basic_Utils
begin

abbreviation map_both where
  "map_both f \<equiv> map_prod f f"

(* A \<G> consists of a set of interface states and
   a set of rules for automaton \<A> and one for \<B>. *)
type_synonym ('q, 'f) gtt = "('q, 'f) ta \<times> ('q, 'f) ta"

abbreviation gtt_rules where
  "gtt_rules \<G> \<equiv> ta_rules (fst \<G>) \<union> ta_rules (snd \<G>)"
abbreviation gtt_eps where
  "gtt_eps \<G> \<equiv> ta_eps (fst \<G>) \<union> ta_eps (snd \<G>)"
abbreviation gtt_states where
  "gtt_states \<G> \<equiv> ta_states (fst \<G>) \<union> ta_states (snd \<G>)"
abbreviation gtt_syms where
  "gtt_syms \<G> \<equiv> ta_syms (fst \<G>) \<union> ta_syms (snd \<G>)"

(* We define two notions of GTT acceptance, and show equivalence of each of them *)

inductive gtt_accept :: "('q, 'f) gtt \<Rightarrow> ('f, 'q) term \<Rightarrow> ('f, 'q) term \<Rightarrow> bool" for \<G> where
  refl [intro]: "gtt_accept \<G> t t"
| join [intro]: "q \<in> ta_res (fst \<G>) s \<Longrightarrow> q \<in> ta_res (snd \<G>) t \<Longrightarrow> gtt_accept \<G> s t"
| ctxt [intro]: "length ss = length ts \<Longrightarrow>
                 \<forall>i < length ss. gtt_accept \<G> (ss ! i) (ts ! i) \<Longrightarrow>
                 gtt_accept \<G> (Fun f ss) (Fun f ts)"

hide_fact (open) refl (* don't shadow HOL.refl *)

inductive gtt_accept' :: "('q, 'f) gtt \<Rightarrow> ('f, 'q) term \<Rightarrow> ('f, 'q) term \<Rightarrow> bool"
  for \<G> where
  mctxt [intro]: "length ss = length ts \<Longrightarrow> num_holes C = length ss \<Longrightarrow>
    \<forall>i < length ts. \<exists>q. q \<in> ta_res (fst \<G>) (ss ! i) \<and> q \<in> ta_res (snd \<G>) (ts ! i) \<Longrightarrow>
    gtt_accept' \<G> (fill_holes C ss) (fill_holes C ts)"

abbreviation gtt_lang where
  "gtt_lang \<G> \<equiv> Restr {(s, t). gtt_accept \<G> s t} {t. ground t}"

(* Lemmas abount gtt_accept' *)
lemma gtt_accept'_refl [simp]:
  "gtt_accept' \<G> t t"
proof -
  have "gtt_accept' \<G> (fill_holes (mctxt_of_term t) [])
                      (fill_holes (mctxt_of_term t) [])" by fastforce
  thus ?thesis by simp
qed

lemma gtt_accept'_join[simp]:
  assumes "q \<in> ta_res (fst \<G>) s" "q \<in> ta_res (snd \<G>) t"
  shows "gtt_accept' \<G> s t"
proof -
  have "gtt_accept' \<G> (fill_holes MHole [s]) (fill_holes MHole [t])"
    using assms by (intro gtt_accept'.intros) auto
  thus ?thesis by simp
qed

lemma gtt_cont_extrct:
  assumes "gtt_accept' \<G> s t"
  shows "\<exists>C ss ts. (s = fill_holes C ss) \<and> (t = fill_holes C ts) \<and>
         (length ss = length ts) \<and> (num_holes C = length ss ) \<and>
         (\<forall>i < length ts. \<exists>q. q \<in> ta_res (fst \<G>) (ss ! i) \<and>
                              q \<in> ta_res (snd \<G>) (ts ! i))"
  using assms by (auto elim!: gtt_accept'.cases)

(* The following lemma proves that the two definitions of gtt_accept are equivalent *)

lemma gtt_accept_equiv:
  "gtt_accept \<G> s t \<longleftrightarrow> gtt_accept' \<G> s t"
proof (rule iffI)
  assume "gtt_accept \<G> s t"
  thus "gtt_accept' \<G> s t"
  proof (induction rule: gtt_accept.induct)
    case (ctxt ss ts f)
    let ?Prop = "\<lambda>x i. \<exists>C ss' ts'. x = (C, (ss', ts')) \<and> length ss' = length ts' \<and>
        (num_holes C = length ss') \<and>
        (ss ! i = fill_holes C ss') \<and> (ts ! i = fill_holes C ts') \<and>
        (\<forall>j < length ts'. \<exists>q. q \<in> ta_res (fst \<G>) (ss' ! j) \<and>
                              q \<in> ta_res (snd \<G>) (ts' ! j)) \<and>
        gtt_accept' \<G> (fill_holes C ss') (fill_holes C ts')"
    { fix i
      assume "i < length ss"
      hence "gtt_accept' \<G> (ss ! i) (ts ! i)" using ctxt(2) by blast
      hence "\<exists>x C ss' ts'. x = (C, (ss', ts')) \<and> length ss' = length ts' \<and>
        (num_holes C = length ss') \<and>
        (ss ! i = fill_holes C ss') \<and> (ts ! i = fill_holes C ts') \<and>
        (\<forall>j < length ts'. \<exists>q. q \<in> ta_res (fst \<G>) (ss' ! j) \<and>
                              q \<in> ta_res (snd \<G>) (ts' ! j)) \<and>
        gtt_accept' \<G> (fill_holes C ss') (fill_holes C ts')"
        by (auto elim!: gtt_accept'.cases)
    } note inner = this
    then obtain xs where xs: "length xs = length ss \<and> (\<forall>i<length ss. ?Prop (xs ! i) i)"
      using construct_exI'[of "length ss" ?Prop] by blast
    hence ss_i_ts_i:
      "\<forall>i<length xs. ss ! i = fill_holes ((map fst xs) ! i) ((map (fst \<circ> snd) xs) ! i) \<and>
                     ts ! i = fill_holes ((map fst xs) ! i) ((map (snd \<circ> snd) xs) ! i) \<and>
                     num_holes ((map fst xs) ! i) = length ((map (fst \<circ> snd) xs) ! i) \<and>
                     num_holes ((map fst xs) ! i) = length ((map (snd \<circ> snd) xs) ! i)"
       by (smt comp_apply fst_conv length_map map_nth nth_map snd_conv)
    define ss' ts' where ss': "ss' = concat (map (fst \<circ> snd) xs)" and
                         ts': "ts' = concat (map (snd \<circ> snd) xs)"
    hence lengths_eq: "length ss' = length ts'" using ss_i_ts_i
      by (metis (no_types, lifting) eq_length_concat_nth length_map)
    have map_eq: "map (num_holes \<circ> fst) xs = map (length \<circ> (fst \<circ> snd)) xs"
      using ss_i_ts_i by (metis (no_types, lifting) List.map.compositionality map_eq_conv')
    hence num_holes: "num_holes (MFun f (map fst xs)) = length ss'"
      using ss' ts' by (auto simp: length_concat nth_map[of _ xs] map_eq)
    have "\<forall>i<length xs. \<forall>j < length ((map (snd \<circ> snd) xs) ! i).
                       \<exists>q. q \<in> ta_res (fst \<G>) (((map (fst \<circ> snd) xs) ! i) ! j) \<and>
                           q \<in> ta_res (snd \<G>) (((map (snd \<circ> snd) xs) ! i) ! j)"
      using xs by (smt comp_apply fst_conv length_map map_nth nth_map snd_conv)
    hence join_state: "\<forall>j < length ts'. \<exists>q. q \<in> ta_res (fst \<G>) (ss' ! j) \<and>
                                            q \<in> ta_res (snd \<G>) (ts' ! j)"
      using ss_i_ts_i (*nth_concat_two_lists[of _ "map (fst \<circ> snd) xs" "map (snd \<circ> snd) xs"]*)
        unfolding ss' ts' by (smt length_map nth_concat_two_lists ss_i_ts_i)
    have "i < length (map (num_holes \<circ> fst) xs) \<Longrightarrow>
          length (map (fst \<circ> snd) xs ! i) = map (num_holes \<circ> fst) xs ! i" for i
      using ss_i_ts_i by fastforce
    hence f_ss: "Fun f ss = fill_holes (MFun f (map fst xs)) ss'"
      using xs ss_i_ts_i partition_by_concat_id[of "map (fst \<circ> snd) xs" "map (num_holes \<circ> fst) xs"]
      unfolding ss' by simp (smt atLeastLessThan_iff map_eq_conv map_nth set_upt ss_i_ts_i)
    have "i < length (map (num_holes \<circ> fst) xs) \<Longrightarrow>
          length (map (snd \<circ> snd) xs ! i) = map (num_holes \<circ> fst) xs ! i" for i
      using ss_i_ts_i by fastforce
    hence f_ts: "Fun f ts = fill_holes (MFun f (map fst xs)) ts'"
      using xs ss_i_ts_i partition_by_concat_id[of "map (snd \<circ> snd) xs" "map (num_holes \<circ> fst) xs"]
      unfolding ss' ts' \<open>length ss = length ts\<close>
      by simp (smt atLeastLessThan_iff map_eq_conv map_nth set_upt ss_i_ts_i)
    show ?case using join_state lengths_eq num_holes unfolding f_ss f_ts
      by (intro gtt_accept'.intros) auto
  qed auto
next
  assume "gtt_accept' \<G> s t"
  thus "gtt_accept \<G> s t"
  proof (induction rule: gtt_accept'.induct)
    case (mctxt ss ts C) thus ?case
    proof (induction C arbitrary: ss ts)
      case MHole
      then obtain s' t' where "ss = [s']" "ts = [t']" by (cases ss; cases ts) fastforce+
      thus ?case using MHole(3) by auto
    next
      case (MFun f Cs)
      moreover { fix i
        assume asm: "i < length Cs"
        hence lengths_eq: "length (partition_holes ss Cs ! i) = length (partition_holes ts Cs ! i)"
          using MFun(2,3) by auto
        have num_holes_length: "num_holes (Cs ! i) = length (partition_holes ss Cs ! i)"
          using asm MFun(3) by auto
        have join: "\<forall>j<length (partition_holes ts Cs ! i).
              \<exists>q. q \<in> ta_res (fst \<G>) ((partition_holes ss Cs ! i) ! j) \<and>
                  q \<in> ta_res (snd \<G>) ((partition_holes ts Cs ! i) ! j)"
          using MFun(2,3,4)
          by (smt asm length_map lengths_eq nth_map num_holes.simps(3) num_holes_length
              partition_by_nth_nth)
        have "gtt_accept \<G> (fill_holes (Cs ! i) (partition_holes ss Cs ! i))
                            (fill_holes (Cs ! i) (partition_holes ts Cs ! i))"
          using asm MFun(1)[OF _ lengths_eq num_holes_length join] by auto
      }
      ultimately show ?case using length_upt[of 0 "length Cs"] unfolding fill_holes.simps
        by (intro gtt_accept.intros(3)) auto
    qed auto
  qed
qed

text \<open>Inverse of GTTs.\<close>

lemma gtt_accept_swap [simp]:
  "gtt_accept (prod.swap \<G>) s t \<longleftrightarrow> gtt_accept \<G> t s"
  unfolding gtt_accept_equiv
  by (intro iffI; elim gtt_accept'.cases; metis gtt_accept'.intros fst_swap snd_swap)

lemma gtt_lang_swap:
  "(gtt_lang (A, B))\<inverse> = gtt_lang (B, A)"
  using gtt_accept_swap[of "(A, B)"] by auto

lemma accept'_closed_ctxt:
  assumes "length ss = length ts"
  and "\<forall>i < length ts. gtt_accept' \<G> (ss ! i) (ts ! i)"
  and "num_holes C = length ss"
  shows "gtt_accept' \<G> (fill_holes C ss) (fill_holes C ts)"
  using assms
proof (induction C arbitrary: ss ts)
  case (MVar x)
  then show ?case by auto
next
  case MHole
  then show ?case
    by (metis One_nat_def fill_holes_MHole lessI num_holes.simps(2))
next
  case (MFun f Cs)
  let ?sss = "(partition_holes ss Cs)"
  let ?tss = "(partition_holes ts Cs)"
  obtain Css Cts where Css_Cts:"fill_holes (MFun f Cs) ss = Fun f Css"
    "fill_holes (MFun f Cs) ts = Fun f Cts"
    "length Cts = length Css"
    by simp
  {fix i
   assume loc_asm: "i < length Cs"
   hence "length (?sss ! i) = length (?tss ! i)" using assms(1) MFun by simp
   moreover have "num_holes (Cs ! i) = length (?sss ! i)" using assms(2) loc_asm MFun by force
   moreover have "\<forall>j < length (?tss ! i). gtt_accept' \<G> ((?sss ! i) ! j)  ((?tss ! i) ! j)"
      using loc_asm MFun assms(3) by (metis calculation(1) calculation(2) length_map nth_map
                  num_holes.simps(3) partition_by_nth_nth(1) partition_by_nth_nth(2))
   ultimately have "gtt_accept' \<G> (fill_holes (Cs ! i) (?sss ! i)) (fill_holes (Cs ! i) (?tss ! i))"
     using MFun(1) loc_asm by force
   hence "gtt_accept \<G> (fill_holes (Cs!i) (?sss!i)) (fill_holes (Cs ! i) (?tss ! i))"
     using gtt_accept_equiv by blast}
  hence "\<forall>i < length Cts. gtt_accept \<G> (Css ! i) (Cts ! i)" using Css_Cts by fastforce
  hence "gtt_accept \<G> (Fun f Css) (Fun f Cts)" using Css_Cts(3) by auto
  then show ?case using gtt_accept_equiv Css_Cts by metis
qed

(* The following Lemmas are about ta_res' *)
lemma ta_res'_id:
  "t \<in> ta_res' \<A> t"
  by (induction t) auto

lemma ta_res_res'_inclusion:
  fixes q \<A> l
  shows "(q \<in> ta_res \<A> l) \<longrightarrow> Var q \<in> ta_res' \<A> l"
  by (induction "\<A>" "l" rule: ta_res.induct) fastforce+

abbreviation mctxt_of_terms where
  "mctxt_of_terms s t \<equiv> (mctxt_of_term s) \<sqinter> (mctxt_of_term t)"
abbreviation residuals where
  "residuals s t \<equiv> inf_mctxt_args (mctxt_of_term s) (mctxt_of_term t)"

lemma term_mctxt [simp]:
  "term_of_mctxt (MFun f (map mctxt_of_term ts)) = Fun f ts"
  by (metis mctxt_of_term.simps(2) term_of_mctxt_mctxt_of_term_id)

lemma map_zip_fun:
  assumes "length xs = length ys"
  shows "map (\<lambda>(x,y). f x y) (zip (map g xs) (map g ys)) =
         map (\<lambda>(x,y). f (g x) (g y)) (zip xs ys)"
  by (simp add: map_nth_eq_conv)

lemma terms_to_common_residuals:
  fixes s t :: "('f, 'v) term"
  shows "s = fill_holes (mctxt_of_terms s t) (map term_of_mctxt (residuals s t)) \<and>
         t = fill_holes (mctxt_of_terms s t) (map term_of_mctxt (residuals t s))"
proof (induction s arbitrary: t)
  case (Var x) thus ?case
  proof (induction t)
    case (Var y)
    have "num_holes (MVar y \<sqinter> MVar x) = num_holes (MVar x \<sqinter> MVar y)" by simp
    define n_holes where "n_holes = (\<lambda> (x::'v) y. num_holes (MVar x \<sqinter> MVar y))"
    have "(n_holes y x = n_holes x y \<and> Var y = Var x) \<or>
          (n_holes y x = n_holes x y \<and> y \<noteq> x)" using n_holes_def by simp
    thus ?case by fastforce
  next
    case (Fun f ts)
    have "Fun f ts = term_of_mctxt (MFun f (map mctxt_of_term ts))"
      by (metis mctxt_of_term.simps(2) term_of_mctxt_mctxt_of_term_id)
    hence "Fun f ts = fill_holes (mctxt_of_terms (Var x) (Fun f ts))
                  (map term_of_mctxt (residuals (Fun f ts) (Var x)))"
      using term_mctxt[of f ts] by fastforce
    thus ?case by auto
  qed
next
  case (Fun g ss) thus ?case
  proof (induction t arbitrary: g ss)
    case (Var y)
    have "Fun g ss = fill_holes (mctxt_of_terms (Var y) (Fun g ss))
                  (map term_of_mctxt (residuals (Fun g ss) (Var y)))"
      using term_mctxt[of g ss] by fastforce
    thus ?case using term_mctxt[of g ss] by fastforce
  next
    case (Fun f ts) thus ?case
    proof (cases "f = g \<and> length ts = length ss")
      case True
      hence "mctxt_of_terms (Fun g ss) (Fun f ts) =
               MFun f (map (\<lambda>(s',t'). mctxt_of_terms s' t') (zip ss ts))"
        using True by (simp add: zip_nth_conv)
      let ?Cs = "map (\<lambda>(s',t'). mctxt_of_terms s' t') (zip ss ts)" and
          ?ts = "residuals (Fun g ss) (Fun f ts)" and
          ?ts' = "map (\<lambda>(s',t'). inf_mctxt_args s' t') (zip (map mctxt_of_term ss) (map mctxt_of_term ts))" and
          ?ts'' = "map (\<lambda>(s',t'). residuals s' t') (zip ss ts)" and
          ?ss = "residuals (Fun f ts) (Fun g ss)" and
          ?ss' = "map (\<lambda>(t',s'). inf_mctxt_args t' s') (zip (map mctxt_of_term ts) (map mctxt_of_term ss))" and
          ?ss'' = "map (\<lambda>(t',s'). residuals t' s') (zip ts ss)"
      { fix s' t' :: "('f, 'v) Term.term" and  ss' ts'
        assume "s' \<in> set ss'" "t' \<in> set ts'"
        hence "length (residuals s' t') = num_holes (mctxt_of_terms s' t')"
              "length (residuals s' t') = num_holes (mctxt_of_terms t' s')"
          using num_holes_inf_mctxt by (metis, metis inf_mctxt_comm)
      } note residuals_num_holes = this
      have "i < length (map num_holes ?Cs) \<Longrightarrow>
             length (?ts'' ! i) = (map num_holes ?Cs) ! i" for i
        using residuals_num_holes(1)[of _ ss _ ts] True by force
      hence "partition_by ?ts (map num_holes ?Cs) = ?ts'"
        using partition_by_concat_id[of ?ts' "map num_holes ?Cs"] True by simp
      hence part1: "map (\<lambda>(s', t'). residuals s' t') (zip ss ts) = partition_holes ?ts ?Cs"
        using True map_zip_fun by metis
      have "i < length (map num_holes ?Cs) \<Longrightarrow>
             length (?ss'' ! i) = (map num_holes ?Cs) ! i" for i
        using residuals_num_holes(2)[of _ ts _ ss] True by force
      hence "partition_by ?ss (map num_holes ?Cs) = ?ss'"
        using partition_by_concat_id[of ?ss' "map num_holes ?Cs"] True by simp
      hence part2: "map (\<lambda>(t', s'). residuals t' s') (zip ts ss) = partition_holes ?ss ?Cs"
        using True map_zip_fun by metis
      have push_map_inside: "(map term_of_mctxt) \<circ> (\<lambda>(x, y). residuals x y) =
            (\<lambda>(x, y). map term_of_mctxt (residuals x y))" by auto
      { fix i
        assume "i < length ss"
        hence ith_in_set: "ss ! i \<in> set ss" using True by simp+
        hence IH': "ss ! i = fill_holes (mctxt_of_terms (ss ! i) (ts ! i))
                                   (map term_of_mctxt (residuals (ss ! i) (ts ! i)))"
              "ts ! i = fill_holes (mctxt_of_terms (ss ! i) (ts ! i))
                                   (map term_of_mctxt (residuals (ts ! i) (ss ! i)))"
          using Fun(2)[of "ss ! i" "ts ! i"] True by (meson set_zip_leftD)+
        have "ss ! i = fill_holes (map (\<lambda>(x, y). mctxt_of_terms x y) (zip ss ts) ! i)
                         (map (\<lambda>(x, y). map term_of_mctxt (residuals x y)) (zip ss ts) ! i)"
             "ts ! i = fill_holes (map (\<lambda>(x, y). mctxt_of_terms x y) (zip ss ts) ! i)
                         (map (\<lambda>(x, y). map term_of_mctxt (residuals x y)) (zip ts ss) ! i)"
          using IH' True nth_map[of i "zip ss ts"] \<open>i < length ss\<close> by force+
      } note inner = this
      have "Fun g ss = fill_holes (mctxt_of_terms (Fun g ss) (Fun f ts))
                                  (map term_of_mctxt (residuals (Fun g ss) (Fun f ts)))"
        using True part1[symmetric]
        by (simp add: map_map_partition_by[symmetric] map_zip_fun push_map_inside)
           (smt inner(1) atLeastLessThan_iff map_eq_conv map_nth set_upt) (*TODO*)
      moreover have "Fun f ts = fill_holes (mctxt_of_terms (Fun g ss) (Fun f ts))
                                  (map term_of_mctxt (residuals (Fun f ts) (Fun g ss)))"
        using True part2[symmetric]
        by (simp add: map_map_partition_by[symmetric] map_zip_fun push_map_inside)
           (smt inner(2) atLeastLessThan_iff map_eq_conv map_nth set_upt) (*TODO*)
      ultimately show ?thesis using True by blast
    next
      case False
      thus ?thesis using term_mctxt[of f ts] term_mctxt[of g ss] by fastforce
    qed
  qed
qed


lemma gtt_accept_exI:
  assumes "gtt_accept \<G> s t" "ground s" "ground t"
  shows "\<exists>u. u \<in> ta_res' (fst \<G>) s \<and> u \<in> ta_res' (snd \<G>) t"
  using assms
proof (induction)
  case (refl t)
  thus ?case using ta_res'_id by blast
next
  case (join q s t)
  thus ?case using ta_res_res'_inclusion by fast
next
  case (ctxt ss ts f)
  { fix i
    assume asm: "i < length ss"
    hence "ground (ts ! i)" "ground (ss ! i)" using ctxt by auto
    hence "gtt_accept \<G> (ss ! i) (ts ! i)"
          "\<exists>u\<^sub>i. u\<^sub>i \<in> ta_res' (fst \<G>) (ss ! i) \<and> u\<^sub>i \<in> ta_res' (snd \<G>) (ts ! i)"
      using asm ctxt(1,2) by blast+
    hence "\<exists>us. (us ! i) \<in> ta_res' (fst \<G>) (ss ! i) \<and> (us ! i) \<in> ta_res' (snd \<G>) (ts ! i)"
      using asm by (metis nth_list_update_eq)
  }
  hence inner: "\<forall>i < length ss.
           \<exists>us. (us ! i) \<in> ta_res' (fst \<G>) (ss ! i) \<and> (us ! i) \<in> ta_res' (snd \<G>) (ts ! i)"
    by auto
  have "\<exists>us. length us = length ss \<and> (\<forall>i < length ss.
            (us ! i) \<in> ta_res' (fst \<G>) (ss ! i) \<and> (us ! i) \<in> ta_res' (snd \<G>) (ts ! i))"
    using construct_exI[OF inner] by auto
  then obtain us where "length us = length ss \<and> (\<forall>i < length ss.
            (us ! i) \<in> ta_res' (fst \<G>) (ss ! i) \<and> (us ! i) \<in> ta_res' (snd \<G>) (ts ! i))"
    by blast
  hence "Fun f us \<in> ta_res' (fst \<G>) (Fun f ss) \<and>
         Fun f us \<in> ta_res' (snd \<G>) (Fun f ts)" using ctxt(1) by fastforce
  thus ?case by blast
qed

lemma ta_res_stable:
  assumes "q \<in> ta_res \<A> t" "ta_rules \<A> \<subseteq> ta_rules \<A>'" "ta_eps \<A> \<subseteq> ta_eps \<A>'"
  shows "q \<in> ta_res \<A>' t"
  by (metis assms subsetD[OF ta_res_mono'])

lemma gtt_accept_mono:
  assumes "ta_eps (fst \<G>) \<subseteq> ta_eps (fst \<G>')" "ta_rules (fst \<G>) \<subseteq> ta_rules (fst \<G>')"
    "ta_eps (snd \<G>) \<subseteq> ta_eps (snd \<G>')" "ta_rules (snd \<G>) \<subseteq> ta_rules (snd \<G>')"
    "gtt_accept \<G> s t"
  shows "gtt_accept \<G>' s t"
  using assms(5) unfolding gtt_accept_equiv
  by (elim gtt_accept'.cases) (metis gtt_accept'.intros assms(1-4) subsetD[OF ta_res_mono'])

lemma gtt_accept_stable:
  assumes "gtt_accept (\<A>, \<B>) l r"
    "ta_rules \<A> \<subseteq> ta_rules \<A>'" "ta_rules \<B> \<subseteq> ta_rules \<B>'"
    "ta_eps \<A> \<subseteq> ta_eps \<A>'" "ta_eps \<B> \<subseteq> ta_eps \<B>'"
  shows "gtt_accept (\<A>', \<B>') l r"
  using gtt_accept_mono[of "(\<A>, \<B>)" "(\<A>', \<B>')" l r] assms by simp

lemma ta_res_ta_final_ignored:
  assumes "q \<in> ta_res (ta.make Q\<^sub>f \<Delta> \<E>) t"
  shows "q \<in> ta_res (ta.make Q\<^sub>f' \<Delta> \<E>) t"
  using assms(1) unfolding ta.make_def
proof (induction "\<lparr>ta_final = Q\<^sub>f, ta_rules = \<Delta>, ta_eps = \<E>\<rparr>" t arbitrary: q rule: ta_res.induct)
  case (2 f ts)
  obtain q' qs where accepted_\<Delta>: "f qs \<rightarrow> q' \<in> \<Delta>" "(q', q) \<in> \<E>\<^sup>*" "length qs = length ts"
      "\<forall>i<length ts. qs ! i \<in> ta_res \<lparr>ta_final = Q\<^sub>f, ta_rules = \<Delta>, ta_eps = \<E>\<rparr> (ts ! i)"
    using 2(2) by auto
  moreover have "\<forall>i<length ts. qs ! i \<in> ta_res \<lparr>ta_final = Q\<^sub>f', ta_rules = \<Delta>, ta_eps = \<E>\<rparr> (ts ! i)"
    using 2(1) accepted_\<Delta>(4) by auto
  ultimately show ?case using accepted_\<Delta>(1-3) by auto
qed simp

abbreviation fmap_states_gtt where
  "fmap_states_gtt f \<equiv> map_both (fmap_states_ta f)"

lemma ground_map_vars_term_simp:
  "ground t \<Longrightarrow> map_term f g t = map_term f undefined t"
  by (induct t) auto

lemma ground_map_vars_mctxt_simp:
  "ground_mctxt C \<Longrightarrow> map_vars_mctxt f C = map_vars_mctxt undefined C"
  by (induct C) auto

lemma gtt_lang_fmap_states_gtt:
  assumes "inj_on f (gtt_states \<G>)"
  shows "gtt_lang (fmap_states_gtt f \<G>) = map_both adapt_vars ` gtt_lang \<G>"
proof (intro set_eqI iffI, goal_cases lr rl)
  case (lr x)
  obtain C ss ts where x: "x = (fill_holes C ss, fill_holes C ts)"
       "length ss = num_holes C" "length ts = num_holes C"
       "\<And>i. i < num_holes C \<Longrightarrow> \<exists>q. q \<in> ta_res (fmap_states_ta f (fst \<G>)) (ss ! i) \<and>
          q \<in> ta_res (fmap_states_ta f (snd \<G>)) (ts ! i)"
       "ground (fill_holes C ss)" "ground (fill_holes C ts)"
    using lr by (auto 0 0 simp: gtt_accept_equiv elim!: gtt_accept'.cases)
  { fix i assume i: "i < num_holes C"
    obtain q where q: "q \<in> ta_res (fmap_states_ta f (fst \<G>)) (ss ! i)"
      "q \<in> ta_res (fmap_states_ta f (snd \<G>)) (ts ! i)" using x(4)[OF i] by blast
    have g: "ground (ss ! i)" "ground (ts ! i)" using x(2,3,5,6) by (auto simp: ground_fill_holes i)
    then have "\<exists>p. p \<in> ta_res (fst \<G>) (adapt_vars (ss ! i)) \<and> p \<in> ta_res (snd \<G>) (adapt_vars (ts ! i))"
      using assms[unfolded inj_on_Un] assms x(2,3,5,6) q
         ta_res_fmap_states_inv[of f "fst \<G>" "ss ! i" q] ta_res_fmap_states_inv[of f "snd \<G>" "ts ! i" q]
        inj_onD[OF assms _ UnI1[OF subsetD[OF ta_res_states]] UnI2[OF subsetD[OF ta_res_states]],
          of _ _ "adapt_vars (ss ! i)" "adapt_vars (ts ! i)"]
      by (auto) }
  then have "gtt_accept' \<G> (adapt_vars (fill_holes C ss)) (adapt_vars (fill_holes C ts))"
    using x(2,3) by (auto simp: adapt_vars_def map_vars_term_fill_holes)
  then show ?case using x(1,5,6) by (auto 0 0 simp: image_def prod.map_comp gtt_accept_equiv
      intro!: bexI[of _ "map_prod adapt_vars adapt_vars x"])
next
  case (rl x)
  obtain C ss ts where x: "x = (adapt_vars (fill_holes C ss), adapt_vars (fill_holes C ts))"
    "length ss = num_holes C" "length ts = num_holes C"
    "\<And>i. i < num_holes C \<Longrightarrow> \<exists>q. q \<in> ta_res (fst \<G>) (ss ! i) \<and> q \<in> ta_res (snd \<G>) (ts ! i)"
    "ground (fill_holes C ss)" "ground (fill_holes C ts)"
    using rl by (auto 0 0 simp: gtt_accept_equiv image_def elim!: gtt_accept'.cases)
  from x(4) obtain qs where qs: "qs i \<in> ta_res (fst \<G>) (ss ! i) \<and> qs i \<in> ta_res (snd \<G>) (ts ! i)"
    if "i < num_holes C" for i by metis
  have "gtt_accept' (fmap_states_gtt f \<G>) (fst x) (snd x)"
    using gtt_accept'.intros[of "map (map_vars_term f) ss" "map (map_vars_term f) ts"
      "map_vars_mctxt f C" "fmap_states_gtt f \<G>"] x(2,3,5,6)
      ta_res_fmap_states_ta[OF conjunct1[OF qs], where h = f]
      ta_res_fmap_states_ta[OF conjunct2[OF qs], where h = f]
    by (auto simp: map_vars_term_fill_holes x(1) adapt_vars_def ground_fill_holes
      ground_map_vars_term_simp[of _ _ f] ground_map_vars_mctxt_simp[of _ f]
      cong: map_cong)
  then show ?case using x(1,2,3,5,6) by (auto simp: gtt_accept_equiv)
qed

definition gtt_only_reach where
  "gtt_only_reach = map_both ta_only_reach"

lemma gtt_only_reach_lang:
  "gtt_lang (gtt_only_reach G) = gtt_lang G"
proof (intro equalityI subrelI, goal_cases LR RL)
  case (LR s t)
  then obtain C ss ts qs where
    s: "s = fill_holes C ss" "length ss = num_holes C" "ground s" and
    t: "t = fill_holes C ts" "length ts = num_holes C" "ground t" and
    q: "\<And>i. i < num_holes C \<Longrightarrow> \<exists>q. q \<in> ta_res (fst (gtt_only_reach G)) (ss ! i) \<and>
                        q \<in> ta_res (snd (gtt_only_reach G)) (ts ! i)"
    by (auto 0 0 simp: gtt_accept_equiv elim!: gtt_accept'.cases)
  have "\<exists>q. q \<in> ta_res (fst G) (ss ! i) \<and> q \<in> ta_res (snd G) (ts ! i)" if "i < num_holes C" for i
    using q[of i] subsetD[OF ta_res_mono[OF ta_restrict_subset]] that by (fastforce simp: gtt_only_reach_def)
  then show ?case using s t by (auto simp: gtt_accept_equiv)
next
  case (RL s t)
  then obtain C ss ts qs where
    s: "s = fill_holes C ss" "length ss = num_holes C" "ground s" and
    t: "t = fill_holes C ts" "length ts = num_holes C" "ground t" and
    q: "\<And>i. i < num_holes C \<Longrightarrow> \<exists>q. q \<in> ta_res (fst G) (ss ! i) \<and> q \<in> ta_res (snd G) (ts ! i)"
    by (auto 0 0 simp: gtt_accept_equiv elim!: gtt_accept'.cases)
  have "\<exists>q. q \<in> ta_res (fst (gtt_only_reach G)) (ss ! i) \<and> q \<in> ta_res (snd (gtt_only_reach G)) (ts ! i)"
    if "i < num_holes C" for i
    using q[of i] ta_res_only_reach[OF refl, of "ss ! i" "fst G", THEN conjunct2]
      ta_res_only_reach[OF refl, of "ts ! i" "snd G", THEN conjunct2] that s t
    by (simp add: ground_fill_holes all_set_conv_all_nth)
      (auto simp: ground_vars_term_empty set_conv_nth gtt_only_reach_def)
  then show ?case using s t by (auto simp: gtt_accept_equiv)
qed

lemma gtt_only_reach_syms:
  "gtt_syms (gtt_only_reach G) \<subseteq> gtt_syms G"
  by (auto simp: gtt_only_reach_def ta_restrict_def)
    (metis (mono_tags, lifting) image_iff mem_Collect_eq select_convs(2) ta_syms_def)+

definition gtt_only_prod where
  "gtt_only_prod G =
    (let iface = ta_states ((fst G)\<lparr>ta_final := {}\<rparr>) \<inter> ta_states ((snd G)\<lparr>ta_final := {}\<rparr>) in
    map_both ta_only_prod ((fst G)\<lparr>ta_final := iface\<rparr>, (snd G)\<lparr>ta_final := iface\<rparr>))"

lemma gtt_only_prod_lang:
  "gtt_lang (gtt_only_prod G) = gtt_lang G"
proof (intro equalityI subrelI, goal_cases LR RL)
  case (LR s t)
  then obtain C ss ts qs where
    s: "s = fill_holes C ss" "length ss = num_holes C" "ground s" and
    t: "t = fill_holes C ts" "length ts = num_holes C" "ground t" and
    q: "\<And>i. i < num_holes C \<Longrightarrow> \<exists>q. q \<in> ta_res (fst (gtt_only_prod G)) (ss ! i) \<and>
                        q \<in> ta_res (snd (gtt_only_prod G)) (ts ! i)"
    by (auto 0 0 simp: gtt_accept_equiv elim!: gtt_accept'.cases)
  have "\<exists>q. q \<in> ta_res (fst G) (ss ! i) \<and> q \<in> ta_res (snd G) (ts ! i)" if "i < num_holes C" for i
    using q[of i] subsetD[OF ta_res_mono'[of "fst (gtt_only_prod G)" "fst G"]]
      subsetD[OF ta_res_mono'[of "snd (gtt_only_prod G)" "snd G"]] that
      ta_restrict_subset[of "_\<lparr> ta_final := _ \<rparr>"]
    by (fastforce simp: gtt_only_prod_def Let_def ta_subset_def)
  then show ?case using s t by (auto simp: gtt_accept_equiv)
next
  case (RL s t)
  then obtain C ss ts qs where
    s: "s = fill_holes C ss" "length ss = num_holes C" "ground s" and
    t: "t = fill_holes C ts" "length ts = num_holes C" "ground t" and
    q: "\<And>i. i < num_holes C \<Longrightarrow> \<exists>q. q \<in> ta_res (fst G) (ss ! i) \<and> q \<in> ta_res (snd G) (ts ! i)"
    by (auto 0 0 simp: gtt_accept_equiv elim!: gtt_accept'.cases)
  let ?iface = "ta_states (fst G\<lparr>ta_final := {}\<rparr>) \<inter> ta_states (snd G\<lparr>ta_final := {}\<rparr>)"
  { fix i assume i: "i < num_holes C"
    guess q using q[OF i] by (elim exE conjE) note q = this
    have g: "ground (ss ! i)" "ground (ts ! i)" using s t i by (auto simp: ground_fill_holes)
    have "q \<in> ?iface" using q g
      subset_trans[OF ta_res_mono'[of A "A\<lparr>ta_final := {}\<rparr>" for A] ta_res_states[of _ "_\<lparr>ta_final := {} \<rparr>"], THEN subsetD]
      by fastforce
    have "q \<in> ta_res (fst (gtt_only_prod G)) (ss ! i)" "q \<in> ta_res (snd (gtt_only_prod G)) (ts ! i)"
      using q g by (auto simp: gtt_only_prod_def Let_def intro!: ta_productive_final
        ta_res_only_prod subsetD[OF ta_res_mono'[of A "A\<lparr>ta_final := ?iface\<rparr>" for A]]
        subset_trans[OF ta_res_mono'[of A "A\<lparr>ta_final := {}\<rparr>" for A] ta_res_states[of _ "_\<lparr>ta_final := {} \<rparr>"], THEN subsetD])
    then have "\<exists>q. q \<in> ta_res (fst (gtt_only_prod G)) (ss ! i) \<and> q \<in> ta_res (snd (gtt_only_prod G)) (ts ! i)"
      by auto
  }
  then show ?case using s t by (auto simp: gtt_accept_equiv)
qed

lemma gtt_only_prod_syms:
  "gtt_syms (gtt_only_prod G) \<subseteq> gtt_syms G"
  by (auto simp: gtt_only_prod_def ta_restrict_def)
    (smt fst_conv snd_conv image_iff mem_Collect_eq select_convs(2) ta_syms_def)+

definition trim_gtt where
  "trim_gtt = gtt_only_prod \<circ> gtt_only_reach"

lemma trim_gtt_lang:
  "gtt_lang (trim_gtt G) = gtt_lang G"
  unfolding trim_gtt_def comp_def gtt_only_prod_lang gtt_only_reach_lang ..

lemma trim_gtt_prod_syms:
  "gtt_syms (trim_gtt G) \<subseteq> gtt_syms G"
  unfolding trim_gtt_def using subset_trans[OF gtt_only_prod_syms gtt_only_reach_syms] by simp

end
