theory GTT_Compose_Impl
  imports GTT_Compose GTT_Impl Horn_Autoref TA.Tree_Automata_Autoref_Setup
begin

definition \<Delta>\<^sub>\<epsilon>_rules :: "('q, 'f) ta \<Rightarrow> ('q, 'f) ta \<Rightarrow> ('q \<times> 'q) horn set" where
  "\<Delta>\<^sub>\<epsilon>_rules A B =
    {zip ps qs \<rightarrow>\<^sub>h (p, q) |f ps p qs q. f ps \<rightarrow> p \<in> ta_rules A \<and> f qs \<rightarrow> q \<in> ta_rules B \<and> length ps = length qs} \<union>
    {[(p, q)] \<rightarrow>\<^sub>h (p', q) |p p' q. (p, p') \<in> ta_eps A} \<union>
    {[(p, q)] \<rightarrow>\<^sub>h (p, q') |p q q'. (q, q') \<in> ta_eps B}"

locale \<Delta>\<^sub>\<epsilon>_horn =
  fixes A :: "('q, 'f) ta" and B :: "('q, 'f) ta"
begin

sublocale horn "\<Delta>\<^sub>\<epsilon>_rules A B" .

lemma \<Delta>\<^sub>\<epsilon>_infer0:
  "infer0 = {(p, q) |f p q. f [] \<rightarrow> p \<in> ta_rules A \<and> f [] \<rightarrow> q \<in> ta_rules B}"
  unfolding horn.infer0_def \<Delta>\<^sub>\<epsilon>_rules_def
  using zip_Nil[of "[]"] zip_inj[of "[]" "[]"] by auto force+

lemma \<Delta>\<^sub>\<epsilon>_infer1:
  "infer1 pq X = {(p, q) |f ps p qs q. f ps \<rightarrow> p \<in> ta_rules A \<and> f qs \<rightarrow> q \<in> ta_rules B \<and> length ps = length qs \<and>
    (fst pq, snd pq) \<in> set (zip ps qs) \<and> set (zip ps qs) \<subseteq> insert pq X} \<union>
    {(p', snd pq) |p p'. (p, p') \<in> ta_eps A \<and> p = fst pq} \<union>
    {(fst pq, q') |q q'. (q, q') \<in> ta_eps B \<and> q = snd pq}"
  unfolding \<Delta>\<^sub>\<epsilon>_rules_def horn_infer1_union
  apply (intro arg_cong2[of _ _ _ _ "(\<union>)"])
    by (auto simp: horn.infer1_def simp flip: ex_simps(1)) force+

lemma \<Delta>\<^sub>\<epsilon>_sound:
  "\<Delta>\<^sub>\<epsilon> A B = saturate"
proof (intro set_eqI iffI, goal_cases lr rl)
  case (lr x) obtain p q where x: "x = (p, q)" by (cases x)
  show ?case using lr unfolding x
  proof (induct)
    case (\<Delta>\<^sub>\<epsilon>_cong f ps p qs q) show ?case
      apply (intro infer[of "zip ps qs" "(p, q)"])
      subgoal using \<Delta>\<^sub>\<epsilon>_cong(1-3) by (auto simp: \<Delta>\<^sub>\<epsilon>_rules_def)
      subgoal using \<Delta>\<^sub>\<epsilon>_cong(3,5) by (auto simp: zip_nth_conv)
      done
  next
    case (\<Delta>\<^sub>\<epsilon>_eps1 p p' q) then show ?case
      by (intro infer[of "[(p, q)]" "(p', q)"]) (auto simp: \<Delta>\<^sub>\<epsilon>_rules_def)
  next
    case (\<Delta>\<^sub>\<epsilon>_eps2 q q' p) then show ?case
      by (intro infer[of "[(p, q)]" "(p, q')"]) (auto simp: \<Delta>\<^sub>\<epsilon>_rules_def)
  qed
next
  case (rl x) obtain p q where x: "x = (p, q)" by (cases x)
  show ?case using rl unfolding x
  proof (induct)
    case (infer as a) then show ?case
      using \<Delta>\<^sub>\<epsilon>_cong[of _ "map fst as" "fst a" A "map snd as" "snd a" B]
        \<Delta>\<^sub>\<epsilon>_eps1[of _ "fst a" A "snd a" B] \<Delta>\<^sub>\<epsilon>_eps2[of _ "snd a" B "fst a" A]
      by (auto simp: \<Delta>\<^sub>\<epsilon>_rules_def)
  qed
qed

end

definition \<Delta>\<^sub>\<epsilon>_infer0_list where
  "\<Delta>\<^sub>\<epsilon>_infer0_list A B = map (map_prod r_rhs r_rhs) (
     filter (\<lambda>(ra, rb). case (ra, rb) of (f ps \<rightarrow> p, g qs \<rightarrow> q) \<Rightarrow> f = g \<and> ps = [] \<and> qs = []) (List.product (tal_rules A) (tal_rules B)))"

definition \<Delta>\<^sub>\<epsilon>_infer1_list :: "('q, 'f) ta_list \<Rightarrow> ('q, 'f) ta_list \<Rightarrow> 'q \<times> 'q \<Rightarrow> ('q \<times> 'q) list \<Rightarrow> ('q \<times> 'q) list" where
  "\<Delta>\<^sub>\<epsilon>_infer1_list A B pq bs =
    map (map_prod r_rhs r_rhs) (filter (\<lambda>(ra, rb). case (ra, rb) of (f ps \<rightarrow> p, g qs \<rightarrow> q) \<Rightarrow> f = g \<and> length ps = length qs \<and> (fst pq, snd pq) \<in> set (zip ps qs) \<and> set (zip ps qs) \<subseteq> set ((fst pq, snd pq) # bs)) (List.product (tal_rules A) (tal_rules B))) @
    map (\<lambda>(p, p'). (p', snd pq)) (filter (\<lambda>(p, p') \<Rightarrow> p = fst pq) (tal_eps A)) @
    map (\<lambda>(q, q'). (fst pq, q')) (filter (\<lambda>(q, q') \<Rightarrow> q = snd pq) (tal_eps B))"

locale \<Delta>\<^sub>\<epsilon>_list =
  fixes A :: "('q, 'f) ta_list" and B :: "('q, 'f) ta_list"
begin

sublocale \<Delta>\<^sub>\<epsilon>_horn "ta_of A" "ta_of B" .

sublocale l: horn_list "\<Delta>\<^sub>\<epsilon>_rules (ta_of A) (ta_of B)" "\<Delta>\<^sub>\<epsilon>_infer0_list A B" "\<Delta>\<^sub>\<epsilon>_infer1_list A B"
  apply (unfold_locales)
  unfolding \<Delta>\<^sub>\<epsilon>_horn.\<Delta>\<^sub>\<epsilon>_infer0 \<Delta>\<^sub>\<epsilon>_horn.\<Delta>\<^sub>\<epsilon>_infer1 \<Delta>\<^sub>\<epsilon>_infer0_list_def \<Delta>\<^sub>\<epsilon>_infer1_list_def set_append Un_assoc[symmetric]
   apply (force simp: r_rhs_def map_prod_def split: ta_rule.splits)
  apply (intro arg_cong2[of _ _ _ _ "(\<union>)"])
  by (auto 0 0 simp: r_rhs_def map_prod_def) (force split: ta_rule.splits)+

lemmas infer = l.infer0 l.infer1
lemmas saturate_impl_sound = l.saturate_impl_sound
lemmas saturate_impl_complete = l.saturate_impl_complete

end

definition \<Delta>\<^sub>\<epsilon>_impl :: "('q, 'f) ta_list \<Rightarrow> ('q, 'f) ta_list \<Rightarrow> ('q \<times> 'q) list option" where
  "\<Delta>\<^sub>\<epsilon>_impl A B = horn_list_impl.saturate_impl (\<Delta>\<^sub>\<epsilon>_infer0_list A B) (\<Delta>\<^sub>\<epsilon>_infer1_list A B)"

lemma \<Delta>\<^sub>\<epsilon>_impl_sound:
  "\<Delta>\<^sub>\<epsilon>_impl A B = Some xs \<Longrightarrow> set xs = \<Delta>\<^sub>\<epsilon> (ta_of A) (ta_of B)"
  using \<Delta>\<^sub>\<epsilon>_list.saturate_impl_sound unfolding \<Delta>\<^sub>\<epsilon>_impl_def \<Delta>\<^sub>\<epsilon>_horn.\<Delta>\<^sub>\<epsilon>_sound .

lemma \<Delta>\<^sub>\<epsilon>_impl_complete:
  fixes A B :: "('q, 'f) ta_list"
  shows "\<Delta>\<^sub>\<epsilon>_impl A B \<noteq> None"
proof -
  have *: "\<Delta>\<^sub>\<epsilon> (ta_of A) (ta_of B) \<subseteq> ta_states (ta_of A) \<times> ta_states (ta_of B)"
    by (auto simp: \<Delta>\<^sub>\<epsilon>_def' ta_of_def subsetD[OF ta_res_states])
  have "finite (\<Delta>\<^sub>\<epsilon> (ta_of A) (ta_of B))"
    by (cases A; cases B; rule finite_subset[OF *])
      (auto simp: ta_of_def ta_states_def r_states_def intro!: finite_cartesian_product)
  then show ?thesis unfolding \<Delta>\<^sub>\<epsilon>_impl_def
    by (intro \<Delta>\<^sub>\<epsilon>_list.saturate_impl_complete) (auto simp: \<Delta>\<^sub>\<epsilon>_horn.\<Delta>\<^sub>\<epsilon>_sound)
qed

definition GTT_comp_impl :: "('q, 'f) gtt_list \<Rightarrow> ('q, 'f) gtt_list \<Rightarrow> (_, 'f) gtt_list option" where
  "GTT_comp_impl G1 G2 = (case (\<Delta>\<^sub>\<epsilon>_impl (snd G1) (fst G2), \<Delta>\<^sub>\<epsilon>_impl (fst G2) (snd G1)) of
     (Some xs, Some ys) \<Rightarrow> Some
     (ta_list [] (tal_rules (fst G1) @ tal_rules (fst G2)) (tal_eps (fst G1) @ tal_eps (fst G2) @ xs),
      ta_list [] (tal_rules (snd G1) @ tal_rules (snd G2)) (tal_eps (snd G1) @ tal_eps (snd G2) @ ys))
     | _ \<Rightarrow> None
     )"

lemma GTT_comp_impl_sound:
  "GTT_comp_impl G1 G2 = Some G \<Longrightarrow> gtt_of G = GTT_comp (gtt_of G1) (gtt_of G2)"
  by (auto simp: GTT_comp_def GTT_comp_impl_def split: option.splits dest!: \<Delta>\<^sub>\<epsilon>_impl_sound
   intro!: ta_of_eq_ta_makeI)

lemma GTT_comp_impl_complete:
  "GTT_comp_impl G1 G2 \<noteq> None"
  using \<Delta>\<^sub>\<epsilon>_impl_complete[of "snd G1" "fst G2"] \<Delta>\<^sub>\<epsilon>_impl_complete[of "fst G2" "snd G1"]
  by (auto simp: GTT_comp_impl_def)

text \<open>Same thing with automatic refinement.\<close>

definition \<Delta>\<^sub>\<epsilon>_infer0_rbt where
  "\<Delta>\<^sub>\<epsilon>_infer0_rbt A B = map (map_prod r_rhs r_rhs) (
     filter (\<lambda>(ra, rb). case (ra, rb) of (f ps \<rightarrow> p, g qs \<rightarrow> q) \<Rightarrow> f = g \<and> ps = [] \<and> qs = [])
        (List.product (RBT_Impl.keys (ta_rules_impl A)) (RBT_Impl.keys (ta_rules_impl B))))"

definition \<Delta>\<^sub>\<epsilon>_infer1_rbt where
  "\<Delta>\<^sub>\<epsilon>_infer1_rbt A B pq X =
    map (map_prod r_rhs r_rhs) (filter (\<lambda>(ra, rb). case (ra, rb) of (f ps \<rightarrow> p, g qs \<rightarrow> q) \<Rightarrow> f = g \<and> length ps = length qs \<and>
     (fst pq, snd pq) \<in> set (zip ps qs) \<and> set (zip ps qs) \<subseteq> set (pq # (RBT_Impl.keys X))) (List.product (RBT_Impl.keys (ta_rules_impl A)) (RBT_Impl.keys (ta_rules_impl B)))) @
    map (\<lambda>(p, p'). (p', snd pq)) (filter (\<lambda>(p, p') \<Rightarrow> p = fst pq) (ta_eps_impl A)) @
    map (\<lambda>(q, q'). (fst pq, q')) (filter (\<lambda>(q, q') \<Rightarrow> q = snd pq) (ta_eps_impl B))"

lemma keys_list:
  fixes X :: "('a :: compare_order) set"
  assumes "(X', X) \<in> \<langle>R\<rangle>dflt_rs_rel"
  assumes "PREFER_id R"
  shows "(RBT_Impl.keys X', X) \<in> \<langle>R\<rangle>list_set_rel"
  using assms
    linorder.rbt_lookup_keys[OF linorder_class.linorder_axioms ord.is_rbt_rbt_sorted, of "X'"]
  by (auto simp: map2set_rel_def rbt_map_rel_def rbt_map_rel'_def list_set_rel_def br_def
    lt_of_comp_post_simp linorder_class.distinct_keys compare_order_class.ord_defs)

lemmas keys_it = list_set_rel_imp_it_set_rel[OF keys_list]

lemma ta_rules_list:
  assumes "(A', A) \<in> dflt_ta_rel"
  shows "(RBT_Impl.keys (ta_rules_impl A'), ta_rules A) \<in> \<langle>Id\<rangle>list_set_rel"
  using fun_relD[OF ta_impl_autoref(2), OF assms]
   keys_list[of "ta_rules_impl A'" "ta_rules A"]
  by (auto simp: lt_of_comp_post_simp compare_order_class.ord_defs)

lemmas ta_rules_it = list_set_rel_imp_it_set_rel[OF ta_rules_list]

lemma \<Delta>\<^sub>\<epsilon>_infer0_rbt:
  assumes A': "(A', A) \<in> dflt_ta_rel"
  assumes B': "(B', B) \<in> dflt_ta_rel"
  shows "(\<Delta>\<^sub>\<epsilon>_infer0_rbt A' B', horn.infer0 (\<Delta>\<^sub>\<epsilon>_rules A B)) \<in> \<langle>Id\<rangle>it_set_rel"
  unfolding \<Delta>\<^sub>\<epsilon>_horn.\<Delta>\<^sub>\<epsilon>_infer0 \<Delta>\<^sub>\<epsilon>_infer0_rbt_def
  using assms[THEN ta_rules_it]
  by (auto simp: it_set_rel_def br_def r_rhs_def split: ta_rule.split) force+

lemma \<Delta>\<^sub>\<epsilon>_infer1_rbt:
  assumes A': "(A', A) \<in> dflt_ta_rel"
  assumes B': "(B', B) \<in> dflt_ta_rel"
  assumes pq': "(pq', pq) \<in> Id"
  assumes X': "(X', X) \<in> \<langle>Id\<rangle>dflt_rs_rel"
  shows "(\<Delta>\<^sub>\<epsilon>_infer1_rbt A' B' pq' X', horn.infer1 (\<Delta>\<^sub>\<epsilon>_rules A B) pq X) \<in> \<langle>Id\<rangle>it_set_rel"
  unfolding \<Delta>\<^sub>\<epsilon>_horn.\<Delta>\<^sub>\<epsilon>_infer1 \<Delta>\<^sub>\<epsilon>_infer1_rbt_def it_set_rel_def list_rel_id_simp Id_O_R
  apply (intro brI TrueI)
  using assms(1,2)[THEN fun_relD[OF ta_impl_autoref(3)], THEN list_set_rel_imp_it_set_rel] assms(3)
    assms(1,2)[THEN ta_rules_it] assms(4)[THEN keys_it] unfolding set_append Un_assoc[symmetric]
  apply (intro arg_cong2[of _ _ _ _ "(\<union>)"])
  by (auto 0 0 intro!: rev_image_eqI simp: it_set_rel_def br_def r_rhs_def simp flip: ex_simps(1) split: ta_rule.split prod.split) force+

schematic_goal \<Delta>\<^sub>\<epsilon>_impl_rbt:
  assumes "(A', A) \<in> dflt_ta_rel"
  assumes "(B', B) \<in> dflt_ta_rel"
  shows "(?f, horn.saturate_impl (\<Delta>\<^sub>\<epsilon>_rules A B)) \<in> \<langle>\<langle>Id\<rangle>dflt_rs_rel\<rangle>nres_rel"
  apply (rule horn_linorder.saturate_impl_rbt)
   apply (rule \<Delta>\<^sub>\<epsilon>_infer0_rbt[OF assms]) 
  apply (intro fun_relI)
  apply (rule \<Delta>\<^sub>\<epsilon>_infer1_rbt[OF assms])
  apply assumption+
  done

concrete_definition \<Delta>\<^sub>\<epsilon>_rbt uses \<Delta>\<^sub>\<epsilon>_impl_rbt
prepare_code_thms \<Delta>\<^sub>\<epsilon>_rbt_def

abbreviation dflt_gtt_rel where
  "dflt_gtt_rel \<equiv> dflt_ta_rel \<times>\<^sub>r dflt_ta_rel"

schematic_goal GTT_comp_rbt_aux:
  assumes [autoref_rules]: "(G1', G1) \<in> dflt_gtt_rel"
  assumes [autoref_rules]: "(G2', G2) \<in> dflt_gtt_rel"
  defines "\<Delta>12 \<equiv> \<Delta>\<^sub>\<epsilon> (snd G1) (fst G2)" and "\<Delta>21 \<equiv> \<Delta>\<^sub>\<epsilon> (fst G2) (snd G1)"
  assumes \<Delta>: "(\<Delta>12', \<Delta>12) \<in> \<langle>Id \<times>\<^sub>r Id\<rangle>dflt_rs_rel" "(\<Delta>21', \<Delta>21) \<in> \<langle>Id \<times>\<^sub>r Id\<rangle>dflt_rs_rel"
  notes [autoref_rules] = \<Delta>[THEN keys_list]
  shows "(?f, GTT_comp G1 G2) \<in> dflt_gtt_rel"
  unfolding GTT_comp_def \<Delta>12_def[symmetric] \<Delta>21_def[symmetric]
  by autoref

concrete_definition GTT_comp_rbt_aux uses GTT_comp_rbt_aux

(* export_code GTT_comp_rbt_aux in Haskell *)

lemma bind_RETURN:
  assumes "(a, RETURN b) \<in> \<langle>R\<rangle>nres_rel"
  assumes "\<And>a'. (a', b) \<in> R \<Longrightarrow> (f' a', f) \<in> \<langle>Q\<rangle>nres_rel"
  shows "(a \<bind> f', f) \<in> \<langle>Q\<rangle>nres_rel"
  using assms 
  apply (auto simp: nres_rel_def bind_def SUP_le_iff split: nres.splits)
  by (metis RES_Sup_RETURN RETURN_ref_RETURND SUP_le_iff)

lemma nres_of_bind:
  "nres_of (f \<bind> g) = nres_of f \<bind> (\<lambda>x. nres_of (g x))"
  by (cases f) auto

schematic_goal GTT_comp_rbt:
  assumes G1: "(G1', G1) \<in> dflt_gtt_rel"
  assumes G2: "(G2', G2) \<in> dflt_gtt_rel"
  notes G1' = param_prod(4,5)[THEN fun_relD, OF G1]
  notes G2' = param_prod(4,5)[THEN fun_relD, OF G2]
  notes * = horn.saturate_impl_sound[unfolded SPEC_eq_is_RETURN, of "\<Delta>\<^sub>\<epsilon>_rules _ _", unfolded \<Delta>\<^sub>\<epsilon>_horn.\<Delta>\<^sub>\<epsilon>_sound[symmetric]]
  shows "(?f, RETURN (GTT_comp G1 G2)) \<in> \<langle>dflt_gtt_rel\<rangle>nres_rel"
  apply (rule HOL.back_subst[of "\<lambda>x. (x, _) \<in> _"])
  apply (rule bind_RETURN[OF conc_trans_additional(1)[OF \<Delta>\<^sub>\<epsilon>_rbt.refine[OF G1'(2) G2'(1), THEN nres_relD] *, THEN nres_relI]])
  apply (rule bind_RETURN[OF conc_trans_additional(1)[OF \<Delta>\<^sub>\<epsilon>_rbt.refine[OF G2'(1) G1'(2), THEN nres_relD] *, THEN nres_relI]])
  apply (rule param_RETURN[THEN fun_relD])
   apply (rule GTT_comp_rbt_aux[OF G1 G2, unfolded prod_rel_id_simp]; assumption+)
  unfolding nres_of_simps(3)[symmetric] nres_of_bind[symmetric]
  by (rule refl)

concrete_definition GTT_comp_rbt uses GTT_comp_rbt

export_code GTT_comp_rbt in Haskell

end