theory GTT_RRn_Impl
  imports GTT_RRn GTT_Impl TA.Tree_Automata_Autoref_Setup
begin

context begin interpretation autoref_syn .
lemma op_map_lookup_pat: "(m :: 'k \<rightharpoonup> 'v) k \<equiv> op_map_lookup $ k $ m" by simp
end

lemma image_impl:
  fixes f :: "'a :: compare \<Rightarrow> 'b :: compare"
  shows "(RBT_Impl.ord.rbt_bulkload (comp2lt compare_res) \<circ> map (\<lambda>x. (f x, ())) \<circ> RBT_Impl.keys, \<lambda>X. f ` X)
    \<in> \<langle>Id\<rangle>comp_rs_rel \<rightarrow> \<langle>Id\<rangle>comp_rs_rel"
proof -
  have bar: "set_to_map ((\<lambda>x. (x, f x)) ` S) = (\<lambda>x. if x \<in> S then Some (f x) else None)" for f S
    by (fastforce simp: set_to_map_def Eps_Opt_def)
  have foo: "map_of (map (\<lambda>x. (x, f x)) xs) = set_to_map ((\<lambda>x. (x, f x)) ` set xs)" for f xs
    unfolding bar by (induct xs) auto
  interpret linorder "comp2le compare_res" "comp2lt compare_res"
    using compare_to_eq_linorder eq_linorder_class_conv by blast
  show ?thesis using foo[of "\<lambda>_. ()" "map f _"]
    by (auto simp: map2set_rel_def rbt_map_rel_def rbt_map_rel'_def in_br_conv set_to_map_dom image_def
      eq_linorder_class_conv[symmetric] compare_to_eq_linorder inj_on_def comp_def
      rbt_lookup_rbt_bulkload rbt_lookup_keys[symmetric] distinct_map
      intro!: linorder.is_rbt_rbtreeify[of "comp2le compare_res"] arg_cong[of _ _ set_to_map])
qed

lemma rbt_lookup_into_class: "ord.rbt_lookup (<) = RBT_Impl.rbt_lookup"
  by (simp add: ord.rbt_lookup_def RBT_Impl.rbt_lookup_def)

lemma list_to_rbt:
  (* PREFER_id ? *)
  "(RBT_Impl.rbt_bulkload \<circ> map (\<lambda>x :: _ :: compare_order. (x, ())), id) \<in> \<langle>Id\<rangle>list_set_rel \<rightarrow> \<langle>Id\<rangle>comp_rs_rel"
proof -
  { fix xs X assume "(xs, X) \<in> \<langle>Id\<rangle>list_set_rel"
    then have "map_of (map (\<lambda>x. (x, ())) xs) x = (if x \<in> X then Some () else None)" for x
      by (auto simp: list_set_rel_def in_br_conv comp_def map_of_eq_None_iff)
  } note [simp] = ext[OF this]
  show ?thesis by (auto simp: map2set_rel_def rbt_map_rel_def rbt_map_rel'_def in_br_conv
    rbt_lookup_into_class lt_of_comp_post_simp ord_defs rbt_lookup_rbt_bulkload
    intro!: relcompI[of _ "\<lambda>x. if x \<in> _ then Some () else None"] split: if_splits)
qed

lemma ta_eps_ref':
  "((\<lambda>(_,_,E,_,_,_,_,_). RBT_Impl.rbt_bulkload (map (\<lambda>k. (k, ())) E)), ta_eps) \<in> dflt_ta_rel \<rightarrow> \<langle>Id \<times>\<^sub>r Id\<rangle>comp_rs_rel"
  using list_to_rbt by (auto simp: dflt_ta_rel_def dest: fun_relD)

schematic_goal [autoref_rules]:
  notes ta_eps_ref'[autoref_rules]
  shows "(?f, ta_states) \<in> dflt_ta_rel \<rightarrow> \<langle>Id\<rangle>comp_rs_rel"
  unfolding ta_states_def r_states_def
  by (autoref (trace, keep_goal) phases: id_op rel_inf fix_rel trans)

lemma [autoref_rules]:
  "(map_ta_rule, map_ta_rule) \<in> (R \<rightarrow> R') \<rightarrow> (S \<rightarrow> S') \<rightarrow> \<langle>S, R\<rangle>ta_rule_rel \<rightarrow> \<langle>S', R'\<rangle>ta_rule_rel"
  using param_map[of R R'] by (auto simp: fun_relD ta_rule_rel_def dest: fun_relD)

lemma set_comprehension_aux:
  "\<And>h X. {h f ps p |f ps p. f ps \<rightarrow> p \<in> X} = case_ta_rule h ` X"
  "\<And>h X Y. {h f ps p g qs q|f ps p g qs q. f ps \<rightarrow> p \<in> X \<and> g qs \<rightarrow> q \<in> Y} =
    (\<lambda>(r, r'). case_ta_rule (case_ta_rule h r) r') ` (X \<times> Y)"
  "\<And>h X. {h p p' |p p'. (p, p') \<in> X} = (\<lambda>(p, p'). h p p') ` X"
  "\<And>h X Y. {h p p' q |p q p'. (p, p') \<in> X \<and> q \<in> Y} = (\<lambda>((p, p'), q). h p p' q) ` (X \<times> Y)"
  "\<And>h X Y. {h p q q' |p q q'. p \<in> X \<and> (q, q') \<in> Y} = (\<lambda>(p, (q, q')). h p q q') ` (X \<times> Y)"
  "\<And>h X. {h q |q. q \<in> X} = h ` X"
  by (fastforce intro: rev_image_eqI split: ta_rule.splits)+

lemma param_zip_fill:
  "(zip_fill, zip_fill) \<in> \<langle>R\<rangle>list_rel \<rightarrow> \<langle>S\<rangle>list_rel \<rightarrow> \<langle>\<langle>R\<rangle>option_rel \<times>\<^sub>r \<langle>S\<rangle>option_rel\<rangle>list_rel"
  apply (refine_vcg)
  subgoal for xs xs' ys ys'
    by (induct xs arbitrary: ys xs' ys'; case_tac ys)
      (auto simp: zip_fill_code elim!: list_relE intro!: param_map[THEN fun_relD, THEN fun_relD])
  done

schematic_goal GTT_to_RR2_ref:
  (* notes op_map_lookup_pat[autoref_op_pat del] (* very helpful for debugging *) *)
  notes [autoref_rules] = param_replicate param_zip_fill
  shows "(?f, (\<lambda>A. (ta_final A, ta_rules A, ta_eps A)) \<circ>\<circ> GTT_to_RR2)
   \<in> \<langle>Id \<times>\<^sub>r Id\<rangle>comp_rs_rel \<rightarrow> dflt_ta_rel \<times>\<^sub>r dflt_ta_rel \<rightarrow>
   (\<langle>\<langle>\<langle>Id\<rangle>option_rel \<times>\<^sub>r \<langle>Id\<rangle>option_rel\<rangle>option_rel\<rangle>comp_rs_rel \<times>\<^sub>r
    \<langle>\<langle>\<langle>Id\<rangle>option_rel \<times>\<^sub>r \<langle>Id\<rangle>option_rel,\<langle>\<langle>Id\<rangle>option_rel \<times>\<^sub>r \<langle>Id\<rangle>option_rel\<rangle>option_rel\<rangle>ta_rule_rel\<rangle>comp_rs_rel \<times>\<^sub>r 
    \<langle>\<langle>\<langle>Id\<rangle>option_rel \<times>\<^sub>r \<langle>Id\<rangle>option_rel\<rangle>option_rel \<times>\<^sub>r \<langle>\<langle>Id\<rangle>option_rel \<times>\<^sub>r \<langle>Id\<rangle>option_rel\<rangle>option_rel\<rangle>list_set_rel)"
  unfolding GTT_to_RR2_def[abs_def] parallel_closure_automaton_def GTT_to_RR2_root_def pair_automaton_def
    ta_make_simps ta.simps ta.make_def comp_def map_prod_def
  unfolding set_comprehension_aux image_comp comp_def
  by autoref

concrete_definition GTT_to_RR2_impl uses GTT_to_RR2_ref

(*
term GTT_to_RR2_impl
thm ta_make_autoref
export_code GTT_to_RR2_impl in Haskell
*)

end