text \<open>GTT for accepting parallel steps of a (left-linear, right-ground) TRS\<close>

theory GTT_TRS
  imports GTT GTT_Transitive_Closure Tree_Automata_Utils
begin

(* In this setup we use option types, The option Some x, plays the same role as e_x
in the text of the paper *)

type_synonym ('f, 'q) term_state = "('f, 'q) term option"

definition cmn_ta_rules :: "('f, 'v) term set \<Rightarrow> (('f, 'v) term option, 'f) ta_rule set" where
  "cmn_ta_rules T \<equiv> {f (map Some ts) \<rightarrow> Some (Fun f ts) |f ts t. Fun f ts \<unlhd> t \<and> t \<in> T}"

definition TRS_fst :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term set" where
  "TRS_fst Ts \<equiv> {fst R |R. R \<in> Ts} "

definition TRS_snd :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term set" where
  "TRS_snd Ts \<equiv> {snd R |R. R \<in> Ts}"

definition TRS_terms :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term set" where
  "TRS_terms Ts \<equiv> TRS_fst Ts \<union> TRS_snd Ts"

definition trs_to_ta_A :: "('f, 'v) trs \<Rightarrow> (('f, 'v) term option, 'f) ta" where
  "trs_to_ta_A Ts \<equiv> ta.make {}
    (cmn_ta_rules (TRS_terms Ts))
    {(Some l, Some r) |l r. (l, r) \<in> Ts}"

definition trs_to_ta_B :: "('f, 'v) trs \<Rightarrow> (('f, 'v) term option, 'f) ta" where
  "trs_to_ta_B Ts \<equiv> ta.make {}
    (cmn_ta_rules (TRS_terms Ts))
    {(Some r, Some l) |l r. (l,r) \<in> Ts}"

abbreviation abs :: "('f, 'v) term \<Rightarrow> ('f, ('f, 'v) term option) term" where
  "abs x \<equiv> adapt_vars x"

abbreviation rep :: "('f, ('f, 'v) term option) term \<Rightarrow> ('f, 'v) term" where
  "rep x \<equiv> adapt_vars x"

abbreviation abs_R :: "(('f, 'v) term \<times> ('f, 'v) term) set \<Rightarrow>
  (('f, ('f, 'v) term option) term \<times> ('f, ('f, 'v) term option) term) set" where
  "abs_R R \<equiv> (\<lambda>t. (abs (fst t), abs (snd t))) ` R"

lemma symmetry_in_G_D:
  "ta_rules (trs_to_ta_B R) = ta_rules (trs_to_ta_A R)"
  "ta_eps (trs_to_ta_B R) = (\<lambda>t. (snd t, fst t)) ` ta_eps (trs_to_ta_A R)"
  unfolding trs_to_ta_B_def trs_to_ta_A_def by force+

(*Finiteness Theorems*)
lemma finite_embedding:
  assumes "finite R"
  shows "finite {(Some l, Some r) |l r. (l, r) \<in> R}"
proof -
  have "{(Some l, Some r) |l r. (l, r) \<in> R} = (\<lambda>(l, r). (Some l, Some r)) ` R"
    by auto
  then show ?thesis using assms by auto
qed

lemma finite_collect:
  assumes "finite S"
  shows "finite {t |t w. t \<unlhd> w \<and> w \<in> S}"
proof -
 have "\<forall>w \<in> S. finite {t. t \<unlhd> w}" using finite_subterms assms by auto
  hence "finite (\<Union> ((\<lambda>w. {t. t \<unlhd> w}) ` S))" using assms by blast
  moreover have "{t |t w. t \<unlhd> w \<and> w \<in> S} = \<Union> ((\<lambda>w. {t. t \<unlhd> w}) ` S)"
    by blast
  ultimately show ?thesis by argo
qed


lemma finite_ta_states:
  assumes "finite R"
  shows "finite (ta_states (trs_to_ta_A R))" "finite (ta_states (trs_to_ta_B R))"
proof -
  have eps: "finite (ta_eps (trs_to_ta_A R))" using assms finite_embedding[OF assms]
    unfolding trs_to_ta_A_def ta.make_def by auto
  moreover have rules: "finite (ta_rules (trs_to_ta_A R))"
  proof -
    have fin: "finite (TRS_terms R)" using assms unfolding TRS_terms_def TRS_fst_def TRS_snd_def
    proof -
      have "\<forall>f. finite {t. \<exists>p. (t :: ('a, 'b) Term.term) = f p \<and> p \<in> R}"
        by (metis (no_types) Collect_mem_eq assms finite_image_set)
      then show "finite ({fst p |p. p \<in> R} \<union> {snd p |p. p \<in> R})"
        by blast
    qed
    let ?fts = "{Fun f ts |f ts t. (\<exists>t \<unrhd> Fun f ts. t \<in> TRS_terms R)}"
    have fin_fun: "finite {Fun f ts |f ts t. \<exists>t \<unrhd> Fun f ts. t \<in> TRS_terms R}"
    proof -
      have "{uu. \<exists>f ts t. uu = Fun f ts \<and> (\<exists>t \<unrhd> Fun f ts. t \<in> TRS_terms R)}
        \<subseteq> {uu. \<exists>t w. uu = t \<and> w \<unrhd> t \<and> w \<in> TRS_terms R}" by blast
      thus ?thesis
        using finite_collect[OF fin]
          finite_subset[of A "{uu. \<exists>t w. uu = t \<and> w \<unrhd> t \<and> w \<in> TRS_terms R}" for A]
        by presburger
    qed
    then have "finite {f map Some ts \<rightarrow> Some (Fun f ts) |f ts. \<exists>t \<unrhd> Fun f ts. t \<in> TRS_terms R}"
    proof -
      have "{f (map Some ts) \<rightarrow> Some (Fun f ts) |f ts. \<exists>t \<unrhd> Fun f ts. t \<in> TRS_terms R} =
          (\<lambda>u. case u of Fun f ts \<Rightarrow> f (map Some ts) \<rightarrow> Some (Fun f ts)) `
          {Fun f ts |f ts t . (\<exists>t \<unrhd> Fun f ts. t \<in> TRS_terms R)}"
        by (smt Collect_cong Term.term.simps(6) setcompr_eq_image)
      then show ?thesis using fin_fun by simp
    qed
    then show ?thesis unfolding trs_to_ta_A_def ta.make_def cmn_ta_rules_def by auto
  qed
  ultimately show "finite (ta_states (trs_to_ta_A R))"
    unfolding trs_to_ta_A_def ta_states_def ta.make_def
    by (simp add: r_states_def)
  show "finite (ta_states (trs_to_ta_B R))"
    using finite_imageI[of "ta_eps (trs_to_ta_A R)" "\<lambda>t. (snd t, fst t)"] eps
      symmetry_in_G_D[of R] rules unfolding ta_states_def trs_to_ta_B_def ta.make_def
    by (simp add: r_states_def)
qed

lemma conv_adpvars_G:
  assumes "ground s" and "s \<unlhd> t \<and> t \<in> TRS_terms R"
  shows "Some s \<in> ta_res (trs_to_ta_A R) (adapt_vars s)"
  using assms
proof (induction s)
  case (Fun f ss)
  have 0: "\<forall>i < length ss. (map (ta_res (trs_to_ta_A R)) (map adapt_vars ss)) ! i =
      ta_res (trs_to_ta_A R) (adapt_vars (ss ! i))"
    by simp
  { fix i
    assume i: "i < length ss"
    hence "ground (ss ! i)" using Fun(2) by simp
    moreover have "ss ! i \<unlhd> t \<and> t \<in> TRS_terms R" using Fun(3)
      by (meson \<open>i < length ss\<close> nth_mem supteq.refl supteq.subt supteq_trans)
    moreover have "adapt_vars (ss ! i) = ss ! i"
      by (simp add: adapt_vars_id calculation(1))
    ultimately have "Some (ss ! i) \<in> ta_res (trs_to_ta_A R) (adapt_vars (ss ! i))"
      using Fun(1) i by fastforce
  } note ref = this
  have 1: "f (map Some ss) \<rightarrow> Some (Fun f ss) \<in> ta_rules (trs_to_ta_A R)"
    unfolding trs_to_ta_A_def cmn_ta_rules_def TRS_terms_def
    by (metis (mono_tags, lifting) CollectI Fun.prems(2) TRS_terms_def ta_make_simps(1))
  thus ?case
  proof -
    have "\<forall>i < length ss. (map Some ss) ! i \<in> map (ta_res (trs_to_ta_A R)) (map adapt_vars ss) ! i"
      using ref by simp
    then show ?thesis using 1 by force
  qed
qed (fastforce)



lemma conv_adpvars_D:
  assumes "ground s" and "s \<unlhd> t \<and> t \<in> TRS_terms R"
  shows "Some s \<in> ta_res (trs_to_ta_B R) (adapt_vars s)"
  using assms
proof (induction s)
  case (Fun f ss)
  have 0: "\<forall>i < length ss. (map (ta_res (trs_to_ta_B R)) (map adapt_vars ss)) ! i =
    ta_res (trs_to_ta_B R) (adapt_vars (ss ! i))"
    by simp
  { fix i
    assume i: "i < length ss"
    hence "ground (ss ! i)" using Fun(2) by simp
    moreover have "ss ! i \<unlhd> t \<and> t \<in> TRS_terms R" using Fun(3)
      by (meson \<open>i < length ss\<close> nth_mem supteq.refl supteq.subt supteq_trans)
    moreover have "adapt_vars (ss ! i) = ss ! i"
      by (simp add: adapt_vars_id calculation(1))
    ultimately have "Some (ss ! i) \<in> ta_res (trs_to_ta_B R) (adapt_vars (ss ! i))"
      using Fun(1) i by fastforce} note ref = this
  have 1: "f (map Some ss) \<rightarrow> Some (Fun f ss) \<in> ta_rules (trs_to_ta_B R)"
    unfolding trs_to_ta_A_def cmn_ta_rules_def TRS_terms_def
    by (metis (mono_tags, lifting) CollectI Fun.prems(2) cmn_ta_rules_def ta_make_simps(1) trs_to_ta_B_def)
  thus ?case
  proof -
    have "\<forall>i < length ss. (map Some ss) ! i \<in> map (ta_res (trs_to_ta_B R)) (map adapt_vars ss) ! i"
      using ref by simp
    then show ?thesis using 1 by force
  qed
qed (fastforce)

(*The following theorem gives a map from rstep acceptance to gtt_acceptance*)

lemma rstep_to_gtt_accept:
  assumes "ground_trs R" "ground u" "ground v" "(u, v) \<in> rstep R"
  shows "gtt_accept (trs_to_ta_A R, trs_to_ta_B R) (abs u) (abs v)"
  using assms(4,3,2,1)
proof (induction rule: rstepE)
  case (rstep s t C \<sigma> l r)
  let ?\<G> = "(trs_to_ta_A R, trs_to_ta_B R)"
  let ?C = "adapt_vars_ctxt C"
  let ?s = "adapt_vars l"
  let ?t = "adapt_vars r"
  have grnd_s_t: "ground l" "ground r" using rstep(3,8) unfolding ground_trs_def by fastforce+
  hence 1: "Some r \<in> ta_res (trs_to_ta_A R) (adapt_vars l)"
  proof -
    from rstep have "(Some l, Some r) \<in> ta_eps (trs_to_ta_A R)" unfolding trs_to_ta_A_def
      by (metis (mono_tags, lifting) mem_Collect_eq ta_make_simps(2))
    hence s_res: "Some r \<in> ta_res (trs_to_ta_A R) (Var (Some l))" by auto
    have "Some l \<in> ta_res (trs_to_ta_A R) (adapt_vars l)"
      using grnd_s_t(1) rstep(3) conv_adpvars_G[of l l R]
      unfolding TRS_terms_def TRS_fst_def by auto
    hence loc: "Var (Some l) \<in> ta_res' (trs_to_ta_A R) (adapt_vars l)"
      using ta_res_res'_inclusion[of "Some l" "trs_to_ta_A R" "adapt_vars l"] by blast
    thus ?thesis using ta_res'_ta_res[OF "loc" "s_res"] by argo
  qed
  hence 2: "Some r \<in> ta_res (trs_to_ta_B R) (adapt_vars r)"
  proof -
    have t_res: "Some r \<in> ta_res (trs_to_ta_B R) (Var (Some r))"
      unfolding trs_to_ta_B_def trs_to_ta_B_def using rule by simp
    have "Some r \<in> ta_res (trs_to_ta_B R) (adapt_vars r)"
      using grnd_s_t(2) rstep(3) conv_adpvars_D[of r r R]
      unfolding TRS_terms_def TRS_snd_def by auto
    hence loc: "Var (Some r) \<in> ta_res' (trs_to_ta_B R) (adapt_vars r)"
      using ta_res_res'_inclusion[of "Some r" "trs_to_ta_B R" "adapt_vars r"] by blast
    thus ?thesis using ta_res'_ta_res[OF "loc" "t_res"] by argo
  qed
  with 1 have IH: "gtt_accept (trs_to_ta_A R, trs_to_ta_B R) (adapt_vars l) (adapt_vars r)"
    by force
  moreover have u_v: "u = C\<langle>l\<rangle>" "v = C\<langle>r\<rangle>" using grnd_s_t ground_subst_apply rstep(1,2,4,5) by metis+
  ultimately have "gtt_accept' ?\<G> (adapt_vars u) (adapt_vars v)"
  proof -
    have loc: "?C\<langle>?s\<rangle> = adapt_vars u" "?C\<langle>?t\<rangle> = adapt_vars v"
      using u_v by simp_all
    moreover have "fill_holes (mctxt_of_ctxt ?C) [?s] = ?C\<langle>?s\<rangle>"
                  "fill_holes (mctxt_of_ctxt ?C) [?t] = ?C\<langle>?t\<rangle>"
      using fill_holes_mctxt_of_ctxt by blast+
    ultimately show ?thesis
      using accept'_closed_ctxt[of "[adapt_vars l]" "[adapt_vars r]" "?\<G>"
           "mctxt_of_ctxt (adapt_vars_ctxt C)"]
            IH
      by (simp add:loc gtt_accept_equiv)
  qed
  thus ?thesis using gtt_accept_equiv by blast
qed

lemma ground_rstep_domain:
  assumes "ground_trs R" "(a, b) \<in> (rstep R)" "ground b"
  shows "ground a"
  using assms(2,1,3)
proof (induction rule: rstepE)
  case (rstep s t C \<sigma> l r)
  have grnd: "ground l" "ground r" using rstep(3,6) unfolding ground_trs_def by fast+
  have "a = C\<langle>l\<rangle>" " b = C\<langle>r\<rangle>"
    using rstep(1,4) ground_subst_apply[OF grnd(1)] apply presburger
    using rstep(2,5) ground_subst_apply[OF grnd(2)] by presburger
  then show ?case using rstep(7) grnd
    using ground_ctxt_apply by blast
qed

lemma rstep_trancl_to_gtt_trancl:
  assumes "ground_trs R"
  shows "Restr (abs_R ((rstep R)\<^sup>*)) {t. ground t}
    \<subseteq> gtt_lang (GTT_trancl (trs_to_ta_A R, trs_to_ta_B R))"
proof -
  { fix s t
    assume asm: "(s, t) \<in> (Restr (abs_R ((rstep R)\<^sup>*)) {t. ground t})"
    hence grnd: "ground s" "ground t" by blast+
    define u v where u_v: "u = rep s" "v = rep t"
    moreover then have rep: "(u, v) \<in> (rstep R)\<^sup>*"
      by (smt IntE Pair_inject adapt_vars_reverse asm grnd(1) grnd(2) imageE prod.collapse)
    hence grnd': "ground u" "ground v" using grnd unfolding u_v by force+
    have "(abs u, abs v) \<in> (gtt_lang (trs_to_ta_A R, trs_to_ta_B R))\<^sup>*"
      using rep grnd' assms(1)
    proof (induction u v rule:rtrancl.induct)
      case (rtrancl_refl a)
      then show ?case by blast
    next
      case (rtrancl_into_rtrancl a b c)
      hence grnd: "ground b" using ground_rstep_domain[of R b c] by fast
      hence "(GTT_TRS.abs a, GTT_TRS.abs b) \<in> (gtt_lang (trs_to_ta_A R, trs_to_ta_B R))\<^sup>*"
        using rtrancl_into_rtrancl(3,4,6) by blast
      moreover have "(GTT_TRS.abs b, GTT_TRS.abs c)
                           \<in> (gtt_lang (trs_to_ta_A R, trs_to_ta_B R))"
       using grnd rtrancl_into_rtrancl(2,5,6) rstep_to_gtt_accept[of R b c]
         by auto
      ultimately show ?case
         by (meson rtrancl.rtrancl_into_rtrancl)
    qed
    hence "(s, t) \<in> (gtt_lang (trs_to_ta_A R, trs_to_ta_B R))\<^sup>*"
      using u_v
      by (simp add: u_v(1) u_v(2) grnd(1) grnd(2))
    moreover have "(s, t) \<in> {t. ground t} \<times> {t. ground t}"
      using grnd by blast
    ultimately have "(s, t) \<in> Restr ((gtt_lang (trs_to_ta_A R, trs_to_ta_B R))\<^sup>*) {t. ground t}"
      by blast
  }
  then have "Restr (abs_R ((rstep R)\<^sup>*)) {t. ground t}
              \<subseteq> Restr ((gtt_lang (trs_to_ta_A R, trs_to_ta_B R))\<^sup>*) {t. ground t}"
    by blast
 then show ?thesis
    using GTT_trancl_lang[of "(trs_to_ta_A R, trs_to_ta_B R)"]
    by (simp add: Restr_rtrancl_gtt_lang_eq_trancl_gtt_lang)
qed

(* The following set of theorem are useful in showing that gtt acceptance
implies an rstep relation. We prove a set of theorem for the ground tree tranducer G,
and replicate it for D*)

lemma Some_ta_res:
  assumes "q \<in> ta_res (trs_to_ta_A R) s" "ground_trs R" "ground s"
  shows "\<exists>p. q = Some p"
  using assms
proof (induction s arbitrary: q)
  case (Var x)
  then show ?case by simp
next
  case (Fun f ss)
  then obtain q' qs where q'_qs: "f qs \<rightarrow> q' \<in> ta_rules (trs_to_ta_A R)"
    "(q', q) \<in> (ta_eps (trs_to_ta_A R))\<^sup>*"
    "length qs = length ss"
    "\<forall>i < length ss. qs ! i \<in> map (ta_res (trs_to_ta_A R)) ss ! i"
    by auto
  then have "\<exists>p'. q' = Some p'" unfolding trs_to_ta_A_def
    by (smt CollectD cmn_ta_rules_def ta_make_simps(1) ta_rule.inject)
  then show ?case using q'_qs(2) unfolding trs_to_ta_A_def
    by (smt CollectD Pair_inject rtranclE ta_make_simps(2))
qed


lemma Some_ta_res_D:
  assumes "q \<in> ta_res (trs_to_ta_B R) s" "ground_trs R" "ground s"
  shows "\<exists>p. q = Some p"
  using assms
proof (induction s arbitrary: q)
  case (Var x)
  then show ?case by simp
next
  case (Fun f ss)
  then obtain q' qs where q'_qs: "f qs \<rightarrow> q' \<in> ta_rules (trs_to_ta_B R)"
    "(q', q) \<in> (ta_eps (trs_to_ta_B R))\<^sup>*"
    "length qs = length ss"
    "\<forall>i < length ss. qs ! i \<in> map (ta_res (trs_to_ta_B R)) ss ! i"
    by auto
  then have "\<exists>p'. q' = Some p'" unfolding trs_to_ta_B_def
    by (smt CollectD cmn_ta_rules_def ta_make_simps(1) ta_rule.inject)
  then show ?case using q'_qs(2) unfolding trs_to_ta_B_def
    by (smt CollectD Pair_inject rtranclE ta_make_simps(2))
qed

lemma ta_eps_tranclp_G:
  assumes "(a, b) \<in> (ta_eps (trs_to_ta_A R))\<^sup>+"
  shows "\<exists>l r. (a = Some l \<and> b = Some r) \<and> (l, r) \<in> R\<^sup>+" using assms
proof (induction rule: trancl_induct)
  case base
  then show ?case unfolding trs_to_ta_A_def ta.make_def
    by auto
next
  case (step c b)
  then obtain l r' where l: "a = Some l \<and> c = Some r'" "(l, r') \<in> R\<^sup>+" by blast
  then obtain r where "b = Some r \<and> (r', r) \<in> R " using step(2)
    unfolding trs_to_ta_A_def make_def by auto
  then show ?case using l by auto
qed

lemma ta_eps_rtraclp_G:
  assumes "(Some l, Some r) \<in> (ta_eps (trs_to_ta_A R))\<^sup>*"
  shows "(l, r) \<in> R\<^sup>*" using assms
proof (cases "Some l = Some r")
  case False
  hence "(Some l, Some r) \<in> (ta_eps (trs_to_ta_A R))\<^sup>+" using assms
    by (meson rtranclD)
  hence "(l, r) \<in> R\<^sup>+" using ta_eps_tranclp_G by blast
  then show ?thesis by fast
qed (blast)


lemma ta_res_G_rstep:
  assumes "ground_trs R" "ground s"
    "Some q \<in> ta_res (trs_to_ta_A R) s"
  shows "(adapt_vars s, q) \<in> (rstep R)\<^sup>*"
  using assms(3,2,1)
proof (induction s arbitrary: q)
  case (Var x)
  then show ?case by simp
next
  case (Fun f ss)
  then obtain qs q' where q'_qs: "f qs \<rightarrow> q' \<in> ta_rules (trs_to_ta_A R)"
    "(q', Some q) \<in> (ta_eps (trs_to_ta_A R))\<^sup>*"
    "length qs = length ss"
    "\<forall>i < length ss. qs ! i \<in> ta_res (trs_to_ta_A R) (ss ! i)" by auto
  have grnd: "\<forall>i < length ss. ground (ss ! i)" using Fun(3) by auto
  hence "\<forall>i < length qs. \<exists>x. Some x = qs ! i"
    using Some_ta_res q'_qs(3,4) assms(1)
    by metis
  then obtain ps where ps: "qs = map Some ps"
    using construct_exI'[of "length qs" "\<lambda>x i. Some x = qs ! i"]
    by (metis map_nth_eq_conv)
  { fix i
    assume i: "i < length ss"
    hence "(adapt_vars (ss ! i), ps ! i) \<in> (rstep R)\<^sup>*"
      using Fun(1) ps q'_qs(3,4) grnd assms(1) by auto }
  hence "(Fun f (map adapt_vars ss), Fun f ps) \<in> (rstep R)\<^sup>*"
    by (metis (mono_tags, lifting) args_rsteps_imp_rsteps length_map nth_map ps q'_qs(3))
  hence 1: "(adapt_vars (Fun f ss), Fun f ps) \<in> (rstep R)\<^sup>*"
    using Fun(3) by simp
  have "f (map Some ps) \<rightarrow> q' \<in> ta_rules (trs_to_ta_A R)"
    using q'_qs(1) ps by blast
  hence "q' = Some (Fun f ps)" unfolding trs_to_ta_A_def ta.make_def cmn_ta_rules_def
    by fastforce
  then have "(Some (Fun f ps), Some q) \<in> (ta_eps (trs_to_ta_A R))\<^sup>*" using q'_qs(2) by argo
  then have "(Fun f ps, q) \<in> R\<^sup>*" using ta_eps_rtraclp_G by blast
  with 1 show ?case
    by (smt rtranclD rtrancl_trancl_trancl subset_rstep trancl_into_rtrancl trancl_mono)
qed

(*Replicating above theorems for the case D*)

lemma ta_eps_tranclp_D:
  assumes "(a, b) \<in> (ta_eps (trs_to_ta_B R))\<^sup>+"
  shows "\<exists>l r. (a = Some r \<and> b = Some l) \<and> (l, r) \<in> R\<^sup>+ " using assms
proof (induction rule: trancl_induct)
  case base
  then show ?case unfolding trs_to_ta_B_def ta.make_def
    by auto
next
  case (step c b)
  then obtain r l' where l: "a = Some r \<and> c = Some l'" "(l', r) \<in> R\<^sup>+" by blast
  then obtain l where "b = Some l \<and> (l, l') \<in> R" using step(2)
    unfolding trs_to_ta_B_def make_def by auto
  then show ?case using l by auto
qed

lemma ta_eps_rtraclp_D:
  assumes "(Some l, Some r) \<in> (ta_eps (trs_to_ta_B R))\<^sup>*"
  shows "(r, l) \<in> R\<^sup>*" using assms
proof (cases "Some l = Some r")
  case False
  hence "(Some l, Some r) \<in> (ta_eps (trs_to_ta_B R))\<^sup>+" using assms
    by (meson rtranclD)
  hence "(r, l) \<in> R\<^sup>+" using ta_eps_tranclp_D by blast
  then show ?thesis by fast
qed (blast)


lemma ta_res_D_rstep:
  assumes "ground_trs R" "ground t"
    "Some q \<in> ta_res (trs_to_ta_B R) t"
  shows "(q, adapt_vars t) \<in> (rstep R)\<^sup>*"
  using assms(3,2,1)
proof (induction t arbitrary: q)
  case (Var x)
  then show ?case by simp
next
  case (Fun f ts)
  then obtain qs q' where q'_qs: "f qs \<rightarrow> q' \<in> ta_rules (trs_to_ta_B R)"
    "(q', Some q) \<in> (ta_eps (trs_to_ta_B R))\<^sup>*"
    "length qs = length ts"
    "\<forall>i < length ts. qs ! i \<in> ta_res (trs_to_ta_B R) (ts ! i)" by auto
  have grnd: "\<forall>i< length ts. ground (ts ! i)" using Fun(3) by auto
  hence "\<forall>i < length qs. \<exists>x. Some x = qs ! i"
    using Some_ta_res_D[of "qs ! i" R "ts ! i" for i] q'_qs(3,4) assms(1)
    by metis
  then obtain ps where ps: "qs = map Some ps"
    using construct_exI'[of "length qs" "\<lambda> x i. Some x = qs ! i"]
    by (metis map_nth_eq_conv)
  { fix i
    assume i: "i < length ts"
    hence "(ps ! i, adapt_vars (ts ! i)) \<in> (rstep R)\<^sup>*"
      using Fun(1) ps q'_qs(3,4) grnd assms(1) by auto }
  hence "(Fun f ps, Fun f (map adapt_vars ts)) \<in> (rstep R)\<^sup>*"
    by (metis (mono_tags, lifting) args_rsteps_imp_rsteps length_map nth_map ps q'_qs(3))
  hence 1: "(Fun f ps, adapt_vars (Fun f ts)) \<in> (rstep R)\<^sup>*"
    using Fun(3) by simp
  have "(f (map Some ps) \<rightarrow> q') \<in> ta_rules (trs_to_ta_B R)"
    using q'_qs(1) ps by blast
  hence "q' = Some (Fun f ps)" unfolding trs_to_ta_B_def ta.make_def cmn_ta_rules_def
    by fastforce
  then have "(Some (Fun f ps), Some q) \<in> (ta_eps (trs_to_ta_B R))\<^sup>*" using q'_qs(2) by argo
  then have "(q, Fun f ps) \<in> R\<^sup>*" using ta_eps_rtraclp_D by blast
  with 1 show ?case
    by (smt rtranclD rtrancl_trancl_trancl subset_rstep trancl_into_rtrancl trancl_mono)
qed

lemma gtt_accept_to_rstep_rtrancl:
  assumes "ground_trs R" "ground u" "ground v"
          "gtt_accept (trs_to_ta_A R, trs_to_ta_B R) u v"
  shows "(rep u, rep v) \<in> (rstep R)\<^sup>*"
  using assms(4,3,2,1)
 proof (induction rule:gtt_accept.induct)
  case (refl t)
  then show ?case by force
next
  case (join q s t)
  obtain p where p: "Some p = q"
    using Some_ta_res[of q R s] join(1,4,5) unfolding fst_conv by blast
  then have "(adapt_vars s, p) \<in> (rstep R)\<^sup>*"
    using ta_res_G_rstep[of R s p] join(1,4,5)
    unfolding fst_conv by blast
  moreover have "(p, adapt_vars t) \<in> (rstep R)\<^sup>*"
    using ta_res_D_rstep[of R t p] join(2,3,5) p
    unfolding snd_conv by blast
  ultimately show ?case by fastforce
next
  case (ctxt ss ts f)
  have ad_Fun: "adapt_vars (Fun f ss) = Fun f (map adapt_vars ss)"
              "adapt_vars (Fun f ss) = Fun f (map adapt_vars ss)"
    using ctxt(4,5) by simp+
  { fix i
    assume i: "i < length ss"
    with ctxt have "gtt_accept (trs_to_ta_A R, trs_to_ta_B R) (ss ! i) (ts ! i)"
      by (auto elim:gtt_accept.cases)
    moreover have "ground (ss ! i) \<and> ground (ts ! i)" using i ctxt(1,3,4) by force
    ultimately have "(adapt_vars (ss ! i), adapt_vars (ts ! i)) \<in> (rstep R)\<^sup>*"
      using ctxt(2,5) i by blast }
  with ad_Fun show ?case
    using args_rsteps_imp_rsteps[of "map adapt_vars ss" "map adapt_vars ts" R f]
    by (simp add: ctxt.hyps)
qed
(*Might want to shorten the proof lemma before*)
lemma abs_R_rtrancl:
  assumes "ground a" "ground b" "ground c" "(a, b) \<in> abs_R (R\<^sup>*)" "(b, c) \<in> abs_R (R\<^sup>*)"
  shows "(a, c) \<in> abs_R (R\<^sup>*)"
proof -
  obtain a' b' c' where obt: "a = abs a'" "b = abs b'" "c = abs c'"
    using assms by blast
  then have "(a', b') \<in> R\<^sup>*" using assms(1,2,4)
    by (smt adapt_vars_reverse fst_conv image_iff relcompE rtrancl_idemp_self_comp snd_conv)
  moreover have "(b', c') \<in> R\<^sup>*" using assms(2,3,5) obt(2,3)
    by (smt adapt_vars_reverse fst_conv image_iff relcompE rtrancl_idemp_self_comp snd_conv)
  ultimately have "(a', c') \<in> R\<^sup>*" by fastforce
  then show ?thesis using obt(1,3)
    by (metis (mono_tags, lifting) fst_conv obt(1) obt(3) rev_image_eqI snd_conv)
qed

lemma gtt_lang_to_rstep_rtrancl:
  assumes "ground_trs R"
  shows "gtt_lang (GTT_trancl (trs_to_ta_A R, trs_to_ta_B R)) \<subseteq> Restr (abs_R ((rstep R)\<^sup>*)) {t. ground t}"
proof -
  have "Restr ((gtt_lang (trs_to_ta_A R, trs_to_ta_B R))\<^sup>*) {t. ground t}
    \<subseteq> Restr (abs_R ((rstep R)\<^sup>*)) {t. ground t}"
  proof -
    { fix u v
      assume asm: "(u, v) \<in> Restr ((gtt_lang (trs_to_ta_A R, trs_to_ta_B R))\<^sup>*) {t. ground t}"
      then have grnd: "ground u" "ground v" by blast+
      moreover have u_v: "(u, v) \<in> (gtt_lang (trs_to_ta_A R, trs_to_ta_B R))\<^sup>*"
        using asm by fast
      have "(u, v) \<in> abs_R ((rstep R)\<^sup>*)"
        using u_v grnd assms
      proof (induction u v rule: rtrancl.induct)
        case (rtrancl_refl a)
        hence "abs (rep a) = a" by auto
        moreover have "(rep a, rep a) \<in> (rstep R)\<^sup>*" by blast
        ultimately show ?case
          by (metis (mono_tags, lifting) fst_conv rev_image_eqI snd_conv)
      next
        case (rtrancl_into_rtrancl a b c)
        then have grnd_b: "ground b" by force
        hence 1: "(a, b) \<in> abs_R ((rstep R)\<^sup>*)"
          using rtrancl_into_rtrancl(3,4,5,6) by fast
        have "(rep b, rep c) \<in> (rstep R)\<^sup>*"
          using grnd_b rtrancl_into_rtrancl(2,5,6) gtt_accept_to_rstep_rtrancl by blast
        hence "(b, c) \<in> abs_R ((rstep R)\<^sup>*)"
          using grnd_b rtrancl_into_rtrancl(5)
          by (smt adapt_vars_adapt_vars fst_conv rev_image_eqI snd_conv)
        with 1 show ?case using abs_R_rtrancl rtrancl_into_rtrancl(4,5) grnd_b by blast
      qed
      hence "(u, v) \<in> Restr (abs_R ((rstep R)\<^sup>*)) {t. ground t}"
        using grnd by blast }
    thus ?thesis by blast
  qed
  thus ?thesis using GTT_trancl_lang[of "(trs_to_ta_A R, trs_to_ta_B R)"]
    by (simp add: Restr_rtrancl_gtt_lang_eq_trancl_gtt_lang)
qed

(*Main Theorem*)
lemma reachability_gtt_equiv:
  assumes "ground_trs R"
  shows "gtt_lang (GTT_trancl (trs_to_ta_A R, trs_to_ta_B R))
    = Restr (abs_R ((rstep R)\<^sup>*)) {t. ground t}"
  using gtt_lang_to_rstep_rtrancl[OF assms] rstep_trancl_to_gtt_trancl[OF assms]
  by auto

end