theory GTT_Transitive_Closure
  imports GTT_Compose
begin

inductive_set \<Delta>_trancl :: "('q, 'f) ta \<Rightarrow> ('q, 'f) ta \<Rightarrow> ('q \<times> 'q) set" for A B where
  \<Delta>_cong: "f ps \<rightarrow> p \<in> ta_rules A \<Longrightarrow> f qs \<rightarrow> q \<in> ta_rules B \<Longrightarrow> length ps = length qs \<Longrightarrow>
   (\<And>i. i < length qs \<Longrightarrow> (ps ! i, qs ! i) \<in> \<Delta>_trancl A B) \<Longrightarrow> (p, q) \<in> \<Delta>_trancl A B"
| \<Delta>_eps1: "(p, p') \<in> ta_eps A \<Longrightarrow> (p, q) \<in> \<Delta>_trancl A B \<Longrightarrow> (p', q) \<in> \<Delta>_trancl A B"
| \<Delta>_eps2: "(q, q') \<in> ta_eps B \<Longrightarrow> (p, q) \<in> \<Delta>_trancl A B \<Longrightarrow> (p, q') \<in> \<Delta>_trancl A B"
| \<Delta>_trans: "(p, q) \<in> \<Delta>_trancl A B \<Longrightarrow> (q, r) \<in> \<Delta>_trancl A B \<Longrightarrow> (p, r) \<in> \<Delta>_trancl A B"

definition GTT_trancl where
  "GTT_trancl G = (ta.make {} (ta_rules (fst G)) (ta_eps (fst G) \<union> \<Delta>_trancl (snd G) (fst G)),
                   ta.make {} (ta_rules (snd G)) (ta_eps (snd G) \<union> \<Delta>_trancl (fst G) (snd G)))"

lemma \<Delta>_trancl_subset_states:
  "\<Delta>_trancl A B \<subseteq> ta_states A \<times> ta_states B"
  apply standard
  subgoal for p q by (induct rule: \<Delta>_trancl.induct)
    (auto simp: ta_states_def r_states_def r_rhs_def split: prod.splits ta_rule.splits)
  done

lemma gtt_syms_GTT_trancl:
  "gtt_syms (GTT_trancl G) = gtt_syms G"
  by (auto simp: GTT_trancl_def ta_syms_def)

lemma \<Delta>_trancl_inv:
  "(\<Delta>_trancl A B)\<inverse> = \<Delta>_trancl B A"
proof -
  have [dest]: "(p, q) \<in> \<Delta>_trancl A B \<Longrightarrow> (q, p) \<in> \<Delta>_trancl B A" for p q A B
    by (induct rule: \<Delta>_trancl.induct) (auto intro: \<Delta>_trancl.intros)
  show ?thesis by auto
qed

lemma GTT_trancl_base:
  "gtt_lang G \<subseteq> gtt_lang (GTT_trancl G)"
  using gtt_accept_mono[of G "GTT_trancl G"] by (auto simp: GTT_trancl_def)

lemma GTT_trancl_trans:
  "gtt_lang (GTT_comp (GTT_trancl G) (GTT_trancl G)) \<subseteq> gtt_lang (GTT_trancl G)"
proof -
  have [dest]: "(p, q) \<in> \<Delta>\<^sub>\<epsilon> (ta.make {} (ta_rules A) (ta_eps A \<union> \<Delta>_trancl B A))
    (ta.make {} (ta_rules B) (ta_eps B \<union> \<Delta>_trancl A B)) \<Longrightarrow> (p, q) \<in> \<Delta>_trancl A B" for p q A B
    by (induct rule: \<Delta>\<^sub>\<epsilon>.induct) (auto intro: \<Delta>_trancl.intros simp: \<Delta>_trancl_inv[of B A, symmetric])
  show ?thesis
    by (auto intro!: gtt_accept_mono[of "GTT_comp (GTT_trancl G) (GTT_trancl G)" "GTT_trancl G"])
      (auto simp: GTT_comp_def GTT_trancl_def)
qed

lemma Restr_rtrancl_gtt_lang_eq_trancl_gtt_lang:
  "Restr ((gtt_lang G)\<^sup>*) {t. ground t} = (gtt_lang G)\<^sup>+"
  by (auto simp: rtrancl_trancl_reflcl simp del: reflcl_trancl dest: tranclD tranclD2)

lemma GTT_trancl_complete:
  "(gtt_lang G)\<^sup>+ \<subseteq> gtt_lang (GTT_trancl G)"
  using GTT_trancl_base subset_trans[OF gtt_comp_lang_complete GTT_trancl_trans]
  by (metis trancl_id trancl_mono_set trans_O_iff)

lemma gtt_lang_ground_ctxt_closed:
  assumes "ground_ctxt C" "(s, t) \<in> gtt_lang \<G>"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> gtt_lang \<G>"
proof -
  { fix i s and ss1 ss2 :: "('f, 'v) term list" assume "i < Suc (length ss1 + length ss2)"
    then have "fill_holes ((map mctxt_of_term ss1 @ MHole # map mctxt_of_term ss2) ! i)
     ((replicate (length ss1) [] @ [s] # replicate (length ss2) []) ! i) = (ss1 @ s # ss2) ! i"
      by (cases "i = length ss1") (auto simp del: upt_Suc simp: nth_append) (* slow *)
  } note [simp] = this
  have *: "fill_holes (MFun f (map mctxt_of_term ss1 @ MHole # map mctxt_of_term ss2)) [s] = Fun f (ss1 @ s # ss2)"
    for f ss1 s ss2 by (auto simp del: upt_Suc intro!: nth_equalityI)
  then have "ctxt.closed {(s, t). gtt_accept \<G> s t}"
    using accept'_closed_ctxt[of "[_]" "[_]" \<G> "MFun _ ((map mctxt_of_term _) @ MHole # (map mctxt_of_term _))"]
    unfolding * by (auto simp: gtt_accept_equiv intro!: one_imp_ctxt_closed)
  then show ?thesis using assms by auto
qed

lemma gtt_langs_ground_ctxt_closed:
  assumes "ground_ctxt C" "(s, t) \<in> (gtt_lang \<G>)\<^sup>*"
  shows "(C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> (gtt_lang \<G>)\<^sup>*"
  using assms(2) gtt_lang_ground_ctxt_closed[OF assms(1), of _ _ \<G>]
  by (induct rule: rtrancl_induct) (auto intro: rtrancl.intros(2))

(* move *)
lemma ground_ctxt_closed_imp_ground_mctxt_closed:
  assumes "\<And>C s t. ground_ctxt C \<Longrightarrow> (s, t) \<in> R\<^sup>* \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> R\<^sup>*"
    "length ss = num_holes C" "length ts = num_holes C" "ground_mctxt C"
    "\<And>i. i < num_holes C \<Longrightarrow> (ss ! i, ts ! i) \<in> Restr (R\<^sup>*) {t. ground t}"
  shows "(fill_holes C ss, fill_holes C ts) \<in> R\<^sup>*"
proof -
  have "i \<le> num_holes C \<Longrightarrow> (fill_holes C ss, fill_holes C (take i ts @ drop i ss)) \<in> R\<^sup>*" for i
  proof (induct i)
    case (Suc i)
    have "num_holes C = Suc (length (take i ts) + length (drop (Suc i) ss))"
      using Suc(2) assms(2,3) by auto
    from fill_holes_ctxt_main[OF this]
    obtain D where [simp]: "\<And>s. D\<langle>s\<rangle> = fill_holes C (take i ts @ s # drop (Suc i) ss)" by metis
    have [simp]: "take i ts @ ss ! i # drop (Suc i) ss = take i ts @ drop i ss"
      "take i ts @ ts ! i # drop (Suc i) ss = take (Suc i) ts @ drop (Suc i) ss"
      using Suc(2) assms(2,3)
      by (simp_all add: Cons_nth_drop_Suc Suc.prems Suc_le_lessD assms(3) take_Suc_conv_app_nth)
    have "ground D\<langle>ss ! i\<rangle>" using assms(2-5) by (auto simp: ground_fill_holes in_set_conv_nth)
    then have "ground_ctxt D" by (metis ground_ctxt_apply)
    then show ?case using Suc(1,2) assms(1)[of D "ss ! i" "ts ! i"] assms(5) by auto
  qed auto
  from this[of "num_holes C"] show ?thesis using assms(2,3) by auto
qed

(* move *)
lemma partition_by_singles:
  "length xs = n \<Longrightarrow> partition_by xs (replicate n (Suc 0)) = map (\<lambda>x. [x]) xs"
  by (induct xs arbitrary: n) auto

(* move *)
lemma map_nth':
  "length xs = n \<Longrightarrow> map ((!) xs) [0..<n] = xs"
  by (auto simp: map_nth)

(* move *)
lemma ground_ctxt_closed_imp_arg_closed:
  assumes "\<And>C s t. ground_ctxt C \<Longrightarrow> (s, t) \<in> R\<^sup>* \<Longrightarrow> (C\<langle>s\<rangle>, C\<langle>t\<rangle>) \<in> R\<^sup>*"
    "length ss = length ts"
    "\<And>i. i < length ts \<Longrightarrow> (ss ! i, ts ! i) \<in> Restr (R\<^sup>*) {t. ground t}"
  shows "(Fun f ss, Fun f ts) \<in> R\<^sup>*"
  using ground_ctxt_closed_imp_ground_mctxt_closed[of R ss "MFun f (replicate (length ts) MHole)" ts]
    assms(2,3) by (auto simp: assms(1) partition_by_singles map_nth' cong: list.map_cong_simp)

lemma trancl_gtt_lang_arg_closed:
  assumes "length ss = length ts" "\<forall>i < length ts. (ss ! i, ts ! i) \<in> (gtt_lang \<G>)\<^sup>+"
  shows "(Fun f ss, Fun f ts) \<in> (gtt_lang \<G>)\<^sup>+"
  using assms(1,2) ground_ctxt_closed_imp_arg_closed[OF _ assms(1), of "gtt_lang \<G>"]
  by (auto simp: gtt_langs_ground_ctxt_closed in_set_conv_nth Restr_rtrancl_gtt_lang_eq_trancl_gtt_lang[symmetric])

lemma \<Delta>_trancl_sound:
  assumes "(p, q) \<in> \<Delta>_trancl A B"
  obtains s t where "(s, t) \<in> (gtt_lang (B, A))\<^sup>+" "p \<in> ta_res A s" "q \<in> ta_res B t"
  using assms
proof (induct arbitrary: thesis rule: \<Delta>_trancl.induct)
  case (\<Delta>_cong f ps p qs q)
  have "\<exists>si ti. (si, ti) \<in> (gtt_lang (B, A))\<^sup>+ \<and> ps ! i \<in> ta_res A (si) \<and>
      qs ! i \<in> ta_res B (ti)" if "i < length qs" for i
    using \<Delta>_cong(5)[OF that] by metis
  then obtain ss ts where
    "\<And>i. i < length qs \<Longrightarrow> (ss i, ts i) \<in> (gtt_lang (B, A))\<^sup>+ \<and> ps ! i \<in> ta_res A (ss i) \<and> qs ! i \<in> ta_res B (ts i)" by metis
  then show ?case using \<Delta>_cong(1-5)
    by (intro \<Delta>_cong(6)[of "Fun f (map ss [0..<length ps])" "Fun f (map ts [0..<length qs])"])
      (auto intro!: trancl_gtt_lang_arg_closed)
next
  case (\<Delta>_eps1 p p' q) then show ?case by (metis r_into_rtrancl ta_res_eps)
next
  case (\<Delta>_eps2 q q' p) then show ?case by (metis r_into_rtrancl ta_res_eps)
next
  case (\<Delta>_trans p q r)
  guess s1 t1 using \<Delta>_trans(2) . note 1 = this
  guess s2 t2 using \<Delta>_trans(4) . note 2 = this
  have "(t1, s2) \<in> gtt_lang (B, A)" using 1(1,3) 2(1,2)
    by (auto simp: Restr_rtrancl_gtt_lang_eq_trancl_gtt_lang[symmetric])
  then have "(s1, t2) \<in> (gtt_lang (B, A))\<^sup>+" using 1(1) 2(1)
    by (meson trancl.trancl_into_trancl trancl_trans)
  then show ?case using 1(2) 2(3) by (auto intro: \<Delta>_trans(5)[of s1 t2])
qed

lemma GTT_trancl_sound_aux:
  assumes "p \<in> ta_res (ta.make {} (ta_rules A) (ta_eps A \<union> \<Delta>_trancl B A)) s" "ground s"
  shows "\<exists>t. (s, t) \<in> (gtt_lang (A, B))\<^sup>+ \<and> p \<in> ta_res A t"
  using assms
proof (induct s arbitrary: p)
  case (Fun f ss)
  obtain qs q where q: "f qs \<rightarrow> q \<in> ta_rules A" "(q, p) \<in> (ta_eps A \<union> \<Delta>_trancl B A)\<^sup>*" "length qs = length ss"
   "\<And>i. i < length ss \<Longrightarrow> qs ! i \<in> ta_res (ta.make {} (ta_rules A) (ta_eps A \<union> \<Delta>_trancl B A)) (ss ! i)"
    using Fun(2) by auto
  have "\<And>i. i < length ss \<Longrightarrow> \<exists>ti. (ss ! i, ti) \<in> (gtt_lang (A, B))\<^sup>+ \<and> qs ! i \<in> ta_res A (ti)"
    using Fun(1)[OF nth_mem q(4)] Fun(3) by fastforce
  then obtain ts where ts: "\<And>i. i < length ss \<Longrightarrow> (ss ! i, ts i) \<in> (gtt_lang (A, B))\<^sup>+ \<and> qs ! i \<in> ta_res A (ts i)"
    by metis
  show ?case using q(2)
  proof (induct rule: rtrancl_induct)
    case base then show ?case using ts q(1,3)
      by (auto intro!: trancl_gtt_lang_arg_closed exI[of _ "Fun f (map ts [0..<length ss])"] exI[of _ qs] exI[of _ q])
  next
    case (step p p')
    guess s' using step(3) .. note s' = this
    show ?case using step(2)
    proof
      assume "(p, p') \<in> ta_eps A" then show ?thesis using s'
        by (auto intro!: exI[of _ s'] simp: r_into_rtrancl ta_res_eps)
    next
      assume "(p, p') \<in> \<Delta>_trancl B A"
      from \<Delta>_trancl_sound[OF this] guess s t . note st = this
      have "(s', s) \<in> gtt_lang (A, B)" using st s'
        by (auto simp: Restr_rtrancl_gtt_lang_eq_trancl_gtt_lang[symmetric])
      then have "(Fun f ss, t) \<in> (gtt_lang (A, B))\<^sup>+"
        using s' st(1) by (meson trancl.trancl_into_trancl trancl_trans)
      then show ?thesis using st(3) by auto
    qed
  qed
qed auto

lemma GTT_trancl_sound:
  "gtt_lang (GTT_trancl G) \<subseteq> (gtt_lang G)\<^sup>+"
proof -
  note [dest] = GTT_trancl_sound_aux
  have "gtt_accept (GTT_trancl G) s t \<Longrightarrow> ground s \<Longrightarrow> ground t \<Longrightarrow> (s, t) \<in> (gtt_lang G)\<^sup>+" for s t
  proof (induct rule: gtt_accept.induct)
    case (join q s t)
    obtain s' where "(s, s') \<in> (gtt_lang G)\<^sup>+" "q \<in> ta_res (fst G) s'" using join(1,3)
      by (auto simp: GTT_trancl_def)
    moreover obtain t' where "(t', t) \<in> (gtt_lang G)\<^sup>+" "q \<in> ta_res (snd G) t'" using join(2,4)
      by (auto simp: GTT_trancl_def gtt_lang_swap[of "fst G" "snd G", symmetric] trancl_converse)
    moreover have "(s', t') \<in> gtt_lang G" using calculation
      by (auto simp: Restr_rtrancl_gtt_lang_eq_trancl_gtt_lang[symmetric])
    ultimately show "(s, t) \<in> (gtt_lang G)\<^sup>+" by (meson trancl.trancl_into_trancl trancl_trans)
  qed (auto intro!: trancl_gtt_lang_arg_closed)
  then show ?thesis by auto
qed

lemma GTT_trancl_lang:
  "gtt_lang (GTT_trancl G) = (gtt_lang G)\<^sup>+"
  using GTT_trancl_sound GTT_trancl_complete by blast

end