theory GTT_Transitive_Closure_Impl
  imports GTT_Transitive_Closure GTT_Compose_Impl GTT_Impl Horn_List
begin

text \<open>This is very similar to @{file "GTT_Compose_Impl.thy"}.\<close>

definition \<Delta>_trancl_rules :: "('q, 'f) ta \<Rightarrow> ('q, 'f) ta \<Rightarrow> ('q \<times> 'q) horn set" where
  "\<Delta>_trancl_rules A B =
    \<Delta>\<^sub>\<epsilon>_rules A B \<union> {[(p, q), (q, r)] \<rightarrow>\<^sub>h (p, r) |p q r. True}"

locale \<Delta>_trancl_horn =
  fixes A :: "('q, 'f) ta" and B :: "('q, 'f) ta"
begin

sublocale horn "\<Delta>_trancl_rules A B" .

lemma \<Delta>_trancl_infer0:
  "infer0 = horn.infer0 (\<Delta>\<^sub>\<epsilon>_rules A B)"
  by (auto simp: \<Delta>\<^sub>\<epsilon>_rules_def \<Delta>_trancl_rules_def horn.infer0_def)

lemma \<Delta>_trancl_infer1:
  "infer1 pq X = horn.infer1 (\<Delta>\<^sub>\<epsilon>_rules A B) pq X \<union>
   {(r, snd pq) |r p'. (r, p') \<in> X \<and> p' = fst pq} \<union>
   {(fst pq, r) |q' r. (q', r) \<in> (insert pq X) \<and> q' = snd pq}"
  unfolding \<Delta>_trancl_rules_def horn_infer1_union Un_assoc
  apply (intro arg_cong2[of _ _ _ _ "(\<union>)"] HOL.refl)
  by (auto simp: horn.infer1_def simp flip: ex_simps(1)) force+

lemma \<Delta>_trancl_sound:
  "\<Delta>_trancl A B = saturate"
proof (intro set_eqI iffI, goal_cases lr rl)
  case (lr x) obtain p q where x: "x = (p, q)" by (cases x)
  show ?case using lr unfolding x
  proof (induct)
    case (\<Delta>_cong f ps p qs q) show ?case
      apply (intro infer[of "zip ps qs" "(p, q)"])
      subgoal using \<Delta>_cong(1-3) by (auto simp: \<Delta>_trancl_rules_def \<Delta>\<^sub>\<epsilon>_rules_def)
      subgoal using \<Delta>_cong(3,5) by (auto simp: zip_nth_conv)
      done
  next
    case (\<Delta>_eps1 p p' q) then show ?case
      by (intro infer[of "[(p, q)]" "(p', q)"]) (auto simp: \<Delta>_trancl_rules_def \<Delta>\<^sub>\<epsilon>_rules_def)
  next
    case (\<Delta>_eps2 q q' p) then show ?case
      by (intro infer[of "[(p, q)]" "(p, q')"]) (auto simp: \<Delta>_trancl_rules_def \<Delta>\<^sub>\<epsilon>_rules_def)
  next
    case (\<Delta>_trans p q r) then show ?case
      by (intro infer[of "[(p,q), (q,r)]" "(p, r)"]) (auto simp: \<Delta>_trancl_rules_def \<Delta>\<^sub>\<epsilon>_rules_def)
  qed
next
  case (rl x) obtain p q where x: "x = (p, q)" by (cases x)
  show ?case using rl unfolding x
  proof (induct)
    case (infer as a) then show ?case
      using \<Delta>_cong[of _ "map fst as" "fst a" A "map snd as" "snd a" B]
        \<Delta>_eps1[of _ "fst a" A "snd a" B] \<Delta>_eps2[of _ "snd a" B "fst a" A]
        \<Delta>_trans[of "fst a" "snd (hd as)" A B "snd a"] by (auto simp: \<Delta>_trancl_rules_def \<Delta>\<^sub>\<epsilon>_rules_def)
  qed
qed

end

definition \<Delta>_trancl_infer0 where
  "\<Delta>_trancl_infer0 A B = \<Delta>\<^sub>\<epsilon>_infer0_list A B"

definition \<Delta>_trancl_infer1 :: "('q, 'f) ta_list \<Rightarrow> ('q, 'f) ta_list \<Rightarrow> 'q \<times> 'q \<Rightarrow> ('q \<times> 'q) list \<Rightarrow> ('q \<times> 'q) list" where
  "\<Delta>_trancl_infer1 A B pq bs =
    \<Delta>\<^sub>\<epsilon>_infer1_list A B pq bs @
    map (\<lambda>(r, p'). (r, snd pq)) (filter (\<lambda>(r, p') \<Rightarrow> p' = fst pq) bs) @
    map (\<lambda>(q', r). (fst pq, r)) (filter (\<lambda>(q', r) \<Rightarrow> q' = snd pq) (pq # bs))"

locale \<Delta>_trancl_list =
  fixes A :: "('q, 'f) ta_list" and B :: "('q, 'f) ta_list"
begin

sublocale \<Delta>_trancl_horn "ta_of A" "ta_of B" .

sublocale l: horn_list "\<Delta>_trancl_rules (ta_of A) (ta_of B)" "\<Delta>_trancl_infer0 A B" "\<Delta>_trancl_infer1 A B"
  apply (unfold_locales)
  unfolding \<Delta>_trancl_rules_def horn_infer0_union horn_infer1_union
    \<Delta>_trancl_infer0_def \<Delta>_trancl_infer1_def \<Delta>\<^sub>\<epsilon>_list.infer set_append
  by (auto simp flip: ex_simps(1) simp: horn.infer0_def horn.infer1_def intro!: arg_cong2[of _ _ _ _ "(\<union>)"])

lemmas saturate_impl_sound = l.saturate_impl_sound
lemmas saturate_impl_complete = l.saturate_impl_complete

end

definition \<Delta>_trancl_impl :: "('q, 'f) ta_list \<Rightarrow> ('q, 'f) ta_list \<Rightarrow> ('q \<times> 'q) list option" where
  "\<Delta>_trancl_impl A B = horn_list_impl.saturate_impl (\<Delta>_trancl_infer0 A B) (\<Delta>_trancl_infer1 A B)"

lemma \<Delta>_trancl_impl_sound:
  "\<Delta>_trancl_impl A B = Some xs \<Longrightarrow> set xs = \<Delta>_trancl (ta_of A) (ta_of B)"
  using \<Delta>_trancl_list.saturate_impl_sound unfolding \<Delta>_trancl_impl_def \<Delta>_trancl_horn.\<Delta>_trancl_sound .

lemma \<Delta>_trancl_impl_complete:
  fixes A B :: "('q, 'f) ta_list"
  shows "\<Delta>_trancl_impl A B \<noteq> None"
proof -
  note * = \<Delta>_trancl_subset_states[of "ta_of A" "ta_of B"]
  have "finite (\<Delta>_trancl (ta_of A) (ta_of B))"
    by (cases A; cases B; rule finite_subset[OF *])
      (auto simp: ta_of_def ta_states_def r_states_def intro!: finite_cartesian_product)
  then show ?thesis unfolding \<Delta>_trancl_impl_def
    by (intro \<Delta>_trancl_list.saturate_impl_complete) (auto simp: \<Delta>_trancl_horn.\<Delta>_trancl_sound)
qed

definition GTT_trancl_impl :: "('q, 'f) gtt_list \<Rightarrow> (_, 'f) gtt_list option" where
  "GTT_trancl_impl G = (case (\<Delta>_trancl_impl (snd G) (fst G), \<Delta>_trancl_impl (fst G) (snd G)) of
     (Some xs, Some ys) \<Rightarrow> Some
     (ta_list [] (tal_rules (fst G)) (tal_eps (fst G) @ xs),
      ta_list [] (tal_rules (snd G)) (tal_eps (snd G) @ ys))
     | _ \<Rightarrow> None
     )"

lemma GTT_trancl_impl_sound:
  "GTT_trancl_impl G = Some H \<Longrightarrow> gtt_of H = GTT_trancl (gtt_of G)"
  by (auto simp: GTT_trancl_def GTT_trancl_impl_def split: option.splits dest!: \<Delta>_trancl_impl_sound
   intro!: ta_of_eq_ta_makeI)

lemma GTT_trancl_impl_complete:
  "GTT_trancl_impl G \<noteq> None"
  using \<Delta>_trancl_impl_complete[of "snd G" "fst G"] \<Delta>_trancl_impl_complete[of "fst G" "snd G"]
  by (auto simp: GTT_trancl_impl_def)

(*
export_code GTT_trancl_impl checking SML
export_code GTT_trancl_impl checking Haskell
export_code GTT_trancl_impl in Haskell
*)

definition \<Delta>\<^sub>_trancl_infer0_rbt where
  "\<Delta>\<^sub>_trancl_infer0_rbt A B = \<Delta>\<^sub>\<epsilon>_infer0_rbt A B"

definition \<Delta>_trancl_infer1_rbt where
  "\<Delta>_trancl_infer1_rbt A B pq X = \<Delta>\<^sub>\<epsilon>_infer1_rbt A B pq X @
    map (\<lambda>(r, p'). (r, snd pq)) (filter (\<lambda>(r, p') \<Rightarrow> p' = fst pq) (RBT_Impl.keys X)) @
    map (\<lambda>(q', r). (fst pq, r)) (filter (\<lambda>(q', r) \<Rightarrow> q' = snd pq) (pq # (RBT_Impl.keys X)))"

lemma \<Delta>_trancl_infer0_rbt:
  assumes A': "(A', A) \<in> dflt_ta_rel"
  assumes B': "(B', B) \<in> dflt_ta_rel"
  shows "(\<Delta>\<^sub>_trancl_infer0_rbt A' B', horn.infer0 (\<Delta>_trancl_rules A B)) \<in> \<langle>Id\<rangle>it_set_rel"
  unfolding \<Delta>_trancl_horn.\<Delta>_trancl_infer0 \<Delta>\<^sub>_trancl_infer0_rbt_def
  by (rule \<Delta>\<^sub>\<epsilon>_infer0_rbt[OF assms])

lemma \<Delta>_trancl_infer1_rbt:
  assumes A': "(A', A) \<in> dflt_ta_rel"
  assumes B': "(B', B) \<in> dflt_ta_rel"
  assumes pq': "(pq', pq) \<in> Id"
  assumes X': "(X', X) \<in> \<langle>Id\<rangle>dflt_rs_rel"
  shows "(\<Delta>_trancl_infer1_rbt A' B' pq' X', horn.infer1 (\<Delta>_trancl_rules A B) pq X) \<in> \<langle>Id\<rangle>it_set_rel"
  unfolding \<Delta>_trancl_infer1_rbt_def \<Delta>_trancl_horn.\<Delta>_trancl_infer1 Un_assoc
  using assms(3) assms(4)[THEN keys_it]
  apply (intro union_it_set_rel \<Delta>\<^sub>\<epsilon>_infer1_rbt[OF assms])
  by (auto 0 0 simp: it_set_rel_def br_def split: prod.split) force+

schematic_goal \<Delta>_trancl_impl_rbt:
  assumes "(A', A) \<in> dflt_ta_rel"
  assumes "(B', B) \<in> dflt_ta_rel"
  shows "(?f, horn.saturate_impl (\<Delta>_trancl_rules A B)) \<in> \<langle>\<langle>Id\<rangle>dflt_rs_rel\<rangle>nres_rel"
  apply (rule horn_linorder.saturate_impl_rbt)
   apply (rule \<Delta>_trancl_infer0_rbt[OF assms])
  apply (intro fun_relI)
  apply (rule \<Delta>_trancl_infer1_rbt[OF assms])
  apply assumption+
  done

concrete_definition \<Delta>_trancl_rbt uses \<Delta>_trancl_impl_rbt
prepare_code_thms \<Delta>_trancl_rbt_def

schematic_goal GTT_trancl_rbt_aux:
  assumes [autoref_rules]: "(G', G) \<in> dflt_gtt_rel"
  defines "\<Delta>12 \<equiv> \<Delta>_trancl (snd G) (fst G)" and "\<Delta>21 \<equiv> \<Delta>_trancl (fst G) (snd G)"
  assumes \<Delta>: "(\<Delta>12', \<Delta>12) \<in> \<langle>Id \<times>\<^sub>r Id\<rangle>dflt_rs_rel" "(\<Delta>21', \<Delta>21) \<in> \<langle>Id \<times>\<^sub>r Id\<rangle>dflt_rs_rel"
  notes [autoref_rules] = \<Delta>[THEN keys_list]
  shows "(?f, GTT_trancl G) \<in> dflt_gtt_rel"
  unfolding GTT_trancl_def \<Delta>12_def[symmetric] \<Delta>21_def[symmetric]
  by autoref

concrete_definition GTT_trancl_rbt_aux uses GTT_trancl_rbt_aux

schematic_goal GTT_trancl_rbt:
  assumes G: "(G', G) \<in> dflt_gtt_rel"
  notes G' = param_prod(4,5)[THEN fun_relD, OF G]
  notes * = horn.saturate_impl_sound[unfolded SPEC_eq_is_RETURN, of "\<Delta>_trancl_rules _ _", unfolded \<Delta>_trancl_horn.\<Delta>_trancl_sound[symmetric]]
  shows "(?f, RETURN (GTT_trancl G)) \<in> \<langle>dflt_gtt_rel\<rangle>nres_rel"
  apply (rule HOL.back_subst[of "\<lambda>x. (x, _) \<in> _"])
  apply (rule bind_RETURN[OF conc_trans_additional(1)[OF \<Delta>_trancl_rbt.refine[OF G'(2) G'(1), THEN nres_relD] *, THEN nres_relI]])
  apply (rule bind_RETURN[OF conc_trans_additional(1)[OF \<Delta>_trancl_rbt.refine[OF G'(1) G'(2), THEN nres_relD] *, THEN nres_relI]])
  apply (rule param_RETURN[THEN fun_relD])
   apply (rule GTT_trancl_rbt_aux[OF G, unfolded prod_rel_id_simp]; assumption+)
  unfolding nres_of_simps(3)[symmetric] nres_of_bind[symmetric]
  by (rule refl)

concrete_definition GTT_trancl_rbt uses GTT_trancl_rbt

export_code GTT_trancl_rbt in Haskell

end