theory LLRG_Confluence
  imports Ground_Confluence Ord.Signature_Extension LV_to_GTT
begin

definition llrg :: "('f,'v) trs \<Rightarrow> bool" where
  "llrg R \<equiv> \<forall>s \<in> R. (linear_term (fst s) \<and> ground (snd s))"

primrec inv_const :: "('f \<times> nat) \<Rightarrow> 'v \<Rightarrow> ('f,'v) term \<Rightarrow> ('f, 'v) term" where
  "inv_const g y (Var x) = (Var x)"
| "inv_const g y (Fun f ss) = (if (Fun f ss = Fun (fst g) []) 
                               then (Var y) else (Fun f (map (inv_const g y) ss)))"

fun inv_const_ctxt :: "('f \<times> nat) \<Rightarrow> 'v \<Rightarrow> ('f,'v) ctxt \<Rightarrow> ('f, 'v) ctxt" where
  "inv_const_ctxt g y Hole = Hole"
| "inv_const_ctxt g y ((More f ss1 C ss2)) 
              = (More f (map (inv_const g y) ss1) (inv_const_ctxt g y C) (map (inv_const g y) ss2))"

definition subin :: "('f,'v) subst \<Rightarrow>('f \<times> nat) \<Rightarrow> 'v  \<Rightarrow> ('f, 'v) subst" where
  "subin f (g::('f \<times> nat)) (y::'v)  =  (\<lambda> (t::'v). inv_const g y (f t))"

lemma inv_cons_sigma:
  fixes g:: "'f"
  defines "\<sigma> \<equiv> \<lambda> t. Fun g []"
  assumes "ground u"
  shows "(inv_const (g,0) y u) \<cdot> \<sigma> = u"
  using assms
proof (induction u)
  case (Fun f ss)
  {fix i
   assume i:"i < length ss"
   hence "ground (ss!i)" using Fun(3) by auto
   hence "inv_const (g, 0) y (ss!i) \<cdot> \<sigma> = (ss!i)" using Fun(1,2) i
     by force}
  then show ?case 
    by (smt \<sigma>_def fst_conv inv_const.simps(2) length_map map_nth_eq_conv subst_apply_term.simps(1) subst_apply_term.simps(2))  
qed (simp)

lemma subst_sig:
  assumes "funas_term u \<subseteq> F"
  shows "funas_term (u \<cdot> (\<lambda> t. Fun g [])) \<subseteq> F \<union> {(g,0)}"
  using assms apply (induction u) apply simp by fastforce

lemma rstep_sig_extn:
  assumes "(u, v) \<in> (sig_step F (rstep R))"
  shows "(u \<cdot> (\<lambda> t. Fun g []), v \<cdot> (\<lambda> t. Fun g [])) \<in> (sig_step (F \<union> {(g,0)}) (rstep R))"
  using subst_sig[of _ F] assms unfolding sig_step_def by auto

lemma rstep_trancl_sig_extn:
  assumes "(u, v) \<in> (sig_step F (rstep R))^*"
  shows "(u \<cdot> (\<lambda> t. Fun g []), v \<cdot> (\<lambda> t. Fun g [])) \<in> (sig_step (F \<union> {(g,0)}) (rstep R))^*"
  using assms 
proof (induction rule:rtrancl_induct)
  case (step v w)
  have "(v \<cdot> (\<lambda> t. Fun g []), w \<cdot> (\<lambda> t. Fun g [])) \<in> (sig_step (F \<union> {(g,0)}) (rstep R))"
    using rstep_sig_extn[OF step(2)] .
  thus ?case using step(3) by simp
qed (simp)

lemma subst_filter:
  fixes g::"'f" and y::"'v"
  defines \<sigma>:"\<sigma> \<equiv> (\<lambda> t. Fun g [])" 
  assumes "ground (u \<cdot> \<rho>)"
  shows "(u \<cdot> \<rho>) = (u \<cdot> (subin \<rho> (g,0) y)) \<cdot> \<sigma>"
  using assms
proof (induction "u \<cdot> \<rho>" arbitrary: u)
  case (Var x)
  then show ?case unfolding subin_def by force  
next
  case (Fun f ss)
  then show ?case
  proof (cases u)
    case (Var z)
    hence ic:"(u \<cdot> (subin \<rho> (g,0) y)) = inv_const (g,0) y (Fun f ss)"
      unfolding subin_def Fun by auto
    moreover have "(inv_const (g,0) y (Fun f ss)) \<cdot> \<sigma> = Fun f ss"
      using inv_cons_sigma[of "Fun f ss" g y] Fun(2,4) unfolding \<sigma> 
      by argo 
    ultimately have "u \<cdot> subin \<rho> (g, 0) y \<cdot> \<sigma> = Fun f ss" 
      by argo
    then show ?thesis unfolding Fun(2)[symmetric] by argo
  next
    case (Fun h ts)
    hence ts_ss:"h = f" "length ss = length ts" using Fun.hyps(2) 
      by fastforce+
    { fix i
      assume i:"i < length ss"
      hence ts_i:"(ts!i) \<cdot> \<rho> = (ss!i)" using ts_ss  Fun.hyps(2) unfolding Fun
        by simp
      moreover have "ground (ss!i)" using Fun.prems i unfolding Fun.hyps(2)[symmetric] by auto
      ultimately have "(ts!i) \<cdot> \<rho> = (ts!i) \<cdot> subin \<rho> (g, 0) y \<cdot> \<sigma>"
        using Fun.hyps(1,3) 
        by (metis \<sigma> i nth_mem)}  
    then show ?thesis using Fun ts_ss
      by (smt map_eq_conv' nth_map subst_apply_term.simps(2) term.distinct(1) term.sel(4) weak_match.elims(2) weak_match_match)  
  qed
qed

lemma inv_const_filter_funas:
  fixes g::"'f" and y::"'v" and u::"('f,'v) term"
  defines \<sigma>:"\<sigma> \<equiv> (\<lambda> t. Fun g [])"
  assumes  "funas_term u \<subseteq> F \<union> {(g,0)}"  "(g,0) \<notin> F"
  shows "funas_term (inv_const (g,0) y u)  \<subseteq> F"
  using assms 
proof (induction u)
 case (Fun f ss)
  then show ?case
    proof (cases "f = g \<and> ss = []")
      case True
      hence "inv_const (g,0) y (Fun f ss) = Var y" by simp
      then show ?thesis 
        by (metis empty_subsetI funas_term.simps(1))   
    next
      case False
      hence "(f,length ss) \<noteq> (g,0)" by fastforce
      hence f:"(f, length ss) \<in> F" using Fun(3) by auto 
      {fix i
       assume i:"i < length ss"
       hence "funas_term (ss!i) \<subseteq> F \<union> {(g, 0)}" using Fun(3) funas_term.simps(2)[of f ss] by force
       hence "funas_term (inv_const (g,0) y (ss!i)) \<subseteq> F" using Fun(1-4) i  by auto}
     hence "\<forall>i < length ss. funas_term (inv_const (g,0) y (ss!i)) \<subseteq> F" by blast  
     hence "\<Union>set (map funas_term (map (inv_const (g,0) y) ss)) \<subseteq> F" 
            apply simp
     proof -
       assume a1: "\<forall>i<length ss. funas_term (inv_const (g, 0) y (ss ! i)) \<subseteq> F"
       obtain nn :: "('f, 'v) Term.term list \<Rightarrow> ('f, 'v) Term.term \<Rightarrow> nat" where
         "\<forall>x0 x1. (\<exists>v2<length x0. x1 = x0 ! v2) = (nn x0 x1 < length x0 \<and> x1 = x0 ! nn x0 x1)"
         by moura
       then have f2: "\<forall>t ts. t \<notin> set ts \<or> nn ts t < length ts \<and> t = ts ! nn ts t"
         by (meson in_set_idx)
       obtain tt :: "('f, 'v) Term.term set \<Rightarrow> (('f, 'v) Term.term \<Rightarrow> ('f \<times> nat) set) \<Rightarrow> ('f \<times> nat) set \<Rightarrow> ('f, 'v) Term.term" where
         "\<forall>x0 x1 x2. (\<exists>v3. x2 = x1 v3 \<and> v3 \<in> x0) = (x2 = x1 (tt x0 x1 x2) \<and> tt x0 x1 x2 \<in> x0)"
         by moura
       then have f3: "\<forall>P f T. P \<notin> f ` T \<or> P = f (tt T f P) \<and> tt T f P \<in> T"
         by (meson imageE)
       obtain PP :: "('f \<times> nat) set \<Rightarrow> ('f \<times> nat) set set \<Rightarrow> ('f \<times> nat) set" where
         "\<forall>x0 x1. (\<exists>v2. v2 \<in> x1 \<and> \<not> v2 \<subseteq> x0) = (PP x0 x1 \<in> x1 \<and> \<not> PP x0 x1 \<subseteq> x0)"
         by moura
       then have f4: "\<forall>P Pa. PP Pa P \<in> P \<and> \<not> PP Pa P \<subseteq> Pa \<or> \<Union>P \<subseteq> Pa"
         by (meson Union_least)
       moreover
       { assume "funas_term (inv_const (g, 0) y (ss ! nn ss (tt (set ss) (funas_term \<circ> inv_const (g, 0) y) (PP F ((funas_term \<circ> inv_const (g, 0) y) ` set ss))))) \<subseteq> F"
         then have "UNION (set ss) (funas_term \<circ> inv_const (g, 0) y) \<subseteq> F \<or> (PP F ((funas_term \<circ> inv_const (g, 0) y) ` set ss) \<noteq> (funas_term \<circ> inv_const (g, 0) y) (tt (set ss) (funas_term \<circ> inv_const (g, 0) y) (PP F ((funas_term \<circ> inv_const (g, 0) y) ` set ss))) \<or> tt (set ss) (funas_term \<circ> inv_const (g, 0) y) (PP F ((funas_term \<circ> inv_const (g, 0) y) ` set ss)) \<notin> set ss) \<or> PP F ((funas_term \<circ> inv_const (g, 0) y) ` set ss) \<notin> (funas_term \<circ> inv_const (g, 0) y) ` set ss \<or> PP F ((funas_term \<circ> inv_const (g, 0) y) ` set ss) \<subseteq> F"
           using f4 f2 by (metis (no_types) comp_def) }
       ultimately show "UNION (set ss) (funas_term \<circ> inv_const (g, 0) y) \<subseteq> F"
         using f3 f2 a1 by meson
     qed 
     then show ?thesis using f  funas_term.simps(2)[of f "map (inv_const (g,0) y) ss"]
       by (metis (no_types, lifting) Un_insert_left empty_subsetI funas_term.elims 
                 insert_subset inv_const.simps(2) length_map sup_bot.left_neutral term.distinct(1))
    qed
qed (simp) 

lemma inv_contxt_const:
  "inv_const (g, 0) y C\<langle>s\<rangle> = (inv_const_ctxt (g,0) y C)\<langle>inv_const (g, 0) y s\<rangle>"
proof (induction C)
  case (More f  ss1 C ss2)
  let ?C' = "More f  ss1 C ss2"
  let ?ss1 = "map (inv_const (g,0) y) ss1"
  let ?ss2 = "map (inv_const (g,0) y) ss2"
  let ?C =  "inv_const_ctxt (g,0) y C"
  have "?C'\<langle>s\<rangle> \<noteq> Fun g []" by simp
  hence 1:"inv_const (g, 0) y ?C'\<langle>s\<rangle> = (Fun f (map (inv_const (g,0) y) (ss1@C\<langle>s\<rangle>#ss2)))"
    by force
  hence 2:"... =   (Fun f (?ss1@(inv_const (g,0) y C\<langle>s\<rangle>)#?ss2))"
    by simp
  with More have 3:
     "... = (Fun f (?ss1@  ?C\<langle>inv_const (g, 0) y s\<rangle> #?ss2))"   
    by blast
  hence 4:"... = (More f ?ss1 ?C ?ss2)\<langle>inv_const (g, 0) y s\<rangle>"
    by auto
  hence 5:"... = (inv_const_ctxt (g,0) y (More f ss1 C ss2))\<langle>inv_const (g, 0) y s\<rangle>"
    by simp
  thus ?case using 1 2 3 4 by argo
qed (auto) 
  
lemma inv_const_subst:
  assumes "funas_term s \<subseteq> F" "(g,0) \<notin> F"
  shows "inv_const (g, 0) y (s \<cdot> \<rho>) = s \<cdot> (subin \<rho> (g,0) y)" using assms
proof (induction s)
  case (Fun f ss)
  then show ?case 
    proof (cases "f = g \<and> ss = []")
      case True
      then show ?thesis using Fun  
        by (simp add: Fun.IH) 
    next
      case False
      let ?ss = "map (\<lambda>i. i \<cdot> \<rho>) ss" 
      have "(Fun f ss) \<cdot> \<rho> = Fun f ?ss"  by force
      hence 1:"inv_const (g, 0) y ((Fun f ss) \<cdot> \<rho>) = Fun f (map (inv_const (g,0) y) ?ss)"
        using False 
        by force
      {fix i
        assume i:"i < length ss"
        hence "funas_term (ss!i) \<subseteq> F" using Fun(2) by fastforce
        with i Fun(1,3) have "inv_const (g, 0) y (?ss!i) =
                                        (ss!i) \<cdot> subin \<rho> (g, 0) y"
          by force}
      with 1 show ?thesis 
        by (smt length_map map_eq_conv' subst_apply_term.simps(2))     
    qed
qed (simp add: subin_def) 

lemma inv_const_rstep:
  assumes "funas_trs R \<subseteq> F" "(s,t) \<in> (rstep R)" "(g,0) \<notin> F"
  shows "(inv_const (g,0) y s, inv_const (g,0) y t) \<in> (rstep R)"
  using assms(2,1,3)
proof (induction rule: rstepE)
  case (rstep s t C \<rho> l r)
  let ?\<rho> = " (subin \<rho> (g,0) y)" 
  let ?ic = "inv_const (g, 0) y"
  have funas:
     "funas_term l \<subseteq> F" "funas_term r \<subseteq> F" 
      using rstep(3,6) unfolding funas_trs_def funas_rule_def
      by force+
    hence 1:"?ic (l \<cdot> \<rho>) = l \<cdot> ?\<rho>" 
          "?ic  (r \<cdot> \<rho>) = r \<cdot> ?\<rho>"
      using rstep(7) by (simp_all add: inv_const_subst)
    hence "(?ic (l \<cdot> \<rho>), ?ic (r \<cdot> \<rho>)) \<in> rstep R"
      using rstep(3) by auto
    hence "((inv_const_ctxt (g,0) y C)\<langle>?ic (l \<cdot> \<rho> )\<rangle>, 
            (inv_const_ctxt (g,0) y C)\<langle>?ic (r \<cdot> \<rho>)\<rangle>) \<in> rstep R"
      by blast 
    then show ?case using inv_contxt_const[symmetric]
      by (simp add: \<open>\<And>y s g C. (inv_const_ctxt (g, 0) y C)\<langle>inv_const (g, 0) y s\<rangle> = inv_const (g, 0) y C\<langle>s\<rangle>\<close> rstep.hyps(1,2,4,5))
qed

lemma vars_term_funas_term:
  assumes "x \<in> vars_term t" "(Var x) \<cdot> \<sigma> = Fun f ss"
  shows "(f, length ss) \<in> funas_term (t \<cdot> \<sigma>)"
  using assms 
proof (induction t)
  case (Var y)
  have "x = y" using Var(1) by simp
  then show ?case  using assms(2) by auto  
next
  case (Fun g ts)
  obtain i where i:"i < length ts" "x \<in> vars_term (ts!i)" using Fun(2)
    using var_imp_var_of_arg by force 
  hence ts_i:"(ts!i) \<in> set ts" by simp
  have f_len:"(f, length ss) \<in> funas_term (ts!i \<cdot> \<sigma>)" using   Fun(1)[OF ts_i i(2) Fun(3)] .
  moreover have "(ts!i \<cdot> \<sigma>) \<unlhd> (Fun g ts \<cdot> \<sigma>)" using i(1) by simp
  hence "funas_term (ts!i \<cdot> \<sigma>) \<subseteq> funas_term (Fun g ts  \<cdot> \<sigma>)"
    using supteq_imp_funas_term_subset by blast
  then show ?case using i(1) f_len  
    using Fun.IH Fun.prems(2) i(2) nth_mem by blast  
qed

lemma const_ctxt:
  assumes "C\<langle>s\<rangle> = Fun f []"
  shows "C = Hole \<and> s = Fun f []"
  using assms
  by (metis ctxt.cop_nil nectxt_imp_supt_ctxt supt_const) 
  
lemma const_subst:
  assumes "l \<cdot> \<rho> = Fun g []" 
  shows "is_Var l \<or> l = Fun g []"
  using assms 
proof (cases l)
  case (Fun f ss)
  hence "f = g \<and> ss = []" using assms by force
  then show ?thesis using Fun by meson
qed (fast)

lemma Fun_ctxt_closed:
  assumes "(u,v) \<in> rstep R"
  shows "(Fun f (ss@(u#ts)), Fun f (ss@(v#ts))) \<in> rstep R"
  using assms by (simp add: ctxt_closed_one ctxt_closed_rstep)

lemma funas_subterm:
  assumes "funas_term (Fun f ss) \<subseteq> F"
  shows "\<forall>i< length ss. funas_term (ss!i) \<subseteq> F"
  using assms by fastforce

lemma funas_term_take_drop:
  assumes "funas_term (Fun f ss) \<subseteq> F" "n \<le> length ss"
  shows "\<Union>(set (map funas_term (take n ss))) \<subseteq> F" "\<Union>(set (map funas_term (drop n ss))) \<subseteq> F"
  using assms set_take_subset apply fastforce 
  by (metis (no_types, lifting) Sup_le_iff assms(1) drop_map funas_term.simps(2) in_set_dropD le_sup_iff) 

lemma subst_seq: 
  assumes "u \<cdot> \<sigma> = l \<cdot> \<rho>" "funas_term u \<subseteq> F" "funas_term l \<subseteq> F" "(h, 0) \<notin>  F" "\<sigma> = (\<lambda> t. Fun h [])" "is_Fun u" "linear_term l"
  shows "\<exists> \<rho>'. u = l \<cdot> \<rho>'"
  using assms 
proof (induction l arbitrary: u)
  case (Var x)
  have "u = Var x \<cdot> (\<lambda> t. u)" by force
  then show ?case 
    by blast
next
  case (Fun g ts)
  have ts_subst:"Fun g ts \<cdot> \<rho> = Fun g (map (\<lambda> t. t \<cdot> \<rho>) ts)" by auto
  obtain f ss where f_ss:"u = Fun f ss" using Fun(7) by auto
  have lhs:"u \<cdot> \<sigma> = Fun f (map  (\<lambda> t. t \<cdot> \<sigma>) ss)"  using f_ss by force
  hence eq:"g = f \<and> length ss = length ts" using ts_subst Fun(2) f_ss 
    by (metis map_eq_imp_length_eq term.inject(2))
  {fix i
    assume i:"i < length ts"
    have ss_ts_i:"ss!i \<cdot> \<sigma> = ts!i \<cdot> \<rho>"  using eq lhs ts_subst
         by (metis Fun.prems(1) i map_nth_conv term.sel(4))
    hence "\<exists>\<rho>\<^sub>i. (ss!i) = (ts!i) \<cdot> \<rho>\<^sub>i"
    proof (cases "ss!i")
      case (Var y)
      hence "ss!i \<cdot> \<sigma> = Fun h []" using Fun(6) 
        by (simp add: Var)
      hence "ts!i \<cdot>\<rho> = Fun h []" using  ss_ts_i by simp
      hence conj:"is_Var (ts!i) \<or> (ts!i) = Fun h []" using const_subst by fast
      {assume asm: "ts!i = Fun h []"
        hence "funas_term (ts!i) = {(h,0)}" by auto
        moreover have "funas_term (ts!i) \<subseteq> F" using i Fun(4) by fastforce
        ultimately have "(h,0) \<in> F"  by simp
        hence False using Fun(5) by argo
      }note contr= this
      with conj obtain w where "ts!i = Var w" by blast
      then have "ss!i = (ts!i) \<cdot> (\<lambda> t. Var y)" using Var by fastforce
      then show ?thesis by fast
    next
      case (Fun f' ss')
       have "funas_term (ss!i) \<subseteq> F" using f_ss Fun.prems(2) i eq by fastforce
       moreover have "funas_term (ts!i) \<subseteq> F" using Fun.prems(3) i by force
       moreover have "is_Fun (ss!i)" using Fun by simp
       moreover have "linear_term (ts!i)" using Fun.prems(7) i by force
       ultimately show ?thesis using ss_ts_i i Fun.prems(4,5) Fun.IH[of "ts!i" "ss!i"]  
         by force
     qed}
   from this obtain r  where r:"(ss!i) = (ts!i) \<cdot> (r i)"  if "i < length ts" for i 
     by metis
   define rs where  "rs= map r  [0..<length ts]"
   then have rs:"\<forall>i < length ts.  (ss!i) = (ts!i) \<cdot> (rs!i) \<and> length rs = length ts"  using r by auto
   then obtain \<rho>' where " \<forall>i < length ts.   (ts!i) \<cdot> (rs!i) = (ts!i) \<cdot> \<rho>'" using Fun(8)  subst_unification_list[of ts rs]
     by auto
   hence "\<forall>i < length ts. (ss!i) = (ts!i) \<cdot> \<rho>'"  using rs by presburger
   hence "Fun f ss = (Fun f ts) \<cdot> \<rho>'" using f_ss 
     by (simp add: \<open>\<forall>i<length ts. ss ! i = ts ! i \<cdot> \<rho>'\<close> \<open>u = Fun f ss\<close> eq nth_equalityI)
  then show ?case using f_ss eq by blast 
qed

lemma subst_commute_rstep:
  fixes u v
  assumes "\<sigma> = (\<lambda> x. (Fun g []))" "funas_trs R \<subseteq> F" "(g, 0) \<notin> F" 
    "(v \<cdot> \<sigma>, u) \<in> rstep R" "llrg R" "funas_term v \<subseteq> F" "funas_term u \<subseteq> F \<union> {(g,0)}"
  shows "\<exists>u'. (v ,u') \<in> rstep R \<and> (u = u' \<cdot> \<sigma>) \<and> funas_term u' \<subseteq> F"
  using assms(4,1,2,3,5,6,7)
proof (induction v arbitrary: u)
  case (Var x)
  hence "(Fun g [], u) \<in> rstep R" by simp
  hence " (\<exists>u'. ((Var x, u') \<in> rstep R) \<and> (u = u' \<cdot> \<sigma>) \<and> funas_term u' \<subseteq> F)"
    using Var  
  proof (induction rule:rstepE)
    case (rstep s t C \<rho> l r)
    hence C_Hole:"C = Hole" using const_ctxt 
      by metis
    hence "u = r \<cdot> \<rho>" using rstep(2,5) by auto
    moreover have grnd:"ground r" using rstep(3) Var(5) unfolding llrg_def by auto
    ultimately have u_r:"u = r" 
      using ground_subst_apply by blast
    hence u_funas:"funas_term u \<subseteq> F" using rstep(3,8) unfolding funas_trs_def funas_rule_def snd_conv 
      by fastforce
    from C_Hole have "l \<cdot> \<rho> = Fun g []" 
      by (simp add: rstep.hyps(1,4))
    hence l:"is_Var l \<or> l = Fun g []" using const_subst by fast
    have "is_Var l" using l
    proof (cases "is_Var l")
      case False
      hence "l = Fun g []" using l by fast
      hence "(g,0) \<in> funas_trs R" using rstep(3) unfolding funas_trs_def funas_rule_def by fastforce
      then show ?thesis using rstep(8,9)
        by blast 
    qed (auto)
    then obtain y where "l = Var y" 
      by fast
    then have "(Var y, r) \<in> R" using u_r rstep(3) by simp
    then have "(Var y \<cdot>(\<lambda> y. Var x), r \<cdot> (\<lambda> y. Var x)) \<in> rstep R" by fast
    hence "(Var x, r) \<in> rstep R" using grnd 
      by (simp add: ground_subst_apply) 
    moreover have "u = r \<cdot> \<sigma>" using grnd u_r by(simp add:ground_subst_apply)
    ultimately show ?case using rstep u_funas grnd
      by (metis u_r)
  qed
  then show ?case by blast
next
  case (Fun f ss)
  show ?case using Fun(2)
  proof (induction rule:rstepE)
    case (rstep s t C \<rho> l r)
    hence grnd:"ground r" using Fun(6) unfolding llrg_def  
      by (metis snd_conv)
    define ss' where ss':"ss' = map (\<lambda>i. (i \<cdot>\<sigma>)) ss"
    hence ss'_Fun:"Fun f ss \<cdot> \<sigma> = Fun f ss'" by simp
    then show ?case
    proof (cases C)
      case Hole
      hence u:"u = r" using grnd rstep.hyps(2,5) 
        by (simp add: rstep.hyps(3) ground_subst_apply)
      have eq:"Fun f ss \<cdot> \<sigma>= l \<cdot> \<rho>" using rstep(1,4) Hole 
        by (metis ctxt.cop_nil)
      moreover have funas_l:"funas_term l \<subseteq> F" using Fun(4) rstep(3) unfolding funas_trs_def funas_rule_def 
        by force
      ultimately obtain \<rho>' where \<rho>': "Fun f ss = l \<cdot> \<rho>'" using subst_seq[OF eq Fun(7) funas_l Fun(5,3)] 
          rstep(3) Fun(6) unfolding llrg_def 
        by (metis fst_conv is_FunI)
      hence "(l \<cdot> \<rho>', r \<cdot> \<rho>') \<in> rstep R" using rstep(3) 
        by fast
      hence "(Fun f ss, r) \<in> rstep R" using grnd \<rho>' 
        by (simp add: \<rho>' ground_subst_apply)
      moreover have "r \<cdot> \<sigma> = u" using u grnd 
        by (simp add: u ground_subst_apply)
      moreover have "funas_term r \<subseteq> F" using rstep(3) Fun(4)  
        unfolding funas_trs_def funas_rule_def 
        by force
      ultimately show ?thesis by blast  
    next
      case (More h ss1 C' ss3)
      hence "Fun f ss' = Fun h (ss1@(C'\<langle>l \<cdot> \<rho>\<rangle>)#ss3)" using ss'_Fun rstep by auto
      hence h_f:"h = f \<and> ss' = (ss1@(C'\<langle>l \<cdot> \<rho>\<rangle>)#ss3)" by simp
      hence "ss'!(length ss1) = (C'\<langle>l \<cdot> \<rho>\<rangle>)" by simp
      hence ss_i:"ss!(length ss1) \<cdot> \<sigma> = (C'\<langle>l \<cdot> \<rho>\<rangle>)" using ss'
        by (smt \<open>h = f \<and> ss' = ss1 @ C'\<langle>l \<cdot> \<rho>\<rangle> # ss3\<close> length_map map_eq_Cons_D map_eq_append_split nth_append_length) 
      with h_f have u:"u = Fun f (ss1@(C'\<langle>r \<cdot> \<rho>\<rangle>)#ss3)" using More
        by (simp add: \<open>ss ! length ss1 \<cdot> \<sigma> = C'\<langle>l \<cdot> \<rho>\<rangle>\<close> rstep.hyps(2,5))
      have rstep_ind:"(ss!(length ss1) \<cdot> \<sigma>, C'\<langle>r \<cdot> \<rho>\<rangle>) \<in> rstep R" using ss_i rstep(3)  
        by blast
      moreover have len:"length ss1 < length ss" using h_f
      proof -
        have f1: "\<forall>ts. ss1 @ (ss' ! length ss1 # ss3) @ ts = ss' @ ts"
          using h_f by auto
        have f2: "length ss = length ss'"
          using ss' by auto
        have "length ss1 \<noteq> length ss'"
          using f1 by (metis (full_types) append_Cons append_Cons_nth_middle ctxt_apply_eq_False)
        then show ?thesis
          using f2 f1 by (metis (no_types) append_Cons append_Cons_nth_left append_Cons_nth_middle ctxt_apply_eq_False nat_neq_iff)
      qed
      hence "length ss1 < length ss'" using ss' by simp
      moreover have "funas_term (ss!length ss1) \<subseteq> F"  "funas_term (C'\<langle>r \<cdot> \<rho>\<rangle>) \<subseteq> F \<union> {(g,0)}"
        using len Fun(7) funas_subterm[of f ss]  apply simp using len Fun(7,8) funas_subterm[of f "(ss1@(C'\<langle>r \<cdot> \<rho>\<rangle>)#ss3)"]  
        by (simp add: More rstep.hyps(2,5))          
      ultimately obtain u' where  u':"(ss!(length ss1), u') \<in> rstep R \<and>( C'\<langle>r \<cdot> \<rho>\<rangle> = u' \<cdot> \<sigma>) \<and> funas_term u' \<subseteq> F" using rstep_ind Fun(1,3-6) 
        using \<open>length ss1 < length ss\<close> 
        by fastforce
      moreover have breakup:"(take (length ss1) ss)@((ss!(length ss1))#(drop (length ss1+1) ss)) = ss" using h_f
          ss'
        using \<open>length ss1 < length ss\<close> id_take_nth_drop by fastforce
      hence ss1_ss3:"map (\<lambda> i. i \<cdot> \<sigma>) (take (length ss1) ss) = ss1" "map (\<lambda> i. i \<cdot> \<sigma>) (drop (length ss1+1) ss) = ss3" using ss' h_f 
         apply (simp add: h_f append_eq_conv_conj take_map) using ss' h_f id_take_nth_drop \<open>length ss1 < length ss\<close> 
        by (metis (no_types, lifting) One_nat_def \<open>length ss1 < length ss'\<close> add.right_neutral add_Suc_right append_eq_conv_conj drop_map list.inject)
      let ?ss1 = "(take (length ss1) ss)" 
      let ?ss3 = "(drop (length ss1+1) ss)"
      let ?u' = "?ss1@(u'#?ss3)"
      have "(Fun f ss, Fun f ?u') \<in> rstep R" using u' breakup Fun_ctxt_closed[of "ss ! length ss1" u' R f ?ss1 ?ss3]
        by presburger
      moreover have "(Fun f (?u')) \<cdot> \<sigma> = u" using ss1_ss3 u' u by auto
      moreover have "funas_term (Fun f ?u') \<subseteq> F" using breakup Fun(7)
      proof -
        have "\<Union>(set (map funas_term ?u')) = (\<Union> set (map funas_term ?ss1) \<union> (funas_term u') \<union> (\<Union> set (map funas_term ?ss3)))"
          by force
        moreover have "\<Union> set (map funas_term ?ss1) \<subseteq> F" using funas_term_take_drop[OF Fun(7)] len by simp
        moreover have "(funas_term u') \<subseteq> F" using u' by simp
        moreover have "\<Union> set (map funas_term ?ss3) \<subseteq> F" using funas_term_take_drop[OF Fun(7)] len by simp
        ultimately have "\<Union>(set (map funas_term ?u')) \<subseteq> F" by blast
        moreover have "(f, length ss) \<in> F" using Fun(7) by simp 

        ultimately show ?thesis
          by (smt Fun.prems(6) breakup funas_term.simps(2) le_sup_iff length_append list.size(4))   
      qed
      ultimately show ?thesis 
        by blast
    qed  
  qed 
qed

lemma subst_commute_sig_step:
  fixes u v
  assumes "\<sigma> = (\<lambda> x. (Fun g []))" "funas_trs R \<subseteq> F" "(g, 0) \<notin> F" 
    "(v \<cdot> \<sigma>, u) \<in> sig_step (F\<union>{(g,0)}) (rstep R)" "llrg R" "funas_term v \<subseteq> F"
  shows "\<exists>u'. (v ,u') \<in> sig_step F (rstep R) \<and> (u = u' \<cdot> \<sigma>)"
  using subst_commute_rstep[OF assms(1-3)] assms unfolding sig_step_def by blast

lemma subst_commute_sig_step_rtrancl:
  fixes u v
  assumes "\<sigma> = (\<lambda> x. (Fun g []))" "funas_trs R \<subseteq> F" "(g, 0) \<notin> F" 
    "(v \<cdot> \<sigma>, u) \<in> (sig_step (F\<union>{(g,0)}) (rstep R))^*" "llrg R" "funas_term v \<subseteq> F"
  shows "\<exists>u'. (v ,u') \<in> (sig_step F (rstep R))^* \<and> (u = u' \<cdot> \<sigma>)"
  using assms(4,1,2,3,5,6)
proof (induction rule:rtrancl_induct)
  case (step y z)
  then obtain u\<^sub>1 where u\<^sub>1:"(v, u\<^sub>1) \<in> (sig_step F (rstep R))\<^sup>*" "y = u\<^sub>1 \<cdot> \<sigma>" by blast 
  hence tran:"(u\<^sub>1 \<cdot> \<sigma>, z) \<in> (sig_step (F\<union>{(g,0)}) (rstep R))" 
    using step.hyps(2) by blast
  have funas_u\<^sub>1:"funas_term u\<^sub>1 \<subseteq> F" 
    using u\<^sub>1(1) step(8) 
    unfolding sig_step_def apply (induction rule:rtrancl_induct) by simp+
  show ?case using subst_commute_sig_step[OF step(4,5,6) tran step(7) funas_u\<^sub>1] 
  proof -
    show ?thesis
      by (meson  u\<^sub>1(1) \<open>\<exists>u'. (u\<^sub>1, u') \<in> sig_step F (rstep R) \<and> z = u' \<cdot> \<sigma>\<close> rtrancl.rtrancl_into_rtrancl)
  qed
qed (fast)

lemma varposs_subterm_extrc:
  assumes "p \<in> varposs (Fun f ts)"
  shows "\<exists> i q. i < length ts \<and>  p = i # q \<and>  q \<in> varposs(ts!i)" 
  using assms by auto

lemma var_tracing:
  assumes "(s,t) \<in> rstep R" "llrg R"  "p \<in> varposs t"
  shows "t |_ p = s |_ p \<and> p \<in> varposs s"
  using assms
proof (induction rule:rstepE)
  case (rstep s t C \<sigma> l r)
  show ?case using rstep(3-) unfolding rstep(1,2)
  proof (induction C arbitrary:s t p)
    case Hole
    hence "t = r \<cdot> \<sigma>"  by force
    moreover have grnd:"ground r" using rstep(3,6) Hole(1) unfolding llrg_def by fastforce
    ultimately have "t = r"  by (simp add: ground_subst_apply) 
    with grnd have "varposs t = {}" 
      by (metis Hole.prems(5) empty_iff ground_vars_term_empty subt_at_imp_supteq subteq_Var_imp_in_vars_term varposs_iff)
    then show ?case using Hole(5) by blast 
  next
    case (More f ss1 C' ss2)
    define ss where ss: "ss= (ss1@(C'\<langle>l \<cdot> \<sigma>\<rangle>)#ss2)" 
    define ts where ts: "ts =  (ss1@(C'\<langle>r \<cdot> \<sigma>\<rangle>)#ss2)" 
    with More ss have  s_t:"t = Fun f ts" "s = Fun f ss"
      by (meson ctxt_apply_term.simps(2))+
    from  More(6) s_t(1) obtain i q where i_q: "i < length ts \<and> p = i # q \<and>  q \<in> varposs(ts!i)" 
      using varposs_subterm_extrc[of p f ts] by blast
    then show ?case
    proof (cases  "i = length (ss1)")
      case True
      hence ss_ts_i:"ss!i = (C'\<langle>l \<cdot> \<sigma>\<rangle>)" "ts!i = (C'\<langle>r \<cdot> \<sigma>\<rangle>)" using ss ts by simp+
      hence "s|_ p = (C'\<langle>l \<cdot> \<sigma>\<rangle>)|_q" "t|_p  = (C'\<langle>r \<cdot> \<sigma>\<rangle>)|_ q" using s_t i_q 
        apply auto[1] 
        by (simp add: \<open>ts ! i = C'\<langle>r \<cdot> \<sigma>\<rangle>\<close> i_q s_t(1))  
      with More(1)[of "C'\<langle>l \<cdot> \<sigma>\<rangle>" "C'\<langle>r \<cdot> \<sigma>\<rangle>"  q] More(5) i_q ss_ts_i(2) 
      show ?thesis
      proof -
        have "q \<in> varposs C'\<langle>r \<cdot> \<sigma>\<rangle>"
          using \<open>ts ! i = C'\<langle>r \<cdot> \<sigma>\<rangle>\<close> i_q by presburger
        then show ?thesis
          by (metis \<open>\<lbrakk>(l, r) \<in> R; C'\<langle>l \<cdot> \<sigma>\<rangle> = C'\<langle>l \<cdot> \<sigma>\<rangle>; C'\<langle>r \<cdot> \<sigma>\<rangle> = C'\<langle>r \<cdot> \<sigma>\<rangle>; llrg R; q \<in> varposs C'\<langle>r \<cdot> \<sigma>\<rangle>\<rbrakk> \<Longrightarrow> C'\<langle>r \<cdot> \<sigma>\<rangle> |_ q = C'\<langle>l \<cdot> \<sigma>\<rangle> |_ q \<and> q \<in> varposs C'\<langle>l \<cdot> \<sigma>\<rangle>\<close> \<open>s |_ p = C'\<langle>l \<cdot> \<sigma>\<rangle> |_ q\<close> \<open>t |_ p = C'\<langle>r \<cdot> \<sigma>\<rangle> |_ q\<close> i_q length_append list.size(4) poss_Cons_poss rstep.hyps(3) rstep.prems(1) s_t(2) ss ss_ts_i(1) term.sel(4) ts varposs_iff)
      qed
    next
      case False
      hence "ts!i = ss!i" using ss ts i_q 
        by (simp add: append_Cons_nth_not_middle)
      moreover have "t|_ (i#q) = (ts!i)|_q" using s_t(1) i_q by simp
      moreover have "s|_ (i#q) = (ss!i)|_q" using s_t(2) i_q by auto
      ultimately have "t|_ (i#q) = s|_ (i#q)" by argo
      then show ?thesis using i_q 
        by (metis \<open>t |_ (i # q) = ts ! i |_ q\<close> \<open>ts ! i = ss ! i\<close> length_append list.size(4) poss_Cons_poss s_t(2) ss term.sel(4) ts varposs_iff) 
    qed 
  qed
qed

lemma var_tracing_rtrancl:
  assumes "(s,t) \<in> (rstep R)^*" "llrg R" "p \<in> varposs t"
  shows "t |_ p = s |_ p \<and> p \<in> varposs s"
  using assms 
proof (induction rule:rtrancl_induct)
  case (step y z)
  have 1:"z |_ p = y |_ p \<and> p \<in> varposs y" using var_tracing[OF step(2,4,5)] .
  hence "y |_ p = s |_ p \<and> p \<in> varposs s" using step(3)[OF step(4)]  by simp 
  thus ?case using 1 by argo
qed (simp)

lemma var_tracing_sig_step:
  assumes "(s,t) \<in> sig_step F (rstep R)" "llrg R" "p \<in> varposs t"
  shows "t |_ p = s |_ p \<and> p \<in> varposs s" using assms var_tracing unfolding sig_step_def by blast

lemma var_tracing_sig_step_rtrancl:
  assumes "(s,t) \<in> (sig_step F (rstep R))^*" "llrg R" "p \<in> varposs t"
  shows "t |_ p = s |_ p \<and> p \<in> varposs s" using var_tracing_rtrancl var_tracing_sig_step assms 
  unfolding sig_step_def 
  by (smt assms(1) rtranclD set_rev_mp sig_stepE trancl_into_rtrancl trancl_sig_subset)

lemma in_eq_poss:
  assumes "s \<noteq> t"
  shows "\<exists>p. s|_ p \<noteq> t |_ p \<and> p \<in> poss s \<inter> poss t"
  using assms
  by (metis IntI empty_pos_in_poss subt_at.simps(1)) 

lemma in_eq_poss_funposs_eq:
  assumes "s \<noteq> t" "funposs s = funposs t" "\<forall>p \<in> funposs s. s |_ p = t |_ p" 
  shows "\<exists>p. s|_ p \<noteq> t |_ p \<and> p \<in> varposs s \<inter> varposs t"
  using in_eq_poss[OF assms(1)] assms(2) 
proof - 
  obtain p where p:" s|_ p \<noteq> t |_ p \<and> p \<in> poss (s) \<inter> poss (t)" using in_eq_poss[OF assms(1)]
    by blast
  hence or:"p \<in> funposs s \<or> p \<in> varposs s"  "p \<in> funposs t \<or> p \<in> varposs t"  using poss_simps apply blast 
    by (metis Diff_Diff_Int Diff_cancel Diff_iff Int_Diff Int_empty_right Un_iff assms(2) empty_iff funposs_mctxt_mctxt_of_term p poss_simps(2) set_funposs_list) 
  {assume asm:"p \<in> funposs s"
   hence False using assms(3) p 
      by blast}
  hence 1:"p \<in> varposs s" using or by satx
  {assume asm:"p \<in> funposs t"
    hence False using assms(2,3) p by blast}
  with 1 p show ?thesis
    using or(2) by blast 
qed

lemma funposs_subterm_extrc:
  assumes "p \<in> funposs (Fun f ts)"
  shows "p = [] \<or> (\<exists> i q. i < length ts \<and>  p = i # q \<and>  q \<in> funposs (ts ! i))" 
  using assms funposs.simps(2)[of f ts] apply simp by blast

lemma funposs_empty:
  assumes "p \<in> funposs (Fun f ts)"
  shows "p = [] \<Longrightarrow> (Fun f ts) |_ p = Fun f ts"
  using assms by auto

lemma funposs_subterm_extrc2:
  assumes "p \<in> funposs (Fun f ts)" "p \<noteq> []"
  shows "(\<exists> i q. i < length ts \<and>  (p = i # q) \<and>  q \<in> funposs(ts!i))" 
  using assms funposs.simps(2)[of f ts] apply simp by blast

lemma varposs_distinct:
  fixes u :: "('f, 'v) term" and g
  defines "\<sigma> \<equiv> \<lambda>y. (Fun g [] :: ('f, 'v) term)"
  assumes "u \<cdot> \<sigma> = v \<cdot> \<sigma>" "funas_term u \<subseteq> F" "funas_term v \<subseteq> F" "(g, 0) \<notin> F" "u \<noteq> v"
  shows "(\<exists>p \<in> varposs u \<inter> varposs v. u |_ p \<noteq> v |_ p)" using assms
proof (induction u arbitrary: v)
  case (Var x)
   have "Var x \<cdot> \<sigma> = Fun g []" by (simp add: \<sigma>_def)
  with Var(2) have "v \<cdot> \<sigma> = Fun g []" by (force simp: \<sigma>_def)
  hence v:"v = Fun g [] \<or> is_Var v" by fastforce
  {
    assume asm:"v = Fun g []"
    hence "(g,0) \<in> funas_term v" by auto 
    hence "(g,0) \<in>  F" using Var(4) by fast
    hence False using Var(5) by simp
  }
  with v obtain y where y:"v = Var y" by fast
  hence "y \<noteq> x" using Var(6) by simp
  hence "(Var x) |_ [] \<noteq> (Var y) |_ []" by simp
  thus ?case using y by fastforce
next
  case (Fun f ss)
  have Fun_eq:"(Fun f (map (\<lambda> t. t \<cdot> \<sigma>) ss))  = v \<cdot>  \<sigma> " using Fun(3) unfolding subst_apply_term.simps(2) .
  then obtain ts where v:"v = Fun f ts" 
    by (smt Collect_mem_eq Fun.hyps(2) Fun.prems(2) UnCI adapt_vars_simps assms(5) 
            funas_term.elims funas_term.simps(2) insert_iff length_map list.map_disc_iff 
            list.simps(8) list.size(3) subsetCE subst_apply_term.simps(1) subst_apply_term.simps(2) 
            subst_apply_term_empty term.distinct(1) term.inject(2) term.sel(4))
  have list_eq:"(map (\<lambda> t. t \<cdot> \<sigma>) ss)  = (map (\<lambda> t. t \<cdot> \<sigma>) ts)" using Fun_eq unfolding v unfolding subst_apply_term.simps(2) 
    by force
  hence len_eq:"length ss = length ts" 
    using map_eq_conv' by blast
  from Fun(7) v obtain i where i:"i < length ss" "(ss!i) \<noteq> (ts!i)" 
    using len_eq nth_equalityI by blast
  with list_eq len_eq have "(ss!i) \<cdot> \<sigma> = (ts!i) \<cdot> \<sigma>"  
    by (meson map_eq_conv')
  moreover have "funas_term (ss!i) \<subseteq> F" "funas_term (ts!i) \<subseteq> F" using i len_eq Fun(4,5) unfolding v
    by force+
  ultimately obtain q where q:"q \<in> varposs(ss!i) \<inter> varposs(ts!i) \<and> (ss!i) |_ q \<noteq> (ts!i) |_ q"
    using Fun(1)[of "ss!i" "ts!i"] using i Fun(2,6) by fastforce
  hence "i#q \<in> varposs (Fun f ss)" "i#q \<in> varposs (Fun f ts)"using i len_eq by simp+
  moreover have "(Fun f ss) |_ (i#q) = (ss!i) |_ q " using i(1) q by auto
  moreover have "(Fun f ts) |_ (i#q) = (ts!i) |_ q" using i(1) len_eq q by auto
  ultimately show ?case using q unfolding v 
    by (metis IntI) 
qed

lemma funas_extn:
  fixes g :: 'f and s :: "('f,'v) term"
  assumes "(funas_term s) \<subseteq> F" "(g,0) \<notin> F"
  shows "funas_term (s \<cdot> (\<lambda> y. Fun g [])) \<subseteq> (F \<union> {(g,0)})"
  using assms by (induction s)  (simp, fastforce)

lemma sig_step_rtrancl:
  assumes "(s,t) \<in> (sig_step (F) (rstep R))^*"
  shows "(s,t) \<in> (rstep R)^*"
  using assms apply (induction rule:rtrancl_induct) apply simp unfolding sig_step_def 
  by auto

lemma sig_step_rtrancl_conv:
  fixes g :: 'f
  defines \<sigma>: "\<sigma> \<equiv> \<lambda>y. (Fun g [] :: ('f, 'v) term)"
  assumes "(s :: ('f,'v) term,t) \<in> (sig_step (F) (rstep R))^*" "(g, 0) \<notin> F"
  shows "(s \<cdot> \<sigma> ,t \<cdot> \<sigma>) \<in> (sig_step (F \<union> {(g,0)}) (rstep R))^*"
  using assms(2,1,3)
proof (induction rule:rtrancl_induct)
  case (step u v)
  hence 1:"(s \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> (sig_step (F \<union> {(g, 0)}) (rstep R))\<^sup>*" by blast
  have u_v:"(u,v) \<in> rstep R" "funas_term u \<subseteq> F" "funas_term v \<subseteq> F"
    using step(2) unfolding sig_step_def by force+
  hence "(u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> (rstep R)" using rsteps_closed_subst by auto 
  hence "(u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> (sig_step (F \<union> {(g, 0)}) (rstep R))"
    unfolding sig_step_def using funas_extn[OF u_v(2) step(5)] funas_extn[OF u_v(3) step(5)] \<sigma>
    by fast
  with 1 show ?case by simp
qed (blast)

lemma funas_rtrancl:
  assumes "funas_term s \<subseteq> F" "(s, t) \<in> (sig_step F (rstep R))\<^sup>*"
  shows "funas_term t \<subseteq> F"
  using assms(2,1) 
  by (induction rule:rtrancl_induct) (auto)

lemma GCR_implies_CR:
  fixes R :: "('f,'v) trs" 
  assumes "(g,0) \<notin> F"
    "CR_on (sig_step (F \<union> {(g,0)}) (rstep R)) {s. ground s \<and> funas_term s \<subseteq> F \<union> {(g,0)}}"
    "llrg R" "funas_trs R \<subseteq> F"
  shows "CR (sig_step F (rstep R))" 
proof -
  define \<sigma> where \<sigma>: "\<sigma> = (\<lambda> (y::'v) . (Fun g [] :: ('f, 'v) term))"
  define r where r: "r = (sig_step (F) (rstep R))" 
  define r' where r': "r' = (sig_step (F\<union>{(g,0)}) (rstep R))"
  {fix s::"('f,'v) term"
    fix u::"('f,'v) term" 
    fix v::"('f,'v) term"
    assume asm: "(s, u) \<in> r\<^sup>* \<and> (s, v) \<in> r\<^sup>*"
    hence neq_funas:"(s \<noteq> u) \<or> (s \<noteq> v) \<Longrightarrow> (funas_term s \<subseteq> F) \<and> (funas_term u \<subseteq> F) \<and> (funas_term v \<subseteq> F)"
      unfolding r sig_step_def rtrancl_def 
      by (metis (no_types, lifting) asm converse_rtranclE r rtrancl.simps sig_stepE)
    from asm have rstep_s: "(s, u) \<in> (rstep R)^*" "(s, v) \<in> (rstep R)^*" 
      unfolding r using sig_step_rtrancl by auto 
    have "(u,v) \<in> join r"
    proof (cases "(s = u) \<and> (s = v)")
      case False
      hence "(s \<noteq> u) \<or> (s \<noteq> v)" by blast
      hence funas_subs:
        "(funas_term s \<subseteq> F)" "(funas_term u \<subseteq> F)"  "(funas_term v \<subseteq> F)" 
        using neq_funas by auto
      have ground:"ground (s \<cdot> \<sigma>)" "ground (u \<cdot> \<sigma>)" "ground (v \<cdot> \<sigma>)" unfolding \<sigma> by fastforce+
      have funas_rel:
        "funas_term (s \<cdot> \<sigma>) \<subseteq> F \<union> {(g,0)}"
        "funas_term (u \<cdot> \<sigma>) \<subseteq> F \<union> {(g,0)}"
        "funas_term (v \<cdot> \<sigma>) \<subseteq> F \<union> {(g,0)}"
        unfolding \<sigma> using funas_extn[OF funas_subs(1) assms(1)]
          funas_extn[OF funas_subs(2) assms(1)]
          funas_extn[OF funas_subs(3) assms(1)].  
      have meet_gcr:"(s \<cdot> \<sigma>, u \<cdot> \<sigma>) \<in> (r')\<^sup>*"  "(s \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> (r')\<^sup>*" 
        unfolding r' using asm unfolding r using assms(1) unfolding \<sigma>
        by (intro sig_step_rtrancl_conv; simp)+
      hence "(u \<cdot> \<sigma>, v \<cdot> \<sigma>) \<in> join r'"
        using assms(2) funas_rel ground  unfolding r' CR_on_def by blast
      then obtain t' where t':"(u \<cdot> \<sigma>, t') \<in> r'^* " "(v \<cdot> \<sigma>, t') \<in> r'^*"
        by blast
      then obtain t\<^sub>1 where t1:"(u, t\<^sub>1) \<in> (sig_step F (rstep R))\<^sup>*" "(t' = t\<^sub>1 \<cdot> \<sigma>)"
        using subst_commute_sig_step_rtrancl[OF \<sigma> assms(4,1)] t'(1) assms(3) 
        unfolding r'
        using funas_subs(2) by blast 
      moreover  obtain t\<^sub>2 where t2:"(v, t\<^sub>2) \<in> (sig_step F (rstep R))\<^sup>*" "(t' = t\<^sub>2 \<cdot> \<sigma>)"
        using  subst_commute_sig_step_rtrancl[OF \<sigma> assms(4,1)] t'(2) assms(3) 
        unfolding r' 
        using funas_subs(3) by blast  
      ultimately have eq:"t\<^sub>1 \<cdot> \<sigma> = t\<^sub>2 \<cdot> \<sigma>" by auto
      have "t\<^sub>1 = t\<^sub>2"
      proof (rule ccontr)
        assume ineq_assm:"t\<^sub>1 \<noteq> t\<^sub>2"
        then obtain p where  p:"p \<in> varposs(t\<^sub>1) \<inter> varposs(t\<^sub>2)" "t\<^sub>1 |_ p \<noteq> t\<^sub>2 |_ p"   
          using varposs_distinct[of t\<^sub>1 g t\<^sub>2 F] eq funas_rtrancl[OF funas_subs(2) t1(1)]  
            funas_rtrancl[OF funas_subs(3) t2(1)] assms(1) 
          unfolding \<sigma> by fast
        have t1_p:"t\<^sub>1 |_ p = u |_ p \<and> p \<in> varposs(u)" 
          using var_tracing_sig_step_rtrancl[OF t1(1) assms(3)] p(1) by blast
        moreover have u_s:"u |_p  = s |_p \<and> p \<in> varposs(s)"
          using var_tracing_sig_step_rtrancl[of s u F R p] asm t1_p assms(3) unfolding r by blast
        ultimately have 1:"t\<^sub>1 |_ p = s |_ p" by argo
        have t2_p:"t\<^sub>2 |_ p = v |_ p \<and> p \<in> varposs(v)" 
          using var_tracing_sig_step_rtrancl[OF t2(1) assms(3)] p(1) by blast 
        moreover have u_s:"v |_p  = s |_p \<and> p \<in> varposs(s)"
          using var_tracing_sig_step_rtrancl[of s v F R p] asm t2_p assms(3) unfolding r by blast
        ultimately have 2:"t\<^sub>2 |_ p = s |_ p" by argo
        with p(2) 1 show False by argo
      qed
      with t1(1) t2(1) show ?thesis using r joinI
        by (simp add: \<open>t\<^sub>1 = t\<^sub>2\<close>)
    qed (auto)}
  thus ?thesis unfolding CR_defs using r by blast
qed

lemma sig_step_invert:
  fixes R :: "('f,'v) trs" and g :: 'f and y
  defines "\<sigma> \<equiv> \<lambda> y. (Fun g []::('f, 'v) term)"
  assumes "(g,0) \<notin> F"
    "funas_trs R \<subseteq> F"
    "(s, t) \<in> (sig_step (F\<union>{(g,0)}) (rstep R))"
    "ground s" "ground t"
  shows "((inv_const (g,0) y s) \<cdot> \<sigma> = s) \<and> ((inv_const (g,0) y t) \<cdot> \<sigma> = t) 
            \<and> (inv_const (g,0) y s, inv_const (g,0) y t) \<in> (sig_step F (rstep R))" 
proof -
  let ?s = "inv_const (g,0) y s"
  let ?t = "inv_const (g,0) y t"
  have s'_t'_rstep:"(s,t) \<in> rstep R" using assms(4) unfolding sig_step_def by simp
  hence s_t_rstep:"(?s,?t) \<in> rstep R" using inv_const_rstep[OF assms(3) "s'_t'_rstep" assms(2)] 
    by (simp)
  have const_cond:"s = ?s \<cdot> \<sigma>" "t = ?t \<cdot> \<sigma>" 
    using inv_cons_sigma[OF assms(5)] inv_cons_sigma[OF assms(6)] unfolding assms(1) by presburger+
  have funas_cond:
    "funas_term s \<subseteq> F \<union> {(g,0)}" "funas_term t \<subseteq> F \<union> {(g,0)}" using assms(4) unfolding sig_step_def
    by simp+
  hence "funas_term ?s \<subseteq> F" "funas_term ?t \<subseteq> F"   
    using inv_const_filter_funas[of s F g] inv_const_filter_funas[of t F g] assms(2) by simp+
  with const_cond s_t_rstep show ?thesis unfolding sig_step_def 
    by (metis (no_types, lifting) case_prodI mem_Collect_eq) 
qed

lemma rstep_ground_llrg:
  assumes "llrg R" "ground s" "(s,t) \<in> rstep R"
  shows "ground t"
  using assms  empty_iff ground_vars_term_empty llrg_def prod.sel(2) rstep_ground subsetI 
  by (metis)

lemma rstep_ground_llrg_rtrancl:
  assumes "llrg R" "ground s" "(s,t) \<in> (rstep R)^*"
  shows "ground t"
  using assms(3,2,1)
  apply (induction rule:rtrancl_induct)
  apply simp
  using rstep_ground_llrg by fast

lemma sig_step_invert_rtrancl:
  fixes R :: "('f,'v) trs" and g :: 'f
  defines "\<sigma> \<equiv> \<lambda>_. Fun g [] :: ('f, 'v) term"
  assumes "(g,0) \<notin> F"
    "funas_trs R \<subseteq> F"
    "(s, t) \<in> (sig_step (F\<union>{(g,0)}) (rstep R))^*"
    "llrg R"
    "ground s \<and> ground t"
  shows "((inv_const (g,0) y s) \<cdot> \<sigma> = s) \<and> ((inv_const (g,0) y t) \<cdot> \<sigma> = t) 
        \<and> (inv_const (g,0) y s, inv_const (g,0) y t) \<in> (sig_step F (rstep R))^*" 
  using assms(4,1,2,3,5,6)
proof (induction rule: rtrancl_induct)
  case base
  have "inv_const (g,0) y s \<cdot> \<sigma> = s"  "(inv_const (g,0) y s, inv_const (g,0) y s) \<in> (sig_step F (rstep R))\<^sup>*"
    using base(5) by (simp_all add: base(1) ground_subst_apply inv_cons_sigma) 
  then show ?case by fast
next
  case (step u t)
  let ?ic = "inv_const (g,0) y"
  have "(s, u) \<in> (rstep R)^*" using step(1) unfolding sig_step_def 
    using sig_step_rtrancl step.hyps(1) by blast
  hence grnd:"ground u" using rstep_ground_llrg_rtrancl  step(7,8)  
    by meson
  hence int:"((?ic s) \<cdot> \<sigma> = s)" "((?ic u) \<cdot> \<sigma> = u)" "((?ic s, ?ic u) \<in> (sig_step F (rstep R))\<^sup>*)"
    using step by blast+
  moreover have "((?ic t) \<cdot> \<sigma> = t)" "((?ic u, ?ic t) \<in> (sig_step F (rstep R))\<^sup>*)"
    using sig_step_invert[OF step(5,6,2) grnd]  step(8) unfolding  step(4) by fast+
  ultimately show ?case by force
qed

lemma join_sig_extn:
  assumes "(u, v) \<in> join (sig_step F (rstep R))"
  shows "(u \<cdot> (\<lambda>_. Fun g []), v \<cdot> (\<lambda>_. Fun g [])) \<in> join (sig_step (F \<union> {(g,0)}) (rstep R))"
proof -
  obtain t where "(u, t) \<in> (sig_step (F) (rstep R))^*"
                 "(v, t) \<in> (sig_step (F) (rstep R))^*"
    using assms by auto
  hence "(u \<cdot> (\<lambda> t. Fun g []), t \<cdot> (\<lambda> t. Fun g [])) \<in> (sig_step (F \<union> {(g,0)}) (rstep R))^*"
        "(v \<cdot> (\<lambda> t. Fun g []), t \<cdot> (\<lambda> t. Fun g [])) \<in> (sig_step (F \<union> {(g,0)}) (rstep R))^*"
    using rstep_trancl_sig_extn[of _ _ F] by blast+
  thus ?thesis by auto
qed

lemma CR_implies_GCR:
  fixes R :: "('f,'v) trs" 
  assumes "(g,0) \<notin> F"
    "CR (sig_step F (rstep R))"
    "llrg R" "funas_trs R \<subseteq> F"
  shows "CR_on (sig_step (F \<union> {(g,0)}) (rstep R)) {s. ground s \<and> funas_term s \<subseteq> F \<union> {(g,0)}}"
proof -
  fix y
  let ?ic = "inv_const (g,0) y :: ('f, 'v) term \<Rightarrow> _"
  define S where S:"S = {(s::('f,'v) term). ground s \<and> funas_term s \<subseteq> (F \<union> {(g,0)})}"
  { fix s u v
    assume asm1: "(s,u) \<in> (sig_step (F\<union>{(g,0)}) (rstep R))^*"  "(s,v) \<in> (sig_step (F\<union>{(g,0)}) (rstep R))^*"
    assume asm2: "s \<in> S" 
    have grnd:"ground u" "ground v"   using rstep_ground_llrg_rtrancl[OF assms(3)] asm1 unfolding sig_step_def
          asm2 unfolding S 
      by (metis (no_types, lifting) S asm1 asm2 sig_step_rtrancl mem_Collect_eq)+
    hence "(?ic s, ?ic u) \<in> (sig_step F (rstep R))^*"  "(?ic u \<cdot> (\<lambda>y. Fun g [])) = u"
          "(?ic s, ?ic v) \<in> (sig_step F (rstep R))^*"  "(?ic v \<cdot> (\<lambda>y. Fun g [])) = v"
      using asm1 asm2 sig_step_invert_rtrancl assms 
      by (simp add: S sig_step_invert_rtrancl)+
    hence "(?ic u, ?ic v) \<in> join (sig_step (F) (rstep R))"
      using assms(2) by blast
    hence "(?ic u \<cdot> (\<lambda> t. Fun g []), ?ic v \<cdot> (\<lambda> t. Fun g [])) \<in> join  (sig_step (F\<union>{(g,0)}) (rstep R))"  
     using join_sig_extn[of _ _ F] by blast
    hence "(u, v) \<in> join  (sig_step (F\<union>{(g,0)}) (rstep R))" using grnd by (simp add: inv_cons_sigma) }
  thus ?thesis unfolding S by auto
qed

lemma CR_GCR_equiv:
  fixes R :: "('f,'v) trs" 
  assumes "llrg R" "funas_trs R \<subseteq> F" "(g,0) \<notin> F"
  shows "CR (sig_step F (rstep R)) = CR_on (sig_step (F \<union> {(g,0)}) (rstep R)) {s. ground s \<and> funas_term s \<subseteq> F \<union> {(g,0)}}"
  using CR_implies_GCR[of g F] GCR_implies_CR[of g F] assms by blast+

end