theory LV_to_GTT_Impl
  imports LV_to_GTT GTT_TRS_Impl TRS.Term_Impl TA.Tree_Automata_Autoref_Setup Refine_Iterator
begin

(* TODO: move; a variation of this is reproved in GTT_RRn_Impl *)
lemma map_of_keys:
  "map_of (map (\<lambda>x. (x, ())) xs) x = (if x \<in> set xs then Some () else None)"
  by (induct xs) auto

lemma trs_to_ta_extra_simps:
  "\<And>f g P. f ` {g x |x. P x} = {f (g x) |x. P x}"
  "\<And>f g P. f ` {g x y |x y. P x y} = {f (g x y) |x y. P x y}"
  "\<And> f X. {f x y |x y. (x, y) \<in> X} = case_prod f ` X"
  "\<And> f X. {f x |x. x \<in> X} = f ` X"
  (* "\<And>f y g. f (let x = y in g x) = (let x = y in f (g x))" *)
  by auto

definition (in fixed_signature) term_TA' ::
  "(('f,'v) term \<Rightarrow> 'b) \<Rightarrow> ('f,'v) term \<Rightarrow> ('b, 'f) ta_rule set" where
  "term_TA' h t = ({(f (map h ts) \<rightarrow> h (Fun f ts))|f ts. Fun f ts \<unlhd> t})"

fun term_TA'_list :: "(('f,'v) term \<Rightarrow> 'b) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('b, 'f) ta_rule list" where
  "term_TA'_list h (Fun f ts) = (TA_rule f (map h ts) (h (Fun f ts))) # concat (map (term_TA'_list h) ts)"
| "term_TA'_list h (Var x) = []"

lemma term_TA'_list:
  shows "fixed_signature.term_TA' h t = set (term_TA'_list h t)"
proof (induct t)
  have *: "s \<rhd> t \<longleftrightarrow> (\<exists>f ss s'. s = Fun f ss \<and> s' \<in> set ss \<and> s' \<unrhd> t)" for f ss t s
    by (metis supt_Fun_imp_arg_supteq set_supteq_into_supt Var_supt term.exhaust)
  case (Fun f ts)
  have *: "fixed_signature.term_TA' h (Fun f ts) = {f (map h ts) \<rightarrow> h (Fun f ts)} \<union> \<Union>((\<lambda>t. fixed_signature.term_TA' h t) ` set ts)"
    unfolding fixed_signature.term_TA'_def
    apply (subst subterm.order.order_iff_strict)
    apply (simp only: * singleton_iff term.simps conj_assoc ex_simps simp_thms cong: rev_conj_cong)
    by auto
  show ?case by (simp add: * Fun)
qed (auto 0 1 simp: fixed_signature.term_TA'_def dest: supteq_Var_id)

definition term_TA'_impl ::
  "(('f, 'v) term \<Rightarrow> 'b :: compare_order) \<Rightarrow> ('f :: compare_order,'v) term \<Rightarrow> (('b, 'f) ta_rule, unit) RBT_Impl.rbt" where
  "term_TA'_impl h t = rbt_bulkload (map (\<lambda>k. (k, ())) (term_TA'_list h t))"

lemma term_TA'_impl:
  assumes "FORCE_id R" "FORCE_id Q"
  shows "(term_TA'_impl, fixed_signature.term_TA') \<in>
   (\<langle>Id :: _ :: compare_order rel,Id :: _ :: compare_order rel\<rangle>term_rel \<rightarrow> R) \<rightarrow> \<langle>Id,Id\<rangle>term_rel \<rightarrow> \<langle>\<langle>Q,Id\<rangle>ta_rule_rel\<rangle>comp_rs_rel"
  using assms
  apply (refine_vcg)
  subgoal for h' h t' t
    by (auto intro!: relcompI[of _ "\<lambda>x. if x \<in> set (term_TA'_list h t) then Some () else None"]
      simp: map_of_keys compare_order_class.ord_defs autoref_post_simps rbt_lookup_rbt_bulkload
      term_TA'_impl_def term_TA'_list map2set_rel_def rbt_map_rel_def rbt_map_rel'_def
      split: if_splits intro!: brI)
  done

lemma ta_make_autoref':
  "FORCE_id R \<Longrightarrow>
   (ta_make_code, ta.make)
    \<in> \<langle>Id\<rangle>comp_rs_rel \<rightarrow> \<langle>\<langle>Id, Id\<rangle>ta_rule_rel\<rangle>comp_rs_rel \<rightarrow> \<langle>R \<times>\<^sub>r R\<rangle>list_set_rel \<rightarrow> dflt_ta_rel"
  using ta_make_autoref by simp

definition Inl_rule where
  "Inl_rule \<equiv> \<lambda>(f :: 'f) n. f (replicate n (Inl (Var ()))) \<rightarrow> Inl (Var ())"

schematic_goal trs_to_ta_L_impl:
  notes [autoref_rules] = term_TA'_impl FORCE_id[of _ fixed_signature.embed_term] FORCE_id[of _ replicate]
    FORCE_id[of _ Inl_rule] ta_make_autoref'
  shows "(?f, fixed_signature.trs_to_ta_L) \<in> \<langle>Id \<times>\<^sub>r Id\<rangle>comp_rs_rel \<rightarrow> \<langle>\<langle>Id :: 'f :: compare_order rel,Id :: 'q :: compare_order rel\<rangle>term_rel \<times>\<^sub>r \<langle>Id,Id\<rangle>term_rel\<rangle>comp_rs_rel \<rightarrow> dflt_ta_rel"
  unfolding fixed_signature.trs_to_ta_L_def fixed_signature.L_trs_ta_set_def
  unfolding trs_to_ta_extra_simps(1,2) fixed_signature.ta_rules_Inl_eq
  unfolding
    fixed_signature.term_to_ta_L_def fixed_signature.term_to_ta_def fmap_states_ta_def
  unfolding trs_to_ta_extra_simps ta.simps ta.make_def
  unfolding ta.make_def[symmetric] fixed_signature.term_TA'_def[of Inl, symmetric] Inl_rule_def[symmetric]
  by (autoref)

concrete_definition trs_to_ta_L_impl uses trs_to_ta_L_impl

definition Inr_rule where
  "Inr_rule \<equiv> \<lambda>(f :: 'f) n. f (replicate n (Inr (Var ()))) \<rightarrow> Inr (Var ())"

schematic_goal trs_to_ta_R_impl:
  notes [autoref_rules] = term_TA'_impl FORCE_id[of _ fixed_signature.embed_term] FORCE_id[of _ replicate]
    FORCE_id[of _ Inr_rule] ta_make_autoref'
  shows "(?f, fixed_signature.trs_to_ta_R) \<in> \<langle>Id \<times>\<^sub>r Id\<rangle>comp_rs_rel \<rightarrow> \<langle>\<langle>Id :: 'f :: compare_order rel,Id :: 'q :: compare_order rel\<rangle>term_rel \<times>\<^sub>r \<langle>Id,Id\<rangle>term_rel\<rangle>comp_rs_rel \<rightarrow> dflt_ta_rel"
  unfolding fixed_signature.trs_to_ta_R_def fixed_signature.R_trs_ta_set_def
  unfolding trs_to_ta_extra_simps(1,2) fixed_signature.ta_rules_Inr_eq
  unfolding
    fixed_signature.term_to_ta_L_def fixed_signature.term_to_ta_def fmap_states_ta_def
  unfolding trs_to_ta_extra_simps ta.simps ta.make_def
  unfolding ta.make_def[symmetric] fixed_signature.term_TA'_def[of Inr, symmetric] Inr_rule_def[symmetric]
  by (autoref)

concrete_definition trs_to_ta_R_impl uses trs_to_ta_R_impl

declare fixed_signature.embed_term_def [code]

definition lv_trs_to_gtt_impl :: "_ \<Rightarrow> (('f :: compare_order, 'v :: compare_order) term \<times> _, unit) RBT_Impl.rbt \<Rightarrow> _" where
  "lv_trs_to_gtt_impl F R = (trs_to_ta_L_impl F R, trs_to_ta_R_impl F R)"

lemma prod_binary_function:
  assumes "(f', f) \<in> P \<rightarrow> Q \<rightarrow> Rf"
  assumes "(g', g) \<in> P \<rightarrow> Q \<rightarrow> Rg"
  shows "(\<lambda>x y. (f' x y, g' x y), \<lambda>x y. (f x y, g x y)) \<in> P \<rightarrow> Q \<rightarrow> Rf \<times>\<^sub>r Rg"
  using assms by (auto intro!: fun_relI prod_relI dest: fun_relD)

lemmas trs_to_gtt_autoref [autoref_rules] =
  prod_binary_function[OF trs_to_ta_L_impl.refine trs_to_ta_R_impl.refine, folded fixed_signature.trs_to_gtt_def lv_trs_to_gtt_impl_def]

(* export_code lv_trs_to_gtt_impl in Haskell *)

end