theory Monadic_Confluence
imports Ground_Confluence Ord.Signature_Extension LV_to_GTT LLRG_Confluence
begin

locale monadic_signature =
  fixes \<F> :: "'f sig"
  assumes monadic_sig: "snd ` \<F> \<subseteq> {0,1}"
     and ex_constant:"\<exists>c \<in> \<F>. snd c = 0"
begin

definition monadic :: "('f, 'v) term  \<Rightarrow> bool" where
  "monadic s \<equiv> (funas_term s \<subseteq> \<F>) "

lemma monadic_subterm: 
  assumes "monadic s" "t \<unlhd> s"
  shows "monadic t" unfolding monadic_def 
  by (meson assms(1) assms(2) dual_order.trans monadic_def supteq_imp_funas_term_subset)

lemma monadic_vars:
  assumes "monadic s"
  shows "card (vars_term s) \<le> 1"
  using assms
proof (induction s)
  case (Fun f ss)
  hence 1:"snd ` (funas_term (Fun f ss)) \<subseteq> {0,1}" unfolding monadic_def using monadic_sig 
    by (meson order.trans subset_image_iff)
  hence 2:"length ss = 0 \<or> length ss = 1" by auto
  then show ?case
  proof (cases "length ss = 0")
    case False
    with 2 have "length ss = 1" by simp
    then obtain s where s:"ss = [s]" 
      by (metis One_nat_def add.right_neutral add_Suc_right le_add_same_cancel2 
                le_imp_less_Suc length_0_conv less_not_refl list.size(4) remdups_adj.cases zero_le)
    hence "vars_term (Fun f ss) = vars_term s" by simp
    moreover have "monadic s" using monadic_subterm[of "Fun f ss" "s"] 
      using Fun.prems s by auto
    ultimately show ?thesis using Fun(1) 
      by (simp add: \<open>vars_term (Fun f ss) = vars_term s\<close> s)  
  qed (auto)
qed (auto)

lemma grnd_ctxt_vars_ctxt:
  "ground_ctxt C \<longleftrightarrow>  vars_ctxt C = {}"
proof (rule iffI, goal_cases)
  case 1 then show ?case
  proof (induction C)
    case (More f ss C ts) 
    have eq:"vars_ctxt (More  f ss C ts) = ((\<Union>s\<in>set ss. vars_term s) \<union> vars_ctxt C \<union> (\<Union>t\<in>set ts. vars_term t))"
      by auto
    have "\<forall>s \<in> set ss. vars_term s ={}" "\<forall>t \<in> set ts. vars_term t ={}"
      using More(2) 
      by (simp add: ground_vars_term_empty)+
    then show ?case using More eq by force 
  qed (simp)
next
  case 2 then show ?case
  proof (induction C)
    case (More f ss C ts)
    hence "\<forall>s \<in> set ss. vars_term s = {}" "\<forall>t \<in> set ts. vars_term t = {}" by force+
    hence "\<forall>s \<in> set ss. ground s" "\<forall>t \<in> set ts. ground t" 
      by (simp add: ground_vars_term_empty)+
    then show ?case using More by auto 
  qed (auto)  
qed

lemma ground_ctxt_monadic:
  assumes "monadic C\<langle>s\<rangle>" "monadic s"
  shows "ground_ctxt C"
  using assms  
proof (induction C)
  case (More f ss C ts)
  hence 1:"snd ` (funas_term (Fun f (ss@C\<langle>s\<rangle>#ts))) \<subseteq> {0,1}" unfolding monadic_def using monadic_sig 
    by (metis ctxt_apply_term.simps(2) doubleton_eq_iff order.trans subset_image_iff)
  hence 2:"(length (ss@C\<langle>s\<rangle>#ts) = 0) \<or>  (length (ss@C\<langle>s\<rangle>#ts) = 1)" by force 
  moreover have C:"ground_ctxt C" using More monadic_subterm by force
  show ?case 
  proof (cases "(length (ss@C\<langle>s\<rangle>#ts) = 0)")
    case False
     hence "(ss@C\<langle>s\<rangle>#ts) = [C\<langle>s\<rangle>]" using 2
       by simp
     hence "More f ss C ts = More f [] C []" using False 
       by (metis Cons_eq_append_conv Nil_is_append_conv list.discI list.inject)
     with C show ?thesis by force
  qed (fast) 
qed (simp)

lemma ground_monadic_eq:
  assumes "ground s" "funas_term s \<subseteq> \<F>" 
  shows "monadic s"
  using assms unfolding monadic_def by simp

lemma ground_ctxt_apply:
  assumes "ground_ctxt C" "ground s"
  shows "ground C\<langle>s\<rangle>" using assms 
  by (simp add: grnd_ctxt_vars_ctxt ground_vars_term_empty vars_term_ctxt_apply)

lemma monadic_subst:
  fixes g :: "'f"
  defines "\<sigma> \<equiv> \<lambda> t. Fun g []"
  assumes "(g,0) \<in> \<F>" "monadic s" 
  shows "funas_term (s \<cdot> \<sigma>) \<subseteq> \<F>"
  using assms 
proof (induction s)
  case (Fun f ss)
  have 1:"Fun f ss \<cdot> \<sigma> = Fun f (map (\<lambda> t. (t \<cdot> \<sigma>)) ss)" by auto  
  {fix i
    assume i:"i < length ss"
    hence "monadic (ss!i)" using Fun(4) monadic_subterm by fastforce
    hence "funas_term ((ss!i)\<cdot> \<sigma>) \<subseteq> \<F>" using Fun(1,2,3) i 
      using monadic_def 
      by auto} note ind = this
  hence int:"\<Union>(set (map funas_term (map (\<lambda> t. (t \<cdot> \<sigma>)) ss))) \<subseteq> \<F>" 
    by (smt Union_least in_set_idx length_map map_nth_eq_conv)
  have "(f, length ss) \<in> \<F>" using monadic_def Fun(4) by force
  hence "(f, length (map (\<lambda> t. t \<cdot> \<sigma>) ss)) \<in> \<F>" by auto 
  thus ?case  using int 1 by simp
qed (simp) 

lemma property_1:
  fixes g :: 'f
  defines "\<sigma>' \<equiv> \<lambda>_. Fun g []"
  assumes "(s,t) \<in> sig_step \<F> (rstep R)" "monadic s" "monadic t" "llrg R" "(g,0) \<in> \<F>"
  shows "(s \<cdot> \<sigma>', t) \<in> sig_step \<F> (rstep R) \<and> (ground t)"       
  using assms 
proof -
  have asm1:"(s,t) \<in> rstep R" "funas_term s \<subseteq> \<F>" "funas_term t \<subseteq> \<F>"  using assms(2) 
    by blast+
  thus ?thesis using assms(3-6,1)
  proof (induction rule:rstepE)
    case (rstep s t C \<sigma> l r)
    hence "r \<cdot> \<sigma> = r" unfolding llrg_def using ground_subst_apply[of r \<sigma>] 
      by (metis ground_subst_apply snd_conv)
    hence t:"t = C\<langle>r\<rangle>" using rstep(5) by argo
    hence "monadic r" using rstep(2,9) monadic_subterm[of t r] by blast
    with t have "ground_ctxt C" using rstep(2,9,10) ground_ctxt_monadic by blast 
    moreover have "ground r" using rstep(3,10) unfolding llrg_def by auto
    ultimately have grnd:"ground t" using t ground_ctxt_apply by blast
    hence "(s \<cdot> \<sigma>', t \<cdot> \<sigma>') \<in>   (rstep R)" using rstep asm1 by blast
    hence step:"(s \<cdot> \<sigma>', t) \<in> (rstep R)" using ground_subst_apply[OF grnd] by metis
    moreover have "funas_term (s \<cdot> \<sigma>') \<subseteq> \<F>" using monadic_subst rstep by blast
    hence "(s \<cdot> \<sigma>', t) \<in> sig_step \<F> (rstep R)" using rstep(2,7) step by blast  
    then show ?case using grnd rstep(1,2) by fast  
  qed
qed

lemma property_1_trancl:
  fixes g :: 'f
  defines "\<sigma>' \<equiv> \<lambda> t. Fun g []"
  assumes "(s,t) \<in> (sig_step \<F> (rstep R))^+" "monadic s" "monadic t" "llrg R" "(g,0) \<in> \<F>"
  shows "(s \<cdot> \<sigma>', t) \<in> (sig_step \<F> (rstep R))^+ \<and> ground t"
  using assms(2,1,3-6)
proof (induction rule:trancl_induct)
  case (base y)
  then show ?case using property_1 by blast  
next
  case (step u v)
  from step(2) have u:"monadic u" unfolding monadic_def by fast
  with step(3,4,5,7,8) have "(s \<cdot> \<sigma>', u) \<in> (sig_step \<F> (rstep R))\<^sup>+" by blast
  hence "(s \<cdot> \<sigma>', v) \<in> (sig_step \<F> (rstep R))\<^sup>+" using step(2) by simp 
  moreover have "ground v" using  property_1[OF step(2) u step(6,7,8)] by force        
  ultimately show ?case by satx
qed

lemma CR_implies_GCR:
  fixes R::"('f,'v) trs" 
  assumes 
    "CR (sig_step \<F> (rstep R))"
    "llrg R" "funas_trs R \<subseteq> \<F>"
  shows "CR_on (sig_step \<F> (rstep R)) {s. ground s}"
  using assms 
proof -
  {fix s u v
    assume "ground s" 
            "(s,u) \<in> (sig_step \<F> (rstep R))^*"
            "(s,v) \<in> (sig_step \<F> (rstep R))^*"
    hence "(u,v) \<in> join (sig_step \<F> (rstep R))"
      using assms(1) by auto}
  then show ?thesis by auto
qed

lemma GCR_implies_CR:
  fixes R :: "('f,'v) trs" 
  assumes 
    "CR_on (sig_step \<F> (rstep R)) {s. ground s}"
    "llrg R" "funas_trs R \<subseteq> \<F>"
  shows "CR (sig_step \<F> (rstep R))"
proof -
  obtain c where c:"(c,0) \<in> \<F>" using ex_constant by force
  define \<sigma> where \<sigma>:"\<sigma> \<equiv> \<lambda> (t::'v). Fun c ([]::('f,'v) term list)"  
  {fix s u v
   assume asm:"(s, u) \<in> (sig_step \<F> (rstep R))\<^sup>*" "(s, v) \<in> (sig_step \<F> (rstep R))\<^sup>*"
   hence "(u,v) \<in> join (sig_step \<F> (rstep R))"
   proof (cases "s = u \<or> s = v")
     case True
     then show ?thesis 
       using asm(1) asm(2) by blast  
   next
     case False
     hence tr:"(s, u) \<in> (sig_step \<F> (rstep R))\<^sup>+" "(s, v) \<in> (sig_step \<F> (rstep R))\<^sup>+"
       by (meson asm rtranclD)+
     hence mon:"monadic s" "monadic u" "monadic v" unfolding monadic_def 
       apply (meson sig_stepE tranclD) 
       apply (meson sig_stepE tr(1) tranclD2)
       by (meson sig_stepE tr(2) tranclD2)
     have grnd:"ground (s\<cdot> \<sigma>)"  using \<sigma>  by (induction s) (simp , force) 
     have r1:"((s \<cdot> \<sigma>, u) \<in> (sig_step \<F> (rstep R))^+)" "(ground u)"
       using property_1_trancl[OF tr(1) mon(1,2) assms(2) c]  \<sigma> by blast+ 
     have r2:"((s \<cdot> \<sigma>, v) \<in> (sig_step \<F> (rstep R))^+)" "(ground v)"
       using property_1_trancl[OF tr(2) mon(1,3) assms(2) c]  \<sigma> by blast+ 
     thus ?thesis using assms(1) r1 grnd unfolding CR_on_def
       by (metis mem_Collect_eq trancl_into_rtrancl)
     
   qed}
  thus ?thesis by auto
qed

(*Statement of equivalence*)
lemma CR_GCR_eq:
  fixes R :: "('f,'v) trs" 
  assumes "llrg R" "funas_trs R \<subseteq> \<F>"
  shows "CR (sig_step \<F> (rstep R)) =  CR_on (sig_step \<F> (rstep R)) {s. ground s}"
  using GCR_implies_CR CR_implies_GCR assms by metis

end

end
