theory RRn_Complement
  imports RRn_Automata TA.Tree_Automata_Det
begin

subsection \<open>Determinization\<close>

text \<open>Determinization is already part of IsaFoR\<close>

lemma deterministic_automaton:
  "RRn_spec n A T \<Longrightarrow> RRn_spec n (ps_ta A) T"
  unfolding RRn_spec_def gta_lang_def ps_ta_lang[symmetric] unfolding gta_lang_def_sym .


subsection \<open>Difference\<close>

definition difference_ta :: "('p, 'f) ta \<Rightarrow> ('q, 'f) ta \<Rightarrow> ('p \<times> 'q option, 'f) ta" where
  "difference_ta A B = ta.make
    {(p, q). p \<in> ta_final A \<and> (q = None \<or> the q \<notin> ta_final B)}
    ({f (zip ps (map Some qs)) \<rightarrow> (p, Some q) |f ps p qs q.
       length ps = length qs \<and> f ps \<rightarrow> p \<in> ta_rules A \<and> f qs \<rightarrow> q \<in> ta_rules B} \<union>
     {f (zip ps qs) \<rightarrow> (p, None) |f ps p qs. length ps = length qs \<and> f ps \<rightarrow> p \<in> ta_rules A \<and>
       (None \<in> set qs \<or> (\<forall>q. f (map the qs) \<rightarrow> q \<notin> ta_rules B))})
    {((p, q), (p', q)) |p q p'. (p, p') \<in> ta_eps A}"

lemma gta_lang_difference_ta:
  assumes "ta_det B"
  shows "gta_lang (difference_ta A B) = gta_lang A - gta_lang B"
proof (intro set_eqI iffI, goal_cases lr rl)
  case (lr t)
  obtain p q where pq: "(p, q) \<in> ta_final (difference_ta A B)"
    "(p, q) \<in> ta_res (difference_ta A B) (term_of_gterm t)" using lr by (auto simp: gta_lang_def elim: ta_langE2)
  have "p \<in> ta_final A" "q = None \<or> the q \<notin> ta_final B" using pq(1) by (auto simp: difference_ta_def)
  moreover have "p \<in> ta_res A (term_of_gterm t) \<and> ta_res B (term_of_gterm t) = {q' |q'. Some q' = q}"
    using pq(2)
  proof (induct t arbitrary: p q rule: gterm.induct)
    case (GFun f ts)
    obtain qs p' q' where pq': "f qs \<rightarrow> (p', q') \<in> ta_rules (difference_ta A B)"
      "((p', q'), (p, q)) \<in> (ta_eps (difference_ta A B))\<^sup>*" "length qs = length ts"
      "\<And>i. i < length ts \<Longrightarrow> (fst (qs ! i), snd (qs ! i)) \<in> ta_res (difference_ta A B) (term_of_gterm (ts ! i))"
      using GFun(2) by auto
    have ps: "f (map fst qs) \<rightarrow> p' \<in> ta_rules A" using pq'(1) by (auto simp: difference_ta_def)
    have p': "(p', p) \<in> (ta_eps A)\<^sup>*" "q' = q" using pq'(2)
      by (induct rule: rtrancl_induct2) (auto simp: difference_ta_def)
    { fix qs' q'' assume "length qs' = length ts"
        "\<And>i. i < length ts \<Longrightarrow> Some (qs' ! i) = snd (qs ! i)"
      then have "qs' = map (the \<circ> snd) qs" using pq'(3)
        by (auto intro!: nth_equalityI simp: arg_cong[of "Some _" _ the, unfolded option.sel])
    } note * = this
    consider "None \<in> set (map snd qs)"
      | "None \<notin> set (map snd qs)" "q' = None"
      | "None \<notin> set (map snd qs)" "q' \<noteq> None" "f (map the (map snd qs)) \<rightarrow> the q' \<in> ta_rules B"
      using pq'(1) by (fastforce simp: difference_ta_def map_map[of _ Some] comp_def simp del: map_map)
    then show ?case
    proof (cases)
      case 1
      then obtain i where "i < length qs" "snd (qs ! i) = None" by (auto simp: set_conv_nth)
      moreover have "q' = None" using 1 pq'(1) by (auto simp: difference_ta_def zip_nth_conv)
      ultimately show ?thesis
        using GFun(1)[OF nth_mem pq'(4)] GFun(1)[OF nth_mem pq'(4), of i]
          assms(1)[unfolded ta_det_def, THEN conjunct1] ps pq'(3) p'
        by (auto intro!: exI[of _ "p'"] exI[of _ "map fst qs"])
    next
      case 2
      { fix xs ys zs assume "xs = zip ys zs" "length ys = length zs"
        then have "ys = map fst xs \<and> zs = map snd xs" by auto
      } note [dest!] = this
      have "f (map (the \<circ> snd) qs) \<rightarrow> q'' \<notin> ta_rules B" for q''
        using pq'(1) 2(1,2) by (auto simp: difference_ta_def zip_map_fst_snd)
      then show ?thesis
        using 2 GFun(1)[OF nth_mem pq'(4)] ps pq'(3) p' assms(1)[unfolded ta_det_def, THEN conjunct1]
        by (auto intro!: exI[of _ "p'"] exI[of _ "map fst qs"] dest!: *)
    next
      case 3
      moreover have "i < length qs \<Longrightarrow> Some (the (snd (qs ! i))) = snd (qs ! i)" for i
        using 3(1) by (cases "snd (qs ! i)") (force simp: set_conv_nth)+
      ultimately show ?thesis
        using GFun(1)[OF nth_mem pq'(4)] assms(1)[unfolded ta_det_def, THEN conjunct1] ps pq'(3) p'(1)
        by (auto intro!: exI[of _ "p'"] exI[of _ "map fst qs"] exI[of _ "map (the \<circ> snd) qs"]
          ta_detE[OF assms(1), of _ "term_of_gterm (GFun f ts)"] dest!: * simp: p'(2) image_def)
    qed
  qed
  ultimately show ?case
    by (auto simp: gta_lang_def image_def intro!: ta_langI2 bexI[of _ "term_of_gterm t"] elim!: ta_langE2)
next
  case (rl t)
  obtain p where p: "p \<in> ta_final A" "p \<in> ta_res A (term_of_gterm t)"
    using rl by (auto simp: gta_lang_def elim: ta_langE2)
  { fix q assume "ta_res B (term_of_gterm t) = {q' |q'. Some q' = q}"
    then have "(p, q) \<in> ta_res (difference_ta A B) (term_of_gterm t)" using p(2)
    proof (induct t arbitrary: p q)
      case (GFun f ts)
      obtain p' ps where p: "f ps \<rightarrow> p' \<in> ta_rules A" "(p', p) \<in> (ta_eps A)\<^sup>*" "length ps = length ts"
        "\<And>i. i < length ts \<Longrightarrow> ps ! i \<in> ta_res A (term_of_gterm (ts ! i))" using GFun(3) by auto
      { fix i assume i: "i < length ts"
        then have "\<exists>qi. ta_res B (term_of_gterm (ts ! i)) = {q' |q'. Some q' = qi}"
          using ta_detE[OF assms(1)] by blast }
      then obtain qs where qs: "ta_res B (term_of_gterm (ts ! i)) = {q' |q'. Some q' = qs i}"
        if "i < length ts" for i using that by metis
      then have qs': "(ps ! i, qs i) \<in> ta_res (difference_ta A B) (term_of_gterm (ts ! i))"
        if "i < length ts" for i using GFun(1)[OF nth_mem _ p(4), of i "qs i"] that by blast
      have "f (zip ps (map qs [0..<length ts])) \<rightarrow> (p', q) \<in> ta_rules (difference_ta A B)"
      proof (cases "\<exists>i < length ts. qs i = None")
        case True then show ?thesis using p(1,3) qs GFun(2)
          by (cases q) (auto 0 3 intro!: exI[of _ ps] exI[of _ "map qs [0..<length ts]"]
            simp: difference_ta_def image_def Ball_def)
      next
        case False define qs' where "qs' = the \<circ> qs"
        have "map qs [0..<length ts] = map (Some \<circ> qs') [0..<length ts]"
          using False by (auto simp: qs'_def)
        then show ?thesis using p(1,3) qs GFun(2) assms(1)[unfolded ta_det_def, THEN conjunct1]
          nth_equalityI[of _ "map qs' [0..<length ts]"]
          by (cases q) (auto dest!: spec[of _ "map qs' [0..<length ts]"]
            intro!: exI[of _ ps] exI[of _ "map qs [0..<length ts]"] exI[of _ "map qs' [0..<length ts]"]
            simp: difference_ta_def image_def Ball_def comp_def cong: list.map_cong_simp)
      qed
      moreover have "((p', q), (p, q)) \<in> (ta_eps (difference_ta A B))\<^sup>*" using p(2)
      proof (induct p rule: rtrancl_induct)
        case (step p'' p) then show ?case
          by (intro rtrancl_into_rtrancl[OF step(3)]) (auto simp: difference_ta_def)
      qed auto
      ultimately show ?case using qs' p(3,4)
        by (auto intro!: exI[of _ p'] exI[of _ q] exI[of _ "zip ps (map qs [0..<length ts])"])
    qed
  } note * = this
  consider "ta_res B (term_of_gterm t) = {}" | q' where "ta_res B (term_of_gterm t) = {q'}"
    using ta_detE[OF assms, of _ "term_of_gterm t"] by blast
  then show ?case
  proof (cases)
    case 1
    moreover have "(p, None) \<in> ta_final (difference_ta A B)" using p(1) by (auto simp: difference_ta_def)
    ultimately show ?thesis using *[of None]
      by (auto simp: gta_lang_def image_def intro!: bexI[of _ "term_of_gterm t"] ta_langI2)
  next
    case 2
    moreover have "q' \<notin> ta_final B" using 2 rl ta_langI2[of "term_of_gterm t" q' B]
      by (force simp: gta_lang_def)
    then have "(p, Some q') \<in> ta_final (difference_ta A B)" using p(1) by (auto simp: difference_ta_def)
    ultimately show ?thesis using *[of "Some q'"] p(1)
      by (auto simp: gta_lang_def image_def intro!: bexI[of _ "term_of_gterm t"] intro: ta_langI2)
  qed
qed

lemma difference_automaton:
  assumes "ta_det B" "RRn_spec n A T" "RRn_spec n B U"
  shows "RRn_spec n (difference_ta A B) (T - U)"
  using assms
  by (auto simp: RRn_spec_def gta_lang_difference_ta) (metis gdecode_gencode)

(* FORT uses a construction for the complement where the signature automaton, but without accepting
states, is adjoined to the given automaton before it's made deterministic, relying on the fact that
the procedure for determinizing automata constructs all reachable states. This would work for us as
well, based on the following two facts from IsaFoR. *)
thm ps_rules_complete ps_rules_sound


subsection \<open>Complement\<close>

text \<open>We can assemble the complement automaton from...\<close>
thm RR1_to_RRn_spec[OF term_automaton]
thm append_automaton
thm difference_automaton

end