theory Refine_Iterator
  imports Refine_Monadic.Refine_Monadic
begin

subsection \<open>Misc\<close>

abbreviation FORCE_id where
  "FORCE_id \<equiv> PREFER REL_FORCE_ID"

lemma PREFER_id:
  "PREFER_id R \<Longrightarrow> (x, x) \<in> R" by simp

lemma FORCE_id:
  "FORCE_id R \<Longrightarrow> (x, x) \<in> R" by simp

subsection \<open>Iterators\<close>

text \<open>Iterators enumerate a set, allowing duplicate elements. We define a corresponding relator.\<close>

definition it_set_rel_def': "it_set_rel R \<equiv> \<langle>R\<rangle>list_rel O br set (\<lambda>_. True)"

lemma it_set_rel_def [refine_rel_defs]:
  "\<langle>R\<rangle>it_set_rel \<equiv> \<langle>R\<rangle>list_rel O br set (\<lambda>_. True)"
  by (auto simp: relAPP_def it_set_rel_def')

lemma list_set_rel_sub_it_set_rel:
  "\<langle>R\<rangle>list_set_rel \<subseteq> \<langle>R\<rangle>it_set_rel"
  by (auto simp: br_def list_set_rel_def it_set_rel_def)

lemmas list_set_rel_imp_it_set_rel = subsetD[OF list_set_rel_sub_it_set_rel]

definition op_set_to_it where
  "op_set_to_it X = SPEC (\<lambda>xs. set xs = X)"

lemma op_set_to_it_ref [autoref_rules]:
  shows "(RETURN, op_set_to_it) \<in> \<langle>R\<rangle>it_set_rel \<rightarrow> \<langle>\<langle>R\<rangle>list_rel\<rangle>nres_rel"
  by (force simp: op_set_to_it_def nres_rel_def it_set_rel_def br_def intro!: RETURN_SPEC_refine)

lemma it_set_rel_Id_implies_finite [simp]:
  shows "(X', X) \<in> \<langle>Id\<rangle>it_set_rel \<Longrightarrow> finite X"
  by (auto simp: it_set_rel_def br_def)

(*
lemma list_set_rel_Id_implies_finite [simp]:
  shows "(X', X) \<in> \<langle>Id\<rangle>list_set_rel \<Longrightarrow> finite X"
  by (auto simp: list_set_rel_def br_def)

lemma list_set_rel_Id_implies_set_eq [simp]:
  shows "(X', X) \<in> \<langle>Id\<rangle>list_set_rel \<Longrightarrow> set X' = X"
  by (auto simp: list_set_rel_def br_def fun_rel_def)

lemma op_set_to_it_ref [autoref_rules]:
  shows "(RETURN, op_set_to_it) \<in> \<langle>R\<rangle>list_set_rel \<rightarrow> \<langle>\<langle>R\<rangle>list_rel\<rangle>nres_rel"
  by (force simp: op_set_to_it_def nres_rel_def list_set_rel_def intro!: RETURN_SPEC_refine)
*)

lemma union_it_set_rel:
  assumes "(A', A) \<in> \<langle>R\<rangle>it_set_rel"
  assumes "(B', B) \<in> \<langle>R\<rangle>it_set_rel"
  shows "(A' @ B', A \<union> B) \<in> \<langle>R\<rangle>it_set_rel"
  using param_append[THEN fun_relD, THEN fun_relD, of A' _ R B'] assms
  by (fastforce simp: it_set_rel_def br_def)

subsection \<open>Iteration\<close>

definition FOREACH_it :: "'b set \<Rightarrow> ('b \<Rightarrow> 'a \<Rightarrow> 'a nres) \<Rightarrow> 'a \<Rightarrow> 'a nres" where
  "FOREACH_it X f s = LIST_FOREACH (\<lambda>_ _. True) (op_set_to_it X) (\<lambda>_. True) f s"

lemma FOREACH_it_rule:
  fixes I :: "'a set \<Rightarrow> 'b \<Rightarrow> bool" and X :: "'a set" and s :: 'b
  assumes I0: "I X s"
  assumes I1: "\<And>x X' s'. finite X \<Longrightarrow> x \<in> X \<Longrightarrow> X' \<subseteq> X \<Longrightarrow> I (insert x X') s' \<Longrightarrow> f x s' \<le> SPEC (I X')"
  assumes F: "\<And>s. finite X \<Longrightarrow> I {} s \<Longrightarrow> \<Phi> s"
  shows "FOREACH_it X f s \<le> SPEC \<Phi>"
  unfolding FOREACH_it_def LIST_FOREACH_def FOREACH_cond_def FOREACH_body_def op_set_to_it_def
  apply (refine_vcg order.trans[OF WHILEIT_weaken WHILEIT_rule[where
    I = "\<lambda>(it, B'). \<exists>xs'. xs = xs' @ it \<and> I (set it) B'" and s = "(xs, _)" for xs]]
    wf_measure[of "\<lambda>(xs, _). length xs"])
  subgoal by simp
  subgoal using I0 by (auto simp: hd_def split: list.splits)
  subgoal for xs s it B' using I1[of "hd it" "set (tl it)" B'] by (force simp: hd_def split: list.splits)
  subgoal by (auto intro: F)
  done

end