theory TreeAutomataTRS
  imports TA.Tree_Automata TA.Exact_Tree_Automata_Completion
begin

inductive ta_trs_rule :: "('f + 'q, 'v) term \<times> ('f + 'q, 'v) term \<Rightarrow> bool" where
  eps [intro]: "ta_trs_rule (State q, State q')"
| rule [intro]: "ta_trs_rule (fqterm f qs, State q)"

inductive is_state :: "('f + 'q, 'v) term \<Rightarrow> bool" where
  state [intro]: "is_state (State q)"

fun states' :: "('f + 'q, 'v) term \<Rightarrow> 'q set" where
  "states' (Var x) = {}"
| "states' (State q) = { q }"
| "states' (Fun (Inr q) (t # ts)) = undefined"
| "states' (Fun (Inl f) ts) = (\<Union>i<length ts. states' (ts ! i))"

abbreviation "ta_rules' \<A> \<equiv> { f qs \<rightarrow> r | f qs r. (fqterm f qs, State r) \<in> \<A> }"
abbreviation "ta_eps' \<A> \<equiv> { (q, q') | q q'. (State q, State q') \<in> \<A>}"
abbreviation "ta_states' \<A> \<equiv> { q | l r q. (q \<in> states' r \<or> q \<in> states' l) \<and> (l, r) \<in> \<A> }"
abbreviation "is_ta \<A> \<equiv> Ball \<A> ta_trs_rule"
abbreviation "trs_to_ta (\<A> :: ('f + 'q, 'v) trs) Q\<^sub>f \<equiv> make Q\<^sub>f (ta_rules' \<A>) (ta_eps' \<A>)"

lemma rev_ta_eps':
  assumes "(q\<^sub>1, q\<^sub>2) \<in> ta_eps' \<A>"
  shows "(State q\<^sub>1, State q\<^sub>2) \<in> \<A>"
using assms by fast

definition ta_accepts_term :: "('f + 'q, 'v) trs \<Rightarrow> 'q set \<Rightarrow> ('f + 'q, 'v) term \<Rightarrow> bool" where
  "ta_accepts_term \<A> Q\<^sub>f t = (ground t \<and> (\<exists>q \<in> Q\<^sub>f. (t, State q) \<in> (rstep \<A>)\<^sup>*))"

lemma rule_in_ta:
  assumes "is_ta \<A>"
  shows "f qs \<rightarrow> p \<in> ta_rules (trs_to_ta \<A> Q\<^sub>f) \<longleftrightarrow> (fqterm f qs, State p) \<in> \<A>"
using assms unfolding make_def by auto

lemma rule_to_rstep:
  "(fqterm f qs, State p) \<in> \<A> \<Longrightarrow> (var2state (Fun f (map Var qs)), State p) \<in> rstep \<A>"
using rstep.rule[of "var2state (Fun f (map Var qs))" "State p"] by force

lemma ground_ta_trs:
  assumes "is_ta \<A>"
  shows "\<forall>(l, r) \<in> \<A>. vars_term r \<subseteq> vars_term l"
proof -
  { fix l r
    assume "(l, r) \<in> \<A>"
    hence "ta_trs_rule (l, r)" using assms by blast
    hence "vars_term r \<subseteq> vars_term l" by (auto elim: ta_trs_rule.cases)
  }
  thus ?thesis by blast
qed

lemma ta_step_ground:
  assumes "(t, Var x) \<in> (rstep \<A>)\<^sup>*" "ground t" "is_ta \<A>"
  shows False
using rsteps_ground[OF _ assms(2,1)] ground_ta_trs[OF assms(3)] by force

lemma ground_var2state[simp]: "ground (var2state t)"
by (induction t) auto

lemma ta_trs_stepE:
  assumes "(s, t) \<in> rstep \<A>" "is_ta \<A>"
  shows "\<exists>C l r. (l, r) \<in> \<A> \<and> ta_trs_rule (l, r) \<and> s = C\<langle>l\<rangle> \<and> t = C\<langle>r\<rangle> \<and> is_state r"
proof -
  obtain C \<sigma> l r where C_l_r: "(l, r) \<in> \<A>" "s = C\<langle>l \<cdot> \<sigma>\<rangle>" "t = C\<langle>r \<cdot> \<sigma>\<rangle>"
    using rstep_imp_C_s_r[OF assms(1)] by blast
  hence trs_rule: "ta_trs_rule (l, r)" using assms(2) by blast
  hence "l \<cdot> \<sigma> = l" "r \<cdot> \<sigma> = r" "is_state r"
    using ground_subst_apply by (auto elim: ta_trs_rule.cases)
  thus ?thesis using C_l_r trs_rule by auto
qed

lemma ta_states_trs_to_ta:
  fixes \<A> :: "('f + 'q, 'v) trs"
  assumes "is_ta \<A>"
  shows "ta_states' \<A> \<union> Q\<^sub>f = ta_states (trs_to_ta \<A> Q\<^sub>f)"
proof (rule equalityI; rule subsetI)
  fix q
  assume "q \<in> ta_states' \<A> \<union> Q\<^sub>f"
  then consider (Q\<^sub>f) "q \<in> Q\<^sub>f" | (in_rules) "q \<in> ta_states' \<A>" by blast
  thus "q \<in> ta_states (trs_to_ta \<A> Q\<^sub>f)"
  proof cases
    case Q\<^sub>f
    thus ?thesis unfolding ta_states_def make_def by fastforce
  next
    case in_rules
    then obtain l r where l_r: "q \<in> states' l \<or> q \<in> states' r" "(l, r) \<in> \<A>" by blast
    hence "ta_trs_rule (l, r)" using assms by blast
    thus ?thesis
    proof (cases rule: ta_trs_rule.cases)
      case (eps p p')
      hence "(p, p') \<in> ta_eps (trs_to_ta \<A> Q\<^sub>f)" using l_r unfolding make_def by simp
      thus ?thesis using l_r unfolding ta_states_def eps by force
    next
      case (rule f qs q')
      hence "f qs \<rightarrow> q' \<in> ta_rules (trs_to_ta \<A> Q\<^sub>f)" using l_r unfolding make_def by auto
      with l_r[unfolded rule] show ?thesis unfolding ta_states_def r_states_def by force
    qed
  qed
next
  fix q
  let ?\<A> = "trs_to_ta \<A> Q\<^sub>f"
  assume asm: "q \<in> ta_states ?\<A>"
  consider (rule) "q \<in> \<Union> (r_states ` ta_rules ?\<A>)" |
           (eps) "\<exists>q'. (q, q') \<in> ta_eps ?\<A> \<or> (q', q) \<in> ta_eps ?\<A>" |
           (final) "q \<in> ta_final ?\<A>" using asm unfolding ta_states_def by fast
  thus "q \<in> ta_states' \<A> \<union> Q\<^sub>f"
  proof cases
    case rule
    then obtain rule where rule: "rule \<in> ta_rules ?\<A>" "q \<in> r_states rule" by blast
    then obtain f qs q' where f_qs_q': "rule = f qs \<rightarrow> q'"
      using ta_rule.exhaust_sel by blast
    hence "q = q' \<or> q \<in> set qs" using rule(2) unfolding r_states_def by simp
    hence "q \<in> states' (fqterm f qs) \<or> q \<in> states' (State q')"
      by (elim disjE) (auto, metis in_set_conv_nth lessThan_iff)
    moreover have in_\<A>: "(fqterm f qs, State q') \<in> \<A>"
      using rule unfolding f_qs_q' make_def by fastforce
    ultimately show ?thesis by blast
  next
    case eps
    thus ?thesis unfolding make_def by force
  qed (auto simp: make_def)
qed

lemma eps_rules_rstep:
  assumes "(q', q) \<in> {(q, q'). (State q, State q') \<in> \<A>}\<^sup>*"
  shows "(State q', State q) \<in> (rstep \<A>)\<^sup>*"
using assms
proof (induction rule: rtrancl_induct)
  case (step y z)
  hence "(State y, State z) \<in> rstep \<A>" by fast
  thus ?case using step(3) by simp
qed blast

lemma steps_in_ta1:
  assumes "q \<in> ta_res (trs_to_ta \<A> Q\<^sub>f) s" "is_ta \<A>"
  shows "(var2state s, State q) \<in> (rstep \<A>)\<^sup>*"
using assms(1)
proof (induction "trs_to_ta \<A> Q\<^sub>f" s arbitrary: q rule: ta_res.induct)
  case (1 q')
  thus ?case unfolding make_def by (auto simp: eps_rules_rstep)
next
  case (2 f ts)
  let ?\<A> = "trs_to_ta \<A> Q\<^sub>f"
  obtain p qs where res: "f qs \<rightarrow> p \<in> ta_rules ?\<A>" "(p, q) \<in> ((ta_eps ?\<A>)\<^sup>*)"
    "length qs = length ts" "\<forall>i < length ts. qs ! i \<in> (map (ta_res ?\<A>) ts) ! i"
    using 2 by auto
  hence inner: "i < length ts \<longrightarrow> (var2state (ts ! i), State (qs ! i)) \<in> (rstep \<A>)\<^sup>*" for i
    using 2(1) by simp
  have "(fqterm f qs, State p) \<in> \<A>" using res(1) rule_in_ta[OF assms(2)] by fast
  hence "(var2state (Fun f (map Var qs)), State p) \<in> rstep \<A>" using rule_to_rstep by metis
  moreover have "(var2state (Fun f ts), var2state (Fun f (map Var qs))) \<in> (rstep \<A>)\<^sup>*"
    using inner args_rsteps_imp_rsteps[of "map var2state ts" "map (var2state \<circ> Var) qs" \<A>] res(3)
    by force
  moreover have "(State p, State q) \<in> (rstep \<A>)\<^sup>*"
    using eps_rules_rstep[of p q \<A>] res(2) unfolding make_def by simp
  ultimately show ?case by auto
qed

lemma state_steps:
  assumes "(State q, p) \<in> (rstep \<A>)\<^sup>*" "is_ta \<A>"
  shows "is_state p"
using assms(1)
proof (induction rule: rtrancl_induct)
  case (step y z)
  obtain C l r where C_l_r: "(l, r) \<in> \<A>" "ta_trs_rule (l, r)" "y = C\<langle>l\<rangle>" "z = C\<langle>r\<rangle>" "is_state r"
    using ta_trs_stepE[OF step(2) assms(2)] by blast
  obtain q' where "C\<langle>l\<rangle> = State q'" using step(3) C_l_r(2,3) by (auto elim: is_state.cases)
  hence "C = \<box>" using Fun_Nil_supt by fastforce
  thus ?case using C_l_r(4,5) by force
qed blast

fun ta_res' :: "('q,'f)ta \<Rightarrow> ('f,'q)term \<Rightarrow> ('f,'q)term set" where
  "ta_res' \<A> (Var q) = {p. \<exists>q'. p = Var q' \<and> (q, q') \<in> (ta_eps \<A>)\<^sup>*}"
| "ta_res' \<A> (Fun f ts) = {q. \<exists>q'. q = Var q' \<and> q' \<in> ta_res \<A> (Fun f ts)} \<union>
                           {t. \<exists>ss. t = Fun f ss \<and> length ss = length ts \<and>
                               (\<forall>i < length ts. ss ! i \<in> ta_res' \<A> (ts ! i))}"

lemma ta_res'_refl[simp]: "t \<in> ta_res' \<A> t"
by (induction t) fastforce+

lemma ta_res_res'_eq:
  fixes q \<A> l
  shows "Var q \<in> ta_res' \<A> l \<longleftrightarrow> q \<in> ta_res \<A> l"
by (induction "\<A>" "l" rule: ta_res.induct) fastforce+

lemma ta_res'_ta_res:
  assumes "t \<in> ta_res' \<A> s" "p \<in> ta_res \<A> t"
  shows "p \<in> ta_res \<A> s"
using assms
proof (induction arbitrary: p t rule: ta_res'.induct)
  case (2 \<A> f ts) show ?case using 2(2-)
  proof (induction t)
    case (Var q)
    show ?thesis using ta_res_ctxt[of q \<A> "Fun f ts" p \<box>] Var by simp
  next
    case (Fun g ss)
    have ss_props: "g = f" "length ss = length ts" "\<forall>i < length ts. ss ! i \<in> ta_res' \<A> (ts ! i)"
      using Fun(2) by (auto elim!: ta_res'.cases)
    obtain qs p' where qs_props: "(g qs \<rightarrow> p') \<in> ta_rules \<A> \<and>
          (p',p) \<in> (ta_eps \<A>)\<^sup>* \<and> length qs = length ss \<and>
          (\<forall>i < length ss. qs ! i \<in> (map (ta_res \<A>) ss) ! i)"
      using Fun(3) by auto
    { fix i
      assume asm: "i < length ts"
      hence ss_i_ts_i: "ss ! i \<in> ta_res' \<A> (ts ! i)" using ss_props by blast
      have "qs ! i \<in> ta_res \<A> (ts ! i)"
        using Fun(3) 2(1)[OF asm ss_i_ts_i, of "qs ! i"] asm ss_props(2) qs_props by force
    } note inner = this
    hence "\<forall>i<length ts. qs ! i \<in> ta_res \<A> (ts ! i)"
      using ss_props(1,2) by blast
    thus ?thesis using qs_props unfolding ss_props by auto
  qed
qed force

lemma rstep_var2state_persists:
  assumes "(var2state s, t) \<in> rstep \<A>"  "is_ta \<A>"
  shows "\<exists>t'. t = var2state t'"
proof -
  obtain C l r where step: "(l, r) \<in> \<A>" "ta_trs_rule (l, r)"
      "var2state s = C\<langle>l\<rangle>" "t = C\<langle>r\<rangle>" "is_state r"
    using ta_trs_stepE[OF assms] by blast
  show ?thesis using step(3,4)
  proof (induction C arbitrary: s t)
    case Hole
    thus ?case using step(5)
      by simp (metis is_state.cases var2state.simps(1))
  next
    case (More f ss1 C ss2)
    let ?i = "length ss1"
    obtain g ts where s_def: "f = Inl g" "s = Fun g ts" using More(2) by (cases s) auto
    have ss1_ss2: "map var2state (take ?i ts) = ss1" "map var2state (drop (Suc ?i) ts) = ss2"
      using More(2) unfolding s_def by (auto, metis append_eq_conv_conj take_map,
      metis One_nat_def Suc_eq_plus1 diff_add_inverse drop_all drop_append drop_map
      length_Cons lessI less_imp_le_nat list.inject nth_drop_0 self_append_conv2 zero_less_Suc)
    have lengths: "length ts = length (ss1 @ C\<langle>r\<rangle> # ss2)" using More(2) unfolding s_def
      by simp (metis add_Suc_right length_Cons length_append length_map)
    have "(map var2state ts) ! ?i = C\<langle>l\<rangle>"
      using More(2) unfolding s_def by simp
    hence ts_i: "var2state (ts ! ?i) = C\<langle>l\<rangle>"
      using nth_map[of ?i ts var2state] lengths by force
    obtain t\<^sub>i' where "(ss1 @ C\<langle>r\<rangle> # ss2) ! ?i = var2state t\<^sub>i'"
       using More(1)[OF ts_i, of "(ss1 @ C\<langle>r\<rangle> # ss2) ! ?i"] by auto
    hence "t = var2state (Fun g (take ?i ts @ t\<^sub>i' # drop (Suc ?i) ts))"
      using More(2,3) lengths ts_i ss1_ss2 unfolding s_def by auto
    thus ?case by blast
  qed
qed

lemma rsteps_var2state_persists:
  assumes "(var2state s, t) \<in> (rstep \<A>)\<^sup>*" "is_ta \<A>"
  shows "\<exists>t'. t = var2state t'"
using assms rstep_var2state_persists
by (induction rule: rtrancl_induct) blast+

lemma var2state_inj: "var2state s = var2state t \<Longrightarrow> s = t"
proof (induction s arbitrary: t)
  case (Var x)
  thus ?case by (cases t) auto
next
  case s: (Fun f ts) thus ?case
  proof (cases t)
    case t: (Fun g ss)
    show ?thesis using s unfolding t
      by auto (metis list.inj_map_strong)
  qed simp
qed

lemma state2var_var2state_id[simp]: "state2var (var2state t) = t"
by (induction t) (simp add: map_idI)+

lemma eps_steps:
  assumes "(State q, p) \<in> (rstep \<A>)\<^sup>*" "p = State q'" "is_ta \<A>"
  shows "q' \<in> ta_res (trs_to_ta \<A> Q\<^sub>f) (Var q)"
using assms(1,2)
proof (induction arbitrary: q' rule: rtrancl_induct)
  case (step y z)
  obtain p' where y_def: "y = State p'"
    using state_steps[OF step(1) assms(3)] by (auto elim: is_state.cases)
  hence ta_res_p': "p' \<in> ta_res (trs_to_ta \<A> Q\<^sub>f) (Var q)" using step(3) by blast
  obtain C l r where C_l_r: "(l, r) \<in> \<A>" "ta_trs_rule (l, r)" "y = C\<langle>l\<rangle>" "z = C\<langle>r\<rangle>" "is_state r"
    using ta_trs_stepE[OF step(2) assms(3)] by blast
  have "C = \<box>" using C_l_r(3) Fun_Nil_supt unfolding y_def by fastforce
  have "(p', q') \<in> (ta_eps (trs_to_ta \<A> Q\<^sub>f))\<^sup>*"
    using C_l_r(1-4) unfolding y_def \<open>C = \<box>\<close> step(4) make_def by fastforce
  thus ?case using C_l_r ta_res_p' unfolding y_def \<open>C = \<box>\<close> step(4) by simp
qed simp

lemma steps_in_ta1':
  assumes "t \<in> ta_res' (trs_to_ta \<A> Q\<^sub>f) s" "is_ta \<A>"
  shows  "(var2state s, var2state t) \<in> (rstep \<A>)\<^sup>*"
using assms(1)
proof (induction "trs_to_ta \<A> Q\<^sub>f" s arbitrary: t rule: ta_res'.induct)
  case (1 q')
  thus ?case unfolding make_def by (auto simp: eps_rules_rstep)
next
  case (2 f ts)
  let ?\<A> = "trs_to_ta \<A> Q\<^sub>f"
  consider (tVar) "\<exists>q'. t = Var q' \<and> q' \<in> ta_res ?\<A> (Fun f ts)" |
           (tFun) "\<exists>ss. t = Fun f ss \<and> length ss = length ts \<and>
                   (\<forall>i < length ts. ss ! i \<in> ta_res' ?\<A> (ts ! i))"
    using 2(2) by force
  thus ?case
  proof cases
    case tVar
    then obtain p q qs where res': "t = Var q" "f qs \<rightarrow> p \<in> ta_rules ?\<A>" "(p, q) \<in> ((ta_eps ?\<A>)\<^sup>*)"
        "length qs = length ts" "\<forall>i < length ts. qs ! i \<in> ta_res ?\<A> (ts ! i)"
      by auto
    hence inner: "i < length ts \<longrightarrow> (var2state (ts ! i), State (qs ! i)) \<in> (rstep \<A>)\<^sup>*" for i
      using 2(1)[of i "Var (qs ! i)"] unfolding ta_res_res'_eq by simp
    have "(fqterm f qs, State p) \<in> \<A>" using res'(2) rule_in_ta[OF assms(2)] by fast
    hence "(var2state (Fun f (map Var qs)), State p) \<in> rstep \<A>" using rule_to_rstep by metis
    moreover have "(var2state (Fun f ts), var2state (Fun f (map Var qs))) \<in> (rstep \<A>)\<^sup>*"
      using inner args_rsteps_imp_rsteps[of "map var2state ts" "map (var2state \<circ> Var) qs" \<A>] res'(4)
      by force
    moreover have "(State p, State q) \<in> (rstep \<A>)\<^sup>*"
      using eps_rules_rstep[of p q \<A>] res'(3) unfolding make_def by simp
    ultimately show ?thesis unfolding res'(1) by auto
  next
    case tFun
    then obtain ss where res': "t = Fun f ss" "length ss = length ts"
        "\<forall>i < length ts. ss ! i \<in> ta_res' ?\<A> (ts ! i)"
      by blast
    hence inner: "i < length ts \<longrightarrow> (var2state (ts ! i), var2state (ss ! i)) \<in> (rstep \<A>)\<^sup>*" for i
      using 2(1)[of i "ss ! i"] unfolding ta_res_res'_eq by simp
    thus ?thesis using res'(1,2) args_rsteps_imp_rsteps[of "map var2state ts" "map var2state ss"]
      by auto
  qed
qed

lemma step_in_ta2':
  assumes "(var2state s, t') \<in> rstep \<A>" "t' = var2state t" "is_ta \<A>"
  shows "t \<in> ta_res' (trs_to_ta \<A> Q\<^sub>f) s"
proof -
  obtain C l r where step: "(l, r) \<in> \<A>" "ta_trs_rule (l, r)"
      "var2state s = C\<langle>l\<rangle>" "t' = C\<langle>r\<rangle>" "is_state r"
    using ta_trs_stepE[OF assms(1,3)] by blast
  obtain p' where r: "r = State p'" using step(5) by (auto elim: is_state.cases)
  show ?thesis using step(3,4) unfolding assms(2)
  proof (induction C arbitrary: s t)
    case Hole
    consider (state) "is_state l" | (fqterm) "\<exists>f qs. l = fqterm f qs"
      using step(2) by (auto elim!: ta_trs_rule.cases)
    thus ?case
    proof cases
      case state
      then obtain p where l_def: "l = State p" by (auto elim: is_state.cases)
      hence "p' \<in> ta_res (trs_to_ta \<A> Q\<^sub>f) (Var p)"
        using step(1) eps_steps[OF _ _ assms(3), of p r p'] unfolding r by fast
      thus ?thesis using Hole var2state_inj
        unfolding l_def r ta_res_res'_eq[symmetric]
        by (metis (mono_tags, lifting) ctxt.cop_nil var2state.simps(1))
    next
      case fqterm
      then obtain f qs where l: "l = fqterm f qs" by blast
      have "p' \<in> ta_res (trs_to_ta \<A> Q\<^sub>f) (Fun f (map Var qs))"
        using step(1) unfolding l r rule_in_ta[OF assms(3), of f qs p' Q\<^sub>f, symmetric]
          ctxt_apply_term.simps(1) by auto
      moreover have "t = Var p'" "s = Fun f (map Var qs)"
        using Hole var2state_inj[of t "Var p'"] var2state_inj[of s "Fun f (map Var qs)"]
          map_nth_conv[of "var2state \<circ> Var" qs State qs] var2state.simps(1)[of "qs ! _"]
        unfolding l r by fastforce+
      ultimately show ?thesis using Hole
        unfolding l r ta_res_res'_eq[symmetric] ctxt_apply_term.simps(1) by fast
    qed
  next
    case (More f ss1 C' ss2)
    obtain g ss1' ss2' l' where s: "ss1 = map var2state ss1'" "ss2 = map var2state ss2'"
        "var2state l' = C'\<langle>l\<rangle>" "s = Fun g (ss1' @ l' # ss2')" "f = Inl g"
      using More(2) map_eq_Cons_conv[of var2state _ "C'\<langle>l\<rangle>" ss2]
            map_eq_append_split[of var2state _ ss1 "C'\<langle>l\<rangle> # ss2"]
      by (auto elim!: var2state.elims) metis
    obtain ts where ts: "t = Fun g ts"
        "map var2state ss1' @ C'\<langle>r\<rangle> # map var2state ss2' = map var2state ts"
      using More(3) s(1,2,5) by (auto elim!: var2state.elims)
    obtain r' where t: "var2state r' = C'\<langle>r\<rangle>"
      by (metis map_eq_Cons_conv map_eq_append_splits(2) ts(2))
    have t_def: "t = Fun g (ss1' @ r' # ss2')"
      using ts s(1,2) inj_map_eq_map[OF injI[of var2state], of "ss1' @ r' # ss2'" ts]
        var2state_inj unfolding t[symmetric] by auto
    have "r' \<in> ta_res' (trs_to_ta \<A> Q\<^sub>f) l'" using More(1)[OF s(3) t(1)] .
    moreover have "length (ss1' @ r' # ss2') = length (ss1' @ l' # ss2')" by simp
    ultimately show ?case unfolding s(4) t_def
      by simp (metis (mono_tags, lifting) ta_res'_refl append_Cons_nth_not_middle nth_append_length)
  qed
qed

lemma ta_res'_trans:
  assumes "t \<in> ta_res' \<A> s" "u \<in> ta_res' \<A> t"
  shows "u \<in> ta_res' \<A> s"
using assms
proof (induction arbitrary: t u rule: ta_res'.induct)
  case (2 \<A> f ts) show ?case using 2(2-)
  proof (induction t)
    case (Var q)
    hence "q \<in> ta_res \<A> (Fun f ts)" by auto
    moreover obtain q' where "u = Var q'" "(q, q') \<in> (ta_eps \<A>)\<^sup>*" using Var(2) by auto
    ultimately show ?case using ta_res_ctxt[of q \<A> "Fun f ts" _ \<box>] by simp
  next
    case (Fun g ss)
    have ss_props: "g = f" "length ss = length ts"
         "\<forall>i < length ts. ss ! i \<in> ta_res' \<A> (ts ! i)" using Fun(2) by fastforce+
    consider (uVar) "\<exists>p. u = Var p" | (uFun) "\<exists>us. u = Fun f us \<and> length us = length ss"
      using Fun(3) unfolding \<open>g = f\<close> by auto
    thus ?case
    proof cases
      case uVar
      then obtain p where p: "u = Var p" by blast
      obtain qs p' where qs_props: "(g qs \<rightarrow> p') \<in> ta_rules \<A> \<and>
          (p',p) \<in> (ta_eps \<A>)\<^sup>* \<and> length qs = length ss \<and>
          (\<forall>i < length ss. qs ! i \<in> ta_res \<A> (ss ! i))"
        using Fun(3) unfolding p by auto
      hence "i < length ts \<longrightarrow> Var (qs ! i) \<in> ta_res' \<A> (ts ! i)" for i
        using 2(1)[of i "ss ! i" "Var (qs ! i)"] ss_props(3)
        unfolding ss_props(2) ta_res_res'_eq by blast
      thus ?thesis using qs_props unfolding p ss_props ta_res_res'_eq by auto
    next
      case uFun
      then obtain us where u: "u = Fun f us" "length us = length ss" by blast
      have "\<forall>i < length ss. us ! i \<in> ta_res' \<A> (ss ! i)"
        using Fun(3) u(2) unfolding u(1) \<open>g = f\<close> by fastforce
      hence "i < length ts \<longrightarrow> us ! i \<in> ta_res' \<A> (ts ! i)" for i
        using 2(1) ss_props(3) unfolding ss_props(2) by blast
      thus ?thesis using u unfolding ss_props(2) by simp
    qed
  qed
qed force

lemma steps_in_ta2':
  assumes "(var2state s, t') \<in> (rstep \<A>)\<^sup>*" "t' = var2state t" "is_ta \<A>"
  shows "t \<in> ta_res' (trs_to_ta \<A> Q\<^sub>f) s"
using assms(1,2)
proof (induction arbitrary: t rule: rtrancl_induct)
  case base
  hence "s = t" using var2state_inj by blast
  thus ?case by simp
next
  case (step y z)
  obtain y' where y_def: "y = var2state y'"
    using rsteps_var2state_persists[OF step(1) assms(3)] by blast
  hence step1: "y' \<in> ta_res' (trs_to_ta \<A> Q\<^sub>f) s" using step(3) by blast
  show ?case using step(2,4) step_in_ta2'[of y' z \<A> t] assms(3)
    ta_res'_trans[OF step1, of t] y_def by blast
qed

lemma step_in_ta2:
  assumes "(var2state s, q') \<in> rstep \<A>" "q' = State q" "is_ta \<A>"
  shows "q \<in> ta_res (trs_to_ta \<A> Q\<^sub>f) s"
proof -
  obtain C l r where step: "(l, r) \<in> \<A>" "ta_trs_rule (l, r)"
      "var2state s = C\<langle>l\<rangle>" "q' = C\<langle>r\<rangle>" "is_state r"
    using ta_trs_stepE[OF assms(1,3)] by blast
  have "C = \<box>" using step(4) Fun_Nil_supt unfolding assms(2) by fastforce
  consider (state) "is_state l" | (fqterm) "\<exists>f qs. l = fqterm f qs"
    using step(2) by (auto elim!: ta_trs_rule.cases)
  thus ?thesis
  proof cases
    case state
    then obtain p where l_def: "l = State p" by (auto elim: is_state.cases)
    hence "q \<in> ta_res (trs_to_ta \<A> Q\<^sub>f) (Var p)"
      using step(1,3,4) eps_steps[of p "State q" \<A> q, OF _ _ assms(3)]
      unfolding assms(2) \<open>C = \<box>\<close> by auto
    thus ?thesis using step(3) unfolding \<open>C = \<box>\<close> l_def
      by simp (metis (mono_tags, lifting) move_eps move_reachable_states
         r_into_rtrancl ta_res_reachable_states_eq)
  next
    case fqterm
    then obtain f qs where "l = fqterm f qs" by blast
    thus ?thesis using step(1,4) rule_in_ta[OF assms(3), of f qs] assms(2)
      unfolding \<open>C = \<box>\<close>
      by simp (metis (mono_tags, lifting) step(3,4) move.trans move_reachable_states
          r_into_rtrancl ta_res_reachable_states_eq)
  qed
qed

lemma steps_in_ta2:
  assumes "(var2state s, q') \<in> (rstep \<A>)\<^sup>*" "q' = State q" "is_ta \<A>"
  shows "q \<in> ta_res (trs_to_ta \<A> Q\<^sub>f) s"
using assms(1,2)
proof (induction arbitrary: q rule: rtrancl_induct)
  case (step y z)
  obtain y' where y_def: "y = var2state y'"
    using rsteps_var2state_persists[OF step(1) assms(3)] by blast
  hence "y' \<in> ta_res' (trs_to_ta \<A> Q\<^sub>f) s" using steps_in_ta2'[OF step(1) _ assms(3)] by fast
  moreover have "q \<in> ta_res (trs_to_ta \<A> Q\<^sub>f) y'"
    using step(2,4) step_in_ta2[OF _ _ assms(3)] unfolding y_def by blast
  ultimately show ?case using ta_res'_ta_res by fast
qed (auto elim: var2state.elims)

lemma ta_res_adapt_vars:
  assumes "q \<in> ta_res \<A> t" "ground t"
  shows "q \<in> ta_res \<A> (adapt_vars t)"
using assms
proof (induction \<A> t arbitrary: q rule: ta_res.induct)
  case (2 \<A> f ts)
  obtain p qs where res: "f qs \<rightarrow> p \<in> ta_rules \<A>" "(p, q) \<in> ((ta_eps \<A>)\<^sup>*)"
    "length qs = length ts" "\<forall>i < length ts. qs ! i \<in> (map (ta_res \<A>) ts) ! i"
    using 2(2) by auto
  { fix i
    assume "i < length ts"
    hence "qs ! i \<in> ta_res \<A> (adapt_vars (ts ! i))"
      using 2(1)[of i "ts ! i" "qs ! i"] 2(3) res by simp
  }
  then show ?case using res by auto
qed simp

lemma ta_lang_trs:
  fixes \<A> :: "('f + 'q, 'v) trs" and t :: "('f, 'q) term"
  assumes "is_ta \<A>" "ground t"
  shows "t \<in> ta_lang (trs_to_ta \<A> Q\<^sub>f) \<longleftrightarrow> ta_accepts_term \<A> Q\<^sub>f (var2state t)"
proof -
  let ?\<A> = "trs_to_ta \<A> Q\<^sub>f"
  show ?thesis
  proof rule
    assume asm: "t \<in> ta_lang ?\<A>"
    then obtain t' q\<^sub>f where t'_q\<^sub>f: "t = adapt_vars t'" "ground t'"
        "q\<^sub>f \<in> ta_final ?\<A>" "q\<^sub>f \<in> ta_res ?\<A> t'"
      by (meson ta_langE)
    hence "q\<^sub>f \<in> ta_res ?\<A> t" "q\<^sub>f \<in> Q\<^sub>f" using ta_res_adapt_vars unfolding make_def by fastforce+
    hence "(var2state t, State q\<^sub>f) \<in> (rstep \<A>)\<^sup>*" using steps_in_ta1[OF _ assms(1)] by blast
    moreover have "ground (var2state t)"
      using ground_fterm_var2_state t'_q\<^sub>f(2) unfolding t'_q\<^sub>f(1)
      by (metis ground_adapt_vars ground_map_funs_term)
    ultimately show "ta_accepts_term \<A> Q\<^sub>f (var2state t)"
     using \<open>q\<^sub>f \<in> Q\<^sub>f\<close> unfolding ta_accepts_term_def by blast
  next
    assume asm: "ta_accepts_term \<A> Q\<^sub>f (var2state t)"
    obtain t' :: "('f, 'q) term" where t_def: "t = adapt_vars t'" "ground t'"
      by (metis adapt_vars_adapt_vars ground_adapt_vars assms(2))
    obtain q\<^sub>f where q\<^sub>f: "q\<^sub>f \<in> Q\<^sub>f" "(var2state t, State q\<^sub>f) \<in> (rstep \<A>)\<^sup>*"
      using asm unfolding ta_accepts_term_def by blast
    hence "q\<^sub>f \<in> ta_final ?\<A>" unfolding make_def by auto
    moreover have "q\<^sub>f \<in> ta_res ?\<A> (adapt_vars t)"
      using ta_res_adapt_vars[OF steps_in_ta2[OF q\<^sub>f(2) _ assms(1)] assms(2)] by blast
    ultimately show "t \<in> ta_lang ?\<A>"
      using ta_langI2[OF assms(2)] by fast
  qed
qed

lemma steps_in_ta:
  assumes "is_ta \<A>"
  shows "q \<in> ta_res (trs_to_ta \<A> Q\<^sub>f) s \<longleftrightarrow> (var2state s, State q) \<in> (rstep \<A>)\<^sup>*"
using steps_in_ta1[OF _ assms] steps_in_ta2[OF _ _ assms] by blast

lemma steps_in_ta':
  assumes "is_ta \<A>"
  shows "t \<in> ta_res' (trs_to_ta \<A> Q\<^sub>f) s \<longleftrightarrow> (var2state s, var2state t) \<in> (rstep \<A>)\<^sup>*"
using steps_in_ta1'[OF _ assms] steps_in_ta2'[OF _ _ assms] by blast

definition eps_ta_trs :: "('q \<times> 'q) set \<Rightarrow> ('f + 'q, 'v) trs" where
  "eps_ta_trs S = { (State x, State y) | x y. (x,y) \<in> S }"

definition rules_ta_trs :: "('q, 'f) ta_rule set \<Rightarrow> ('f + 'q, 'v) trs" where
  "rules_ta_trs S = { (fqterm f qs, State r) | f qs r. f qs \<rightarrow> r \<in> S }"

lemma ta_rules'_inv:
  assumes "\<forall>rule \<in> S. \<exists>f qs r. rule = f qs \<rightarrow> r"
  shows "ta_rules' (rules_ta_trs S) = S"
  using assms unfolding rules_ta_trs_def
  by simp (smt Collect_cong Collect_mem_eq list.inj_map_strong not_arg_cong_Inr term.inject(2))

lemma ta_rules'_eps: "ta_rules' (eps_ta_trs S) = {}"
  unfolding eps_ta_trs_def by fast

lemma ta_rules'_Union: "ta_rules' (\<A> \<union> \<B>) = ta_rules' \<A> \<union> ta_rules' \<B>"
  by blast

definition ta_to_trs :: "('q,'f) ta \<Rightarrow> ('f + 'q, 'v) trs \<times> 'q set" where
  "ta_to_trs \<A> = (eps_ta_trs (ta_eps \<A>) \<union> rules_ta_trs (ta_rules \<A>), ta_final \<A>)"

lemma ta_to_trs_id':
  fixes \<A>' :: "('q,'f) ta"
  assumes \<A>_def: "\<A> = fst (ta_to_trs \<A>')" and
          Q\<^sub>f_def: "Q\<^sub>f = snd (ta_to_trs \<A>')"
  shows "trs_to_ta \<A> Q\<^sub>f = \<A>' \<and> is_ta \<A>"
proof -
  have "ta_eps (trs_to_ta \<A> Q\<^sub>f) = ta_eps \<A>'"
    unfolding eps_ta_trs_def make_def rules_ta_trs_def ta_to_trs_def assms by force
  moreover have "ta_rules (trs_to_ta \<A> Q\<^sub>f) = ta_rules \<A>'"
  proof -
    have asm: "\<forall>t \<in> (ta_rules \<A>'). \<exists>f qs r. t = f qs \<rightarrow> r"
      by (meson r_sym.cases)
    have "ta_rules (trs_to_ta \<A> Q\<^sub>f) = ta_rules (trs_to_ta (rules_ta_trs (ta_rules \<A>')) Q\<^sub>f)"
      unfolding \<A>_def Q\<^sub>f_def make_def
        ta_rules'_Union[of "eps_ta_trs (ta_eps \<A>')" "rules_ta_trs (ta_rules \<A>')"]
        ta_rules'_eps[of "(ta_eps \<A>')"] ta_rules'_inv[OF asm] ta_to_trs_def fst_conv
      by simp
    thus ?thesis unfolding make_def ta_rules'_inv[OF asm] by simp
  qed
  moreover have "ta_final (trs_to_ta \<A> Q\<^sub>f) = ta_final \<A>'"
    unfolding make_def Q\<^sub>f_def ta_to_trs_def by simp
  ultimately have "trs_to_ta \<A> Q\<^sub>f = \<A>'" by fastforce
  thus "trs_to_ta \<A> Q\<^sub>f = \<A>' \<and> is_ta \<A>"
    unfolding \<A>_def Q\<^sub>f_def fst_conv snd_conv eps_ta_trs_def rules_ta_trs_def ta_to_trs_def
    by blast
qed

lemma trs_to_ta_id:
  "trs_to_ta (fst (ta_to_trs \<A>)) (snd (ta_to_trs \<A>)) = \<A>"
  "is_ta (fst (ta_to_trs \<A>))"
  using ta_to_trs_id'[of "fst (ta_to_trs \<A>)" \<A> "snd (ta_to_trs \<A>)"] by blast+

end
