{-# LANGUAGE StandaloneDeriving #-}

import qualified GCR as GCR

import qualified Data.Rewriting.Problem as P
import qualified Data.Rewriting.Rule as R
import qualified Data.Rewriting.Rules as Rs
import qualified Data.Rewriting.Term as T
import System.Environment
import Data.Char (ord)

deriving instance Show GCR.Answer -- YES | NO | MAYBE

main = do
    [f] <- getArgs
    p <- P.parseFileIO f
    let trs = P.allRules (P.rules p)
        -- convert to IsaFoR types (see below)
        trs' = trsConv trs
        -- use the TRS's natural signature
        GCR.Set sig = GCR.funas_trs (GCR.Set trs')
        -- invoke GGCR decision procedure
        res = GCR.lv_gcr_procedure_wrap sig trs'
    print res

-- convert to IsaFoR types

-- use Nat for Char (mapping to Unicode code points)
charConv :: Char -> GCR.Nat
charConv = GCR.nat_of_integer . fromIntegral . ord

stringConv :: String -> [GCR.Nat]
stringConv = map charConv

-- the two term types are isomorphic
type GCR_Term = GCR.Term [GCR.Nat] [GCR.Nat]

termConv :: T.Term String String -> GCR_Term
termConv = T.fold GCR.Var GCR.Fun . T.map stringConv stringConv

-- for rules, the IsaFoR type is a pair.
trsConv :: [R.Rule String String] -> [(GCR_Term, GCR_Term)]
trsConv = map (\r -> (termConv (R.lhs r), termConv (R.rhs r)))
